@@ -20,11 +20,12 @@
   #+gencgc sb-vm:dynamic-space-start
   #-gencgc (extern-alien "current_dynamic_space" unsigned-long))
 
-#+(or x86 x86-64)
+#+gencgc
 (progn
   (declaim (inline dynamic-space-free-pointer))
   (defun dynamic-space-free-pointer ()
-    (extern-alien "dynamic_space_free_pointer" system-area-pointer)))
+    (sap+ (int-sap (current-dynamic-space-start))
+          (* (extern-alien "next_free_page" signed) sb-vm:gencgc-page-bytes))))
 
 (declaim (inline dynamic-usage))
 (defun dynamic-usage ()
@@ -62,6 +63,13 @@
 (define-load-time-global *n-bytes-freed-or-purified* 0)
 (defun gc-reinit ()
   (setq *gc-inhibit* nil)
+  ;; This GC looks utterly unnecessary, but when I tried removing it,
+  ;; I got heap exhaustions sooner than when I left it in.
+  ;; So why is that? Well, it seems like there's some state that is getting
+  ;; saved into the core which needs resetting, such as these SETQs,
+  ;; but how is that affecting the C runtime's decision on when to
+  ;; perform the first auto-triggered GC? How is not doing an explicit
+  ;; GC delaying the decision? It seems wrong.
   (gc)
   (setf *n-bytes-freed-or-purified* 0
         *gc-run-time* 0))
