@@ -122,7 +122,7 @@
 
 ;;; a holder for the FASL file we're reading from
 (defstruct (fasl-input (:conc-name %fasl-input-)
-                       (:constructor make-fasl-input (stream))
+                       (:constructor make-fasl-input (stream print))
                        (:predicate nil)
                        (:copier nil))
   (stream nil :type ansi-stream :read-only t)
@@ -135,7 +135,8 @@
   ;; function calls) while executing other FOPs. SKIP-UNTIL will
   ;; either contain the position where the skipping will stop, or
   ;; NIL if we're executing normally.
-  (skip-until nil :type (or null fixnum)))
+  (skip-until nil :type (or null fixnum))
+  (print nil :type boolean))
 (declaim (freeze-type fasl-input))
 
 ;;; Output the current number of semicolons after a fresh-line.
@@ -681,17 +682,7 @@
 ;;; Return true if we successfully load a group from the stream, or
 ;;; NIL if EOF was encountered while trying to read from the stream.
 ;;; Dispatch to the right function for each fop.
-;;;
-;;; When true, PRINT causes most tlf-equivalent forms to print their primary value.
-;;; This differs from loading of Lisp source, which prints all values of
-;;; only truly-toplevel forms.  This is permissible per CLHS -
-;;;  "If print is true, load incrementally prints information to standard
-;;;   output showing the progress of the loading process. [...]
-;;;   For a compiled file, what is printed might not reflect precisely the
-;;;   contents of the source file, but some information is generally printed."
-;;;
-(defun load-fasl-group (fasl-input print)
-  (declare (ignorable print))
+(defun load-fasl-group (fasl-input)
   (let ((stream (%fasl-input-stream fasl-input))
         (trace *show-fops-p*))
     (unless (check-fasl-header stream)
@@ -733,10 +724,7 @@
                (setf (%fasl-input-deprecated-stuff fasl-input) nil)
                (loader-deprecation-warn
                 it
-                (and (eq (svref stack 1) 'sb-impl::%defun) (svref stack 2))))
-             (when print
-               (load-fresh-line)
-               (prin1 result)))))))))
+                (and (eq (svref stack 1) 'sb-impl::%defun) (svref stack 2)))))))))))
 
 ;; This is the moral equivalent of a warning from /usr/bin/ld that
 ;; "gets() is dangerous." You're informed by both the compiler and linker.
@@ -757,13 +745,14 @@
   (when (zerop (file-length stream))
     (error "attempt to load an empty FASL file:~%  ~S" (namestring stream)))
   (maybe-announce-load stream verbose)
-  (let ((fasl-input (make-fasl-input stream)))
-    (unwind-protect
-         (loop while (load-fasl-group fasl-input print))
-      ;; Nuke the table and stack to avoid keeping garbage on
-      ;; conservatively collected platforms.
-      (nuke-fop-vector (%fasl-input-table fasl-input))
-      (nuke-fop-vector (%fasl-input-stack fasl-input))))
+  (let ((fasl-input (make-fasl-input stream print)))
+    (with-deferred-package-names
+      (unwind-protect
+           (loop while (load-fasl-group fasl-input))
+        ;; Nuke the table and stack to avoid keeping garbage on
+        ;; conservatively collected platforms.
+        (nuke-fop-vector (%fasl-input-table fasl-input))
+        (nuke-fop-vector (%fasl-input-stack fasl-input)))))
   t)
 
 
@@ -795,10 +784,9 @@
 
 ;; %MAKE-INSTANCE does not exist on the host.
 (define-fop 48 :not-host (fop-struct ((:operands size) layout))
-  (let ((res (%make-instance size)) ; number of words excluding header
+  (let ((res (sb-kernel::%new-instance* layout size)) ; number of words excluding header
         ;; Discount the layout from number of user-visible words.
         (n-data-words (- size sb-vm:instance-data-start)))
-    (setf (%instance-wrapper res) layout)
     (with-fop-stack ((stack (operand-stack)) ptr n-data-words)
       (declare (type index ptr))
       (let ((bitmap (wrapper-bitmap layout)))
@@ -824,25 +812,22 @@
   (decf depthoid) ; was bumped by 1 since non-stack args can't encode negatives
   (sb-kernel::load-layout name depthoid inherits length bitmap flags))
 
-;; Allocate a CLOS object. This is used when the compiler detects that
-;; MAKE-LOAD-FORM returned a simple use of MAKE-LOAD-FORM-SAVING-SLOTS,
-;; or possibly a hand-written equivalent (however unlikely).
-(define-fop 68 :not-host (fop-allocate-instance (name) nil)
-  (let ((instance (allocate-instance (find-class (the symbol name)))))
-    (push-fop-table instance (fasl-input))))
-
-;; Fill in object slots as dictated by the second return value from
-;; MAKE-LOAD-FORM-SAVING-SLOTS.
-(define-fop 69 :not-host (fop-set-slot-values ((:operands n-slots) slot-names obj) nil)
-  (let* ((stack (operand-stack))
-         (ptr (fop-stack-pop-n stack n-slots)))
-      (dotimes (i n-slots)
-        (let ((val (svref stack (+ ptr i)))
-              (slot-name (pop slot-names)))
-          (if (unbound-marker-p val)
-              ;; SLOT-MAKUNBOUND-USING-CLASS might do something nonstandard.
-              (slot-makunbound obj slot-name)
-              (setf (slot-value obj slot-name) val))))))
+;;; This is dumped when the compiler detects that MAKE-LOAD-FORM
+;;; returned a simple use of MAKE-LOAD-FORM-SAVING-SLOTS, or possibly
+;;; a hand-written equivalent (however unlikely).
+(define-fop 68 :not-host (fop-instance ((:operands n-slots) name))
+  (let* ((instance (allocate-instance (find-class (the symbol name))))
+         (stack (operand-stack))
+         (ptr (fop-stack-pop-n stack (* 2 n-slots))))
+    (dotimes (i n-slots)
+      (let* ((index (+ ptr (* 2 i)))
+             (value (svref stack index))
+             (slot-name (svref stack (1+ index))))
+        (if (unbound-marker-p value)
+            ;; SLOT-MAKUNBOUND-USING-CLASS might do something nonstandard.
+            (slot-makunbound instance slot-name)
+            (setf (slot-value instance slot-name) value))))
+    instance))
 
 (define-fop 64 (fop-end-group () nil)
   (throw 'fasl-group-end t))
@@ -856,10 +841,6 @@
 
 ;;;; fops for loading symbols
 
-(defstruct (undefined-package (:copier nil))
-  (error nil :read-only t))
-(declaim (freeze-type undefined-package))
-
 ;;; Cold load has its own implementation of all symbol fops,
 ;;; but we have to execute define-fop now to assign their numbers.
 ;;;
@@ -887,14 +868,8 @@
          (aux-fop-intern (length+flag package fasl-input)
            (multiple-value-bind (name length elt-type)
                (read-symbol-name length+flag fasl-input)
-             (if (undefined-package-p package)
-                 (error 'simple-package-error
-                        :format-control "Error finding package for symbol ~s:~% ~a"
-                        :format-arguments
-                        (list (subseq name 0 length)
-                              (undefined-package-error package)))
-                 (push-fop-table (%intern name length package elt-type t)
-                                 fasl-input))))
+             (push-fop-table (%intern name length package elt-type t)
+                             fasl-input)))
          (ensure-hashed (symbol)
            ;; ENSURE-SYMBOL-HASH when vop-translated is flushable since it is
            ;; conceptually just a slot reader, however its actual effect is to fill in
@@ -931,8 +906,9 @@
     (read-char-string-as-varints (fasl-input-stream) package-name)
     (push-fop-table
      (handler-case (find-undeleted-package-or-lose package-name)
-       (simple-package-error (c)
-         (make-undefined-package :error (princ-to-string c))))
+       (simple-package-error (condition)
+         (declare (ignore condition))
+         (make-deferred-package package-name)))
      (fasl-input))))
 
 ;;;; fops for loading numbers
@@ -1014,8 +990,8 @@
   (with-fast-read-byte ((unsigned-byte 8) (fasl-input-stream))
     (let ((tag (fast-read-s-integer 8)))
       (cond #+sb-simd-pack-256
-            ((logbitp 2 tag)
-             (%make-simd-pack-256 (logand tag #b11)
+            ((logbitp 6 tag)
+             (%make-simd-pack-256 (logand tag #b00111111)
                                   (fast-read-u-integer 8)
                                   (fast-read-u-integer 8)
                                   (fast-read-u-integer 8)
@@ -1109,22 +1085,25 @@
 
 ;;;; fops for fixing up circularities
 
-(define-fop 11 (fop-rplaca ((:operands tbl-slot idx) val) nil)
+(define-fop 11 (fop-rplaca ((:operands tbl-slot index) value) nil)
   (let ((obj (ref-fop-table (fasl-input) tbl-slot)))
-    (setf (car (nthcdr idx obj)) val)))
+    (setf (car (nthcdr index obj)) value)))
 
-(define-fop 12 (fop-rplacd ((:operands tbl-slot idx) val) nil)
+(define-fop 12 (fop-rplacd ((:operands tbl-slot index) value) nil)
   (let ((obj (ref-fop-table (fasl-input) tbl-slot)))
-    (setf (cdr (nthcdr idx obj)) val)))
+    (setf (cdr (nthcdr index obj)) value)))
 
-(define-fop 13 (fop-svset ((:operands tbl-slot idx) val) nil)
-  (setf (svref (ref-fop-table (fasl-input) tbl-slot) idx) val))
+(define-fop 13 (fop-svset ((:operands tbl-slot index) value) nil)
+  (setf (svref (ref-fop-table (fasl-input) tbl-slot) index) value))
 
-(define-fop 14 :not-host (fop-structset ((:operands tbl-slot idx) val) nil)
-  (%instance-set (ref-fop-table (fasl-input) tbl-slot) idx val)
-  val) ; I can't remmeber whether this fop needs to return VAL. maybe?
+(define-fop 14 :not-host (fop-structset ((:operands tbl-slot index) value) nil)
+  (%instance-set (ref-fop-table (fasl-input) tbl-slot) index value))
 
-(define-fop 15 (fop-nthcdr ((:operands n) obj))
+(define-fop 15 :not-host (fop-slotset ((:operands tbl-slot index) value slot-name) nil)
+  index
+  (setf (slot-value (ref-fop-table (fasl-input) tbl-slot) slot-name) value))
+
+(define-fop 16 (fop-nthcdr ((:operands n) obj))
   (nthcdr n obj))
 
 ;;;; fops for loading functions
@@ -1137,7 +1116,7 @@
 ;;; fasl file header.)
 
 (define-load-time-global *show-new-code* nil)
-(define-fop 16 :not-host (fop-load-code ((:operands header n-code-bytes n-fixups)))
+(define-fop 17 :not-host (fop-load-code ((:operands header n-code-bytes n-fixups)))
   (let* ((n-simple-funs (read-unsigned-byte-32-arg (fasl-input-stream)))
          (n-named-calls (read-unsigned-byte-32-arg (fasl-input-stream)))
          (n-boxed-words (ash header -1))
@@ -1207,13 +1186,13 @@
 
 ;; this gets you an #<fdefn> object, not the result of (FDEFINITION x)
 ;; cold-loader uses COLD-FDEFINITION-OBJECT instead.
-(define-fop 17 :not-host (fop-fdefn (name))
+(define-fop 18 :not-host (fop-fdefn (name))
   (awhen (deprecated-thing-p 'function name) ; returns the stage of deprecation
     (pushnew (list* it name :function)
              (%fasl-input-deprecated-stuff (fasl-input)) :test 'equal))
   (find-or-create-fdefn name))
 
-(define-fop 18 :not-host (fop-known-fun (name))
+(define-fop 19 :not-host (fop-known-fun (name))
   (%coerce-name-to-fun name))
 
 ;;; This FOP is only encountered in cross-compiled FASLs for cold load,
@@ -1229,19 +1208,41 @@
   (declare (ignore name qualifiers specializer fn)))
 
 ;;; Modify a slot of the code boxed constants.
-(define-fop 19 (fop-alter-code ((:operands index) code value) nil)
+(define-fop 20 (fop-alter-code ((:operands index) code value) nil)
   (flet (#+sb-xc-host
          ((setf code-header-ref) (value code index)
             (write-wordindexed code index value)))
     (setf (code-header-ref code index) value)
     (values)))
 
-(define-fop 20 (fop-fun-entry ((:operands fun-index) code-object))
-  (%code-entry-point code-object fun-index))
+;;; Set the named constant value in the boxed constants, setting up
+;;; backpatching information if the symbol is not yet bound. Forward
+;;; references can occur at load time when non-top-level components
+;;; containing named constant references get loaded before the top
+;;; level form defining the constant gets loaded. This can happen when
+;;; top level lambdas get merged.
+#-sb-xc-host
+(defun named-constant-set (code index name)
+  (cond ((boundp name)
+         (setf (code-header-ref code index) (symbol-global-value name)))
+        (t
+         (push (lambda (value)
+                 (setf (code-header-ref code index) value))
+               (info :variable :forward-references name)))))
+
+(define-fop 121 :not-host (fop-named-constant-set ((:operands index) name code) nil)
+  (named-constant-set code index name))
+
+(define-fop 21 (fop-fun-entry ((:operands fun-index) code-object))
+  (let ((fun (%code-entry-point code-object fun-index)))
+    (when (%fasl-input-print (fasl-input))
+      (load-fresh-line)
+      (format t "~S loaded" fun))
+    fun))
 
 ;;;; assemblerish fops
 
-(define-fop 21 (fop-assembler-code)
+(define-fop 22 (fop-assembler-code)
   (error "cannot load assembler code except at cold load"))
 
 ;;; FOPs needed for implementing an IF operator in a FASL
