@@ -10,230 +10,13 @@
 ;;;; files for more information.
 
 (in-package "SB-KERNEL")
-
-;;;; the NUMBER-DISPATCH macro
 
-(eval-when (:compile-toplevel :load-toplevel :execute)
-
-;;; Grovel an individual case to NUMBER-DISPATCH, augmenting RESULT
-;;; with the type dispatches and bodies. Result is a tree built of
-;;; alists representing the dispatching off each arg (in order). The
-;;; leaf is the body to be executed in that case.
-(defun parse-number-dispatch (vars result types var-types body)
-  ;; Shouldn't be necessary, but avoids a warning in certain lisps that
-  ;; seem to like to warn about self-calls in :COMPILE-TOPLEVEL situation.
-  (named-let parse-number-dispatch ((vars vars) (result result) (types types)
-                                    (var-types var-types) (body body))
-    (cond ((null vars)
-           (unless (null types) (error "More types than vars."))
-           (when (cdr result)
-             (error "Duplicate case: ~S." body))
-           (setf (cdr result)
-                 (sublis var-types body :test #'equal)))
-          ((null types)
-           (error "More vars than types."))
-          (t
-           (flet ((frob (var type)
-                    (parse-number-dispatch
-                     (rest vars)
-                     (or (assoc type (cdr result) :test #'equal)
-                         (car (setf (cdr result)
-                                    (acons type nil (cdr result)))))
-                     (rest types)
-                     (acons `(dispatch-type ,var) type var-types)
-                     body)))
-             (let ((type (first types))
-                   (var (first vars)))
-               (if (and (consp type) (eq (first type) 'foreach))
-                   (dolist (type (rest type))
-                     (frob var type))
-                   (frob var type))))))))
-
-;;; our guess for the preferred order in which to do type tests
-;;; (cheaper and/or more probable first.)
-(defconstant-eqx +type-test-ordering+
-  '(fixnum single-float double-float integer #+long-float long-float
-    sb-vm:signed-word word bignum
-    complex ratio)
-  #'equal)
-
-;;; Should TYPE1 be tested before TYPE2?
-(defun type-test-order (type1 type2)
-  (let ((o1 (position type1 +type-test-ordering+))
-        (o2 (position type2 +type-test-ordering+)))
-    (cond ((not o1) nil)
-          ((not o2) t)
-          (t
-           (< o1 o2)))))
-
-;;; Return an ETYPECASE form that does the type dispatch, ordering the
-;;; cases for efficiency.
-;;; Check for some simple to detect problematic cases where the caller
-;;; used types that are not disjoint and where this may lead to
-;;; unexpected behaviour of the generated form, for example making
-;;; a clause unreachable, and throw an error if such a case is found.
-;;; An example:
-;;;   (number-dispatch ((var1 integer) (var2 float))
-;;;     ((fixnum single-float) a)
-;;;     ((integer float) b))
-;;; Even though the types are not reordered here, the generated form,
-;;; basically
-;;;   (etypecase var1
-;;;     (fixnum (etypecase var2
-;;;               (single-float a)))
-;;;     (integer (etypecase var2
-;;;                (float b))))
-;;; would fail at runtime if given var1 fixnum and var2 double-float,
-;;; even though the second clause matches this signature. To catch
-;;; this earlier than runtime we throw an error already here.
-(defun generate-number-dispatch (vars error-tags cases)
-  ;; Shouldn't be necessary, but avoids a warning in certain lisps that
-  ;; seem to like to warn about self-calls in :COMPILE-TOPLEVEL situation.
-  (named-let generate-number-dispatch ((vars vars) (error-tags error-tags) (cases cases))
-    (if vars
-        (let ((var (first vars))
-              (cases (sort cases #'type-test-order :key #'car)))
-          (flet ((error-if-sub-or-supertype (type1 type2)
-                   (when (or (sb-xc:subtypep type1 type2)
-                             (sb-xc:subtypep type2 type1))
-                     (error "Types not disjoint: ~S ~S." type1 type2)))
-                 (error-if-supertype (type1 type2)
-                   (when (sb-xc:subtypep type2 type1)
-                     (error "Type ~S ordered before subtype ~S."
-                            type1 type2)))
-                 (test-type-pairs (fun)
-                   ;; Apply FUN to all (ordered) pairs of types from the
-                   ;; cases.
-                   (mapl (lambda (cases)
-                           (when (cdr cases)
-                             (let ((type1 (caar cases)))
-                               (dolist (case (cdr cases))
-                                 (funcall fun type1 (car case))))))
-                         cases)))
-            ;; For the last variable throw an error if a type is followed
-            ;; by a subtype, for all other variables additionally if a
-            ;; type is followed by a supertype.
-            (test-type-pairs (if (cdr vars)
-                                 #'error-if-sub-or-supertype
-                                 #'error-if-supertype)))
-          `((typecase ,var
-              ,@(mapcar (lambda (case)
-                          `(,(first case)
-                            ,@(generate-number-dispatch (rest vars)
-                                                        (rest error-tags)
-                                                        (cdr case))))
-                        cases)
-              (t (go ,(first error-tags))))))
-        cases)))
-
-) ; EVAL-WHEN
-
-;;; This is a vaguely case-like macro that does number cross-product
-;;; dispatches. The Vars are the variables we are dispatching off of.
-;;; The Type paired with each Var is used in the error message when no
-;;; case matches. Each case specifies a Type for each var, and is
-;;; executed when that signature holds. A type may be a list
-;;; (FOREACH Each-Type*), causing that case to be repeatedly
-;;; instantiated for every Each-Type. In the body of each case, any
-;;; list of the form (DISPATCH-TYPE Var-Name) is substituted with the
-;;; type of that var in that instance of the case.
-;;;
-;;; [Though it says "_any_ list", it's still an example of how not to perform
-;;; incomplete lexical analysis within a macro imho. Let's say that the body
-;;; code passes a lambda that happens name its args DISPATCH-TYPE and X.
-;;; What happens?
-;;; (macroexpand-1 '(number-dispatch ((x number))
-;;;                  ((float) (f x (lambda (dispatch-type x) (wat))))))
-;;; -> [stuff elided]
-;;;      (TYPECASE X (FLOAT (F X (LAMBDA FLOAT (WAT))))
-;;;
-;;; So the NUMBER-DISPATCH macro indeed substituted for *any* appearance
-;;; just like it says. I wonder if we could define DISPATCH-TYPE as macrolet
-;;; that expands to the type for the current branch, so that it _must_
-;;; be in a for-evaluation position; but maybe I'm missing something?]
-;;;
-;;; As an alternate to a case spec, there may be a form whose CAR is a
-;;; symbol. In this case, we apply the CAR of the form to the CDR and
-;;; treat the result of the call as a list of cases. This process is
-;;; not applied recursively.
-;;;
-;;; Be careful when using non-disjoint types in different cases for the
-;;; same variable. Some uses will behave as intended, others not, as the
-;;; variables are dispatched off sequentially and clauses are reordered
-;;; for efficiency. Some, but not all, problematic cases are detected
-;;; and lead to a compile time error; see GENERATE-NUMBER-DISPATCH above
-;;; for an example.
-(defmacro number-dispatch (var-specs &body cases)
-  (let ((res (list nil))
-        (vars (mapcar #'car var-specs))
-        (block (gensym)))
-    (dolist (case cases)
-      (if (symbolp (first case))
-          (let ((cases (apply (symbol-function (first case)) (rest case))))
-            (dolist (case cases)
-              (parse-number-dispatch vars res (first case) nil (rest case))))
-          (parse-number-dispatch vars res (first case) nil (rest case))))
-
-    (collect ((errors)
-              (error-tags))
-      (dolist (spec var-specs)
-        (let ((var (first spec))
-              (type (second spec))
-              (tag (gensym)))
-          (error-tags tag)
-          (errors tag)
-          (errors
-           (sb-c::internal-type-error-call var type))))
-
-      `(block ,block
-         (tagbody
-            (return-from ,block
-              ,@(generate-number-dispatch vars (error-tags)
-                                          (cdr res)))
-            ,@(errors))))))
-
-;;;; binary operation dispatching utilities
-
-(eval-when (:compile-toplevel :execute)
-
-;;; Return NUMBER-DISPATCH forms for rational X float.
-(defun float-contagion (op x y &optional (rat-types '(fixnum bignum ratio)))
-  `(((single-float single-float) (,op ,x ,y))
-    (((foreach ,@rat-types)
-      (foreach single-float double-float #+long-float long-float))
-     (,op (coerce ,x '(dispatch-type ,y)) ,y))
-    (((foreach single-float double-float #+long-float long-float)
-      (foreach ,@rat-types))
-     (,op ,x (coerce ,y '(dispatch-type ,x))))
-    #+long-float
-    (((foreach single-float double-float long-float) long-float)
-     (,op (coerce ,x 'long-float) ,y))
-    #+long-float
-    ((long-float (foreach single-float double-float))
-     (,op ,x (coerce ,y 'long-float)))
-    (((foreach single-float double-float) double-float)
-     (,op (coerce ,x 'double-float) ,y))
-    ((double-float single-float)
-     (,op ,x (coerce ,y 'double-float)))))
-
-;;; Return NUMBER-DISPATCH forms for bignum X fixnum.
-(defun bignum-cross-fixnum (fix-op big-op)
-  `(((fixnum fixnum) (,fix-op x y))
-    ((fixnum bignum)
-     (,big-op (make-small-bignum x) y))
-    ((bignum fixnum)
-     (,big-op x (make-small-bignum y)))
-    ((bignum bignum)
-     (,big-op x y))))
-
-) ; EVAL-WHEN
-
 ;;;; canonicalization utilities
 
 ;;; If IMAGPART is 0, return REALPART, otherwise make a complex. This is
 ;;; used when we know that REALPART and IMAGPART are the same type, but
 ;;; rational canonicalization might still need to be done.
-#-sb-fluid (declaim (inline canonical-complex))
+(declaim (inline canonical-complex))
 (defun canonical-complex (realpart imagpart)
   (if (eql imagpart 0)
       realpart
@@ -253,7 +36,7 @@
 ;;; Given a numerator and denominator with the GCD already divided
 ;;; out, make a canonical rational. We make the denominator positive,
 ;;; and check whether it is 1.
-#-sb-fluid (declaim (inline build-ratio))
+(declaim (inline build-ratio))
 (defun build-ratio (num den)
   (multiple-value-bind (num den)
       (if (minusp den)
@@ -268,7 +51,7 @@
       (t (%make-ratio num den)))))
 
 ;;; Truncate X and Y, but bum the case where Y is 1.
-#-sb-fluid (declaim (inline maybe-truncate))
+(declaim (inline maybe-truncate))
 (defun maybe-truncate (x y)
   (if (eql y 1)
       x
@@ -419,6 +202,7 @@
 
 (defmacro two-arg-+/- (name op big-op)
   `(defun ,name (x y)
+     (declare (explicit-check))
      (number-dispatch ((x number) (y number))
        (bignum-cross-fixnum ,op ,big-op)
        (float-contagion ,op x y)
@@ -463,6 +247,7 @@
 (two-arg-+/- two-arg-- - subtract-bignum)
 
 (defun two-arg-* (x y)
+  (declare (explicit-check))
   (flet ((integer*ratio (x y)
            (if (eql x 0) 0
                (let* ((ny (numerator y))
@@ -479,8 +264,7 @@
            (canonical-complex (* (realpart x) y) (* (imagpart x) y))))
     (number-dispatch ((x number) (y number))
       (float-contagion * x y)
-
-      ((fixnum fixnum) (multiply-fixnums x y))
+      ((fixnum fixnum) (* x y))
       ((bignum fixnum) (multiply-bignum-and-fixnum x y))
       ((fixnum bignum) (multiply-bignum-and-fixnum y x))
       ((bignum bignum) (multiply-bignums x y))
@@ -635,71 +419,134 @@
 ;;; ROUND and FROUND are not declared inline since they seem too
 ;;; obscure and too big to inline-expand by default. Also, this gives
 ;;; the compiler a chance to pick off the unary float case.
-#-sb-fluid (declaim (inline fceiling ffloor ftruncate))
-(defun ftruncate (number &optional (divisor 1))
-  "Same as TRUNCATE, but returns first value as a float."
-  (declare (explicit-check))
-  (macrolet ((ftruncate-float (rtype)
-               `(let* ((float-div (coerce divisor ',rtype))
-                       (res (%unary-ftruncate (/ number float-div))))
-                  (values res
-                          (- number
-                             (* (coerce res ',rtype) float-div))))))
-    (number-dispatch ((number real) (divisor real))
-      (((foreach fixnum bignum ratio) (or fixnum bignum ratio))
-       (multiple-value-bind (q r)
-           (truncate number divisor)
-         (values (float q) r)))
-      (((foreach single-float double-float #+long-float long-float)
-        (or rational single-float))
-       (if (eql divisor 1)
-           (let ((res (%unary-ftruncate number)))
-             (values res (- number (coerce res '(dispatch-type number)))))
-           (ftruncate-float (dispatch-type number))))
-      #+long-float
-      ((long-float (or single-float double-float long-float))
-       (ftruncate-float long-float))
-      #+long-float
-      (((foreach double-float single-float) long-float)
-       (ftruncate-float long-float))
-      ((double-float (or single-float double-float))
-       (ftruncate-float double-float))
-      ((single-float double-float)
-       (ftruncate-float double-float))
-      (((foreach fixnum bignum ratio)
-        (foreach single-float double-float #+long-float long-float))
-       (ftruncate-float (dispatch-type divisor))))))
+(declaim (inline fceiling ffloor ftruncate))
 
-(defun ffloor (number &optional (divisor 1))
-  "Same as FLOOR, but returns first value as a float."
-  (declare (explicit-check))
-  (multiple-value-bind (tru rem) (ftruncate number divisor)
-    (if (and (not (zerop rem))
-             (if (minusp divisor)
-                 (plusp number)
-                 (minusp number)))
-        (values (1- tru) (+ rem divisor))
-        (values tru rem))))
+#-round-float
+(progn
+  (defun ftruncate (number &optional (divisor 1))
+    "Same as TRUNCATE, but returns first value as a float."
+    (declare (explicit-check))
+    (macrolet ((ftruncate-float (rtype)
+                 `(let* ((float-div (coerce divisor ',rtype))
+                         (res (%unary-ftruncate (/ number float-div))))
+                    (values res
+                            (- number
+                               (* (coerce res ',rtype) float-div))))))
+      (number-dispatch ((number real) (divisor real))
+        (((foreach fixnum bignum ratio) (or fixnum bignum ratio))
+         (multiple-value-bind (q r)
+             (truncate number divisor)
+           (values (float q) r)))
+        (((foreach single-float double-float #+long-float long-float)
+          (or rational single-float))
+         (if (eql divisor 1)
+             (let ((res (%unary-ftruncate number)))
+               (values res (- number (coerce res '(dispatch-type number)))))
+             (ftruncate-float (dispatch-type number))))
+        #+long-float
+        ((long-float (or single-float double-float long-float))
+         (ftruncate-float long-float))
+        #+long-float
+        (((foreach double-float single-float) long-float)
+         (ftruncate-float long-float))
+        ((double-float (or single-float double-float))
+         (ftruncate-float double-float))
+        ((single-float double-float)
+         (ftruncate-float double-float))
+        (((foreach fixnum bignum ratio)
+          (foreach single-float double-float #+long-float long-float))
+         (ftruncate-float (dispatch-type divisor))))))
 
-(defun fceiling (number &optional (divisor 1))
-  "Same as CEILING, but returns first value as a float."
-  (declare (explicit-check))
-  (multiple-value-bind (tru rem) (ftruncate number divisor)
-    (if (and (not (zerop rem))
-             (if (minusp divisor)
-                 (minusp number)
-                 (plusp number)))
-        (values (+ tru 1) (- rem divisor))
-        (values tru rem))))
+  (defun ffloor (number &optional (divisor 1))
+    "Same as FLOOR, but returns first value as a float."
+    (declare (explicit-check))
+    (multiple-value-bind (tru rem) (ftruncate number divisor)
+      (if (and (not (zerop rem))
+               (if (minusp divisor)
+                   (plusp number)
+                   (minusp number)))
+          (values (1- tru) (+ rem divisor))
+          (values tru rem))))
+
+  (defun fceiling (number &optional (divisor 1))
+    "Same as CEILING, but returns first value as a float."
+    (declare (explicit-check))
+    (multiple-value-bind (tru rem) (ftruncate number divisor)
+      (if (and (not (zerop rem))
+               (if (minusp divisor)
+                   (minusp number)
+                   (plusp number)))
+          (values (+ tru 1) (- rem divisor))
+          (values tru rem))))
 
 ;;; FIXME: this probably needs treatment similar to the use of
 ;;; %UNARY-FTRUNCATE for FTRUNCATE.
-(defun fround (number &optional (divisor 1))
-  "Same as ROUND, but returns first value as a float."
-  (declare (explicit-check))
-  (multiple-value-bind (res rem)
-      (round number divisor)
-    (values (float res (if (floatp rem) rem $1.0)) rem)))
+  (defun fround (number &optional (divisor 1))
+    "Same as ROUND, but returns first value as a float."
+    (declare (explicit-check))
+    (multiple-value-bind (res rem)
+        (round number divisor)
+      (values (float res (if (floatp rem) rem $1.0)) rem))))
+
+#+round-float
+(macrolet ((def (name mode docstring)
+             `(defun ,name (number &optional (divisor 1))
+                ,docstring
+                (declare (explicit-check))
+                (macrolet ((ftruncate-float (rtype)
+                             `(let* ((float-div (coerce divisor ',rtype))
+                                     (res (,(case rtype
+                                              (double-float 'sb-kernel:round-double)
+                                              (single-float 'sb-kernel:round-single))
+                                           (/ number float-div)
+                                           ,,mode)))
+                                (values res
+                                        (- number
+                                           (* (coerce res ',rtype) float-div)))))
+                           (unary-ftruncate-float (rtype)
+                             `(let* ((res (,(case rtype
+                                              (double-float 'sb-kernel:round-double)
+                                              (single-float 'sb-kernel:round-single))
+                                           number
+                                           ,,mode)))
+                                (values res (- number res)))))
+                  (number-dispatch ((number real) (divisor real))
+                    (((foreach fixnum bignum ratio) (or fixnum bignum ratio))
+                     (multiple-value-bind (q r)
+                         (,(find-symbol (string mode) :cl) number divisor)
+                       (values (float q) r)))
+                    (((foreach single-float double-float)
+                      (or rational single-float))
+                     (if (eql divisor 1)
+                         (unary-ftruncate-float (dispatch-type number))
+                         (ftruncate-float (dispatch-type number))))
+                    ((double-float (or single-float double-float))
+                     (ftruncate-float double-float))
+                    ((single-float double-float)
+                     (ftruncate-float double-float))
+                    (((foreach fixnum bignum ratio)
+                      (foreach single-float double-float))
+                     (ftruncate-float (dispatch-type divisor))))))))
+  (def ftruncate :truncate
+       "Same as TRUNCATE, but returns first value as a float.")
+
+  (def ffloor :floor
+       "Same as FLOOR, but returns first value as a float.")
+
+  (def fceiling :ceiling
+       "Same as CEILING, but returns first value as a float.")
+  (def fround :round
+    "Same as ROUND, but returns first value as a float.")
+
+  (macrolet ((def (name)
+               `(defun ,name (x mode)
+                  (ecase mode
+                    ,@(loop for m in '(:round :floor :ceiling :truncate)
+                            collect `(,m (,name x ,m)))))))
+
+
+    (def round-single)
+    (def round-double)))
 
 ;;;; comparisons
 
@@ -794,7 +641,7 @@ the first."
                           ((>= ,integer 0)
                            (< (float quot ,float) ,float))))))))))
 
-(eval-when (:compile-toplevel :execute)
+(eval-when (:compile-toplevel :load-toplevel :execute)
 ;;; The INFINITE-X-FINITE-Y and INFINITE-Y-FINITE-X args tell us how
 ;;; to handle the case when X or Y is a floating-point infinity and
 ;;; the other arg is a rational. (Section 12.1.4.1 of the ANSI spec
@@ -820,7 +667,6 @@ the first."
            (,op x (coerce 0 '(dispatch-type x)))
            (if (float-infinity-p x)
                ,infinite-x-finite-y
-               ;; Likewise
                (make-fixnum-float-comparer ,(case op
                                               (> '<)
                                               (< '>)
@@ -830,41 +676,79 @@ the first."
        (,op (coerce x 'double-float) y))
       ((double-float single-float)
        (,op x (coerce y 'double-float)))
-      (((foreach single-float double-float #+long-float long-float) rational)
-       (if (eql y 0)
-           (,op x (coerce 0 '(dispatch-type x)))
-           (if (float-infinity-p x)
-               ,infinite-x-finite-y
-               (,op (rational x) y))))
-      (((foreach bignum fixnum ratio) float)
+      (((foreach single-float double-float #+long-float long-float) ratio)
+       (if (float-infinity-p x)
+           ,infinite-x-finite-y
+           ;; Avoid converting the float into a rational, since it
+           ;; will be taken apart later anyway.
+           (multiple-value-bind (bits exp) (integer-decode-float x)
+             (if (eql bits 0)
+                 (,op 0 y)
+                 (let ((int (if (minusp x) (- bits) bits)))
+                   (if (minusp exp)
+                       (,op (* int (denominator y))
+                            (ash (numerator y) (- exp)))
+                       (,op (ash int exp) y)))))))
+      ((ratio (foreach single-float double-float))
        (if (float-infinity-p y)
            ,infinite-y-finite-x
+           (multiple-value-bind (bits exp) (integer-decode-float y)
+             (if (eql bits 0)
+                 (,op x 0)
+                 (let ((int (if (minusp y) (- bits) bits)))
+                   (if (minusp exp)
+                       (,op (ash (numerator x) (- exp))
+                            (* int (denominator x)))
+                       (,op x (ash int exp))))))))
+      (((foreach single-float double-float) bignum)
+       (if (float-infinity-p x)
+           ,infinite-x-finite-y
+           #+64-bit
+           (,(case op
+               (> 'float-bignum->)
+               (< 'float-bignum-<)
+               (= 'float-bignum-=))
+            x y)
+           #-64-bit
+           (,op (rational x) y)))
+      ((bignum float)
+       (if (float-infinity-p y)
+           ,infinite-y-finite-x
+           #+64-bit
+           (,(case op
+               (> 'float-bignum-<)
+               (< 'float-bignum->)
+               (= 'float-bignum-=))
+            y x)
+           #-64-bit
            (,op x (rational y))))))
   )                                     ; EVAL-WHEN
 
 
 (macrolet ((def-two-arg-</> (name op ratio-arg1 ratio-arg2 &rest cases)
              `(defun ,name (x y)
+                (declare (inline float-infinity-p)
+                         (explicit-check))
                 (number-dispatch ((x real) (y real))
-                                 (basic-compare
-                                  ,op
-                                  :infinite-x-finite-y
-                                  (,op x (coerce 0 '(dispatch-type x)))
-                                  :infinite-y-finite-x
-                                  (,op (coerce 0 '(dispatch-type y)) y))
-                                 (((foreach fixnum bignum) ratio)
-                                  (,op x (,ratio-arg2 (numerator y)
-                                                      (denominator y))))
-                                 ((ratio integer)
-                                  (,op (,ratio-arg1 (numerator x)
-                                                    (denominator x))
-                                       y))
-                                 ((ratio ratio)
-                                  (,op (* (numerator   (truly-the ratio x))
-                                          (denominator (truly-the ratio y)))
-                                       (* (numerator   (truly-the ratio y))
-                                          (denominator (truly-the ratio x)))))
-                                 ,@cases))))
+                  (basic-compare
+                   ,op
+                   :infinite-x-finite-y
+                   (,op x (coerce 0 '(dispatch-type x)))
+                   :infinite-y-finite-x
+                   (,op (coerce 0 '(dispatch-type y)) y))
+                  (((foreach fixnum bignum) ratio)
+                   (,op x (,ratio-arg2 (numerator y)
+                                       (denominator y))))
+                  ((ratio integer)
+                   (,op (,ratio-arg1 (numerator x)
+                                     (denominator x))
+                        y))
+                  ((ratio ratio)
+                   (,op (* (numerator   (truly-the ratio x))
+                           (denominator (truly-the ratio y)))
+                        (* (numerator   (truly-the ratio y))
+                           (denominator (truly-the ratio x)))))
+                  ,@cases))))
   (def-two-arg-</> two-arg-< < floor ceiling
     ((fixnum bignum)
      (bignum-plus-p y))
@@ -881,6 +765,8 @@ the first."
      (plusp (bignum-compare x y)))))
 
 (defun two-arg-= (x y)
+  (declare (inline float-infinity-p)
+           (explicit-check))
   (number-dispatch ((x number) (y number))
     (basic-compare =
                    ;; An infinite value is never equal to a finite value.
@@ -934,37 +820,37 @@ the first."
     (fixnum (lognot (truly-the fixnum number)))
     (bignum (bignum-logical-not number))))
 
-(macrolet ((def (name explicit-check op big-op &optional doc)
+(macrolet ((def (name op big-op &optional doc)
              `(defun ,name (integer1 integer2)
                 ,@(when doc (list doc))
-                ,@(when explicit-check `((declare (explicit-check))))
+                (declare (explicit-check))
                 (let ((x integer1)
                       (y integer2))
                   (number-dispatch ((x integer) (y integer))
                     (bignum-cross-fixnum ,op ,big-op))))))
-  (def two-arg-and nil logand bignum-logical-and)
-  (def two-arg-ior nil logior bignum-logical-ior)
-  (def two-arg-xor nil logxor bignum-logical-xor)
+  (def two-arg-and logand bignum-logical-and)
+  (def two-arg-ior logior bignum-logical-ior)
+  (def two-arg-xor logxor bignum-logical-xor)
   ;; BIGNUM-LOGICAL-{AND,IOR,XOR} need not return a bignum, so must
   ;; call the generic LOGNOT...
-  (def two-arg-eqv nil logeqv (lambda (x y) (lognot (bignum-logical-xor x y))))
-  (def lognand t lognand
+  (def two-arg-eqv logeqv (lambda (x y) (lognot (bignum-logical-xor x y))))
+  (def lognand lognand
        (lambda (x y) (lognot (bignum-logical-and x y)))
        "Complement the logical AND of INTEGER1 and INTEGER2.")
-  (def lognor t lognor
+  (def lognor lognor
        (lambda (x y) (lognot (bignum-logical-ior x y)))
        "Complement the logical OR of INTEGER1 and INTEGER2.")
   ;; ... but BIGNUM-LOGICAL-NOT on a bignum will always return a bignum
-  (def logandc1 t logandc1
+  (def logandc1 logandc1
        (lambda (x y) (bignum-logical-and (bignum-logical-not x) y))
        "Bitwise AND (LOGNOT INTEGER1) with INTEGER2.")
-  (def logandc2 t logandc2
+  (def logandc2 logandc2
        (lambda (x y) (bignum-logical-and x (bignum-logical-not y)))
        "Bitwise AND INTEGER1 with (LOGNOT INTEGER2).")
-  (def logorc1 t logorc1
+  (def logorc1 logorc1
        (lambda (x y) (bignum-logical-ior (bignum-logical-not x) y))
        "Bitwise OR (LOGNOT INTEGER1) with INTEGER2.")
-  (def logorc2 t logorc2
+  (def logorc2 logorc2
        (lambda (x y) (bignum-logical-ior x (bignum-logical-not y)))
        "Bitwise OR INTEGER1 with (LOGNOT INTEGER2)."))
 
@@ -976,8 +862,8 @@ and the number of 0 bits if INTEGER is negative."
     (fixnum
      (logcount #-x86-64
                (truly-the (integer 0
-                                   #.(max sb-xc:most-positive-fixnum
-                                          (lognot sb-xc:most-negative-fixnum)))
+                                   #.(max most-positive-fixnum
+                                          (lognot most-negative-fixnum)))
                           (if (minusp (truly-the fixnum integer))
                               (lognot (truly-the fixnum integer))
                               integer))
@@ -992,6 +878,7 @@ and the number of 0 bits if INTEGER is negative."
 
 (defun logbitp (index integer)
   "Predicate returns T if bit index of integer is a 1."
+  (declare (explicit-check))
   (number-dispatch ((index integer) (integer integer))
     ((fixnum fixnum) (if (< index sb-vm:n-positive-fixnum-bits)
                          (not (zerop (logand integer (ash 1 index))))
@@ -1219,7 +1106,7 @@ and the number of 0 bits if INTEGER is negative."
                 (values n m))
           (* (truncate max (gcd n m)) min)))))
 
-(declaim (inline fixnum-gcd))
+(declaim (maybe-inline fixnum-gcd))
 (defun fixnum-gcd (u v)
   (declare (optimize (safety 0)))
   (locally
@@ -1248,7 +1135,8 @@ and the number of 0 bits if INTEGER is negative."
 ;;; of 0 before the dispatch so that the bignum code doesn't have to worry
 ;;; about "small bignum" zeros.
 (defun two-arg-gcd (u v)
-  (declare (muffle-conditions compiler-note))
+  (declare (muffle-conditions compiler-note)
+           (inline fixnum-gcd))
   (cond ((eql u 0) (abs v))
         ((eql v 0) (abs u))
         (t
@@ -1267,6 +1155,7 @@ and the number of 0 bits if INTEGER is negative."
 ;;; bignum case, we don't bother, since bignum division is expensive,
 ;;; and the test is not very likely to succeed.
 (defun integer-/-integer (x y)
+  (declare (explicit-check))
   (if (and (typep x 'fixnum) (typep y 'fixnum))
       (multiple-value-bind (quo rem) (truncate x y)
         (if (zerop rem)
@@ -1282,10 +1171,10 @@ and the number of 0 bits if INTEGER is negative."
             (if (eql gcd 1)
                 (values x y)
                 (values (truncate x gcd) (truncate y gcd)))
-          (build-ratio num (truly-the (not (integer 0 1)) den))))))
-(declaim (notinline fixnum-gcd))
+          (build-ratio num den)))))
 
 (defun two-arg-/ (x y)
+  (declare (explicit-check))
   (number-dispatch ((x number) (y number))
     (float-contagion / x y (ratio integer))
 
@@ -1378,7 +1267,7 @@ and the number of 0 bits if INTEGER is negative."
                    (significant-half (ash ,arg (- (ash fourth-size 1))))
                    (significant-half-isqrt
                      (if-fixnum-p-truly-the
-                      (integer 1 #.(isqrt sb-xc:most-positive-fixnum))
+                      (integer 1 #.(isqrt most-positive-fixnum))
                       (,recurse significant-half)))
                    (zeroth-iteration (ash significant-half-isqrt
                                           fourth-size)))
@@ -1461,10 +1350,6 @@ and the number of 0 bits if INTEGER is negative."
       (do-mfuns sb-c::*tagged-modular-class*)))
   `(progn ,@(sort (forms) #'string< :key #'cadr)))
 
-;;; KLUDGE: these out-of-line definitions can't use the modular
-;;; arithmetic, as that is only (currently) defined for constant
-;;; shifts.  See also the comment in (LOGAND OPTIMIZER) for more
-;;; discussion of this hack.  -- CSR, 2003-10-09
 #-(or 64-bit 64-bit-registers)
 (defun sb-vm::ash-left-mod32 (integer amount)
   (etypecase integer
@@ -1485,3 +1370,16 @@ and the number of 0 bits if INTEGER is negative."
     (etypecase integer
       (fixnum (sb-c::mask-signed-field fixnum-width (ash integer amount)))
       (integer (sb-c::mask-signed-field fixnum-width (ash (sb-c::mask-signed-field fixnum-width integer) amount))))))
+
+(sb-c::when-vop-existsp (:translate sb-vm::ash-modfx)
+  (defun sb-vm::ash-mod64 (integer amount)
+    (etypecase integer
+      ((unsigned-byte 64) (ldb (byte 64 0) (ash integer amount)))
+      (fixnum (ldb (byte 64 0) (ash (logand integer #xffffffffffffffff) amount)))
+      (bignum (ldb (byte 64 0)
+                   (ash (logand integer #xffffffffffffffff) amount)))))
+
+  (defun sb-vm::ash-modfx (integer amount)
+    (if (minusp integer)
+        (sb-c::mask-signed-field sb-vm:n-fixnum-bits (ash integer amount))
+        (logand most-positive-fixnum (ash integer amount)))))
