@@ -14,16 +14,11 @@
 
 ;;;; type format database
 
-(defstruct (room-info (:constructor make-room-info (mask name kind))
+;;; FIXME: this structure seems to no longer serve a purpose.
+;;; We'd do as well with a simple-vector of (or symbol cons saetp).
+(defstruct (room-info (:constructor make-room-info (name))
                       (:copier nil))
-    ;; the mask applied to HeaderValue to compute object size
-    (mask 0 :type (and fixnum unsigned-byte) :read-only t)
-    ;; the name of this type
-    (name nil :type symbol :read-only t)
-    ;; kind of type (how to reconstitute an object)
-    (kind nil
-          :type (member :other :closure :instance :list :code :fdefn)
-          :read-only t))
+    (name nil :type symbol :read-only t)) ; the name of this type
 (declaim (freeze-type room-info))
 
 (defun room-info-type-name (info)
@@ -33,58 +28,30 @@
 
 (defconstant tiny-boxed-size-mask #xFF)
 (defun compute-room-infos ()
-  (let ((infos (make-array 256 :initial-element nil))
-        (default-size-mask (mask-field (byte 23 0) -1)))
+  (let ((infos (make-array 256 :initial-element nil)))
     (dolist (obj *primitive-objects*)
       (let ((widetag (primitive-object-widetag obj))
             (lowtag (primitive-object-lowtag obj))
             (name (primitive-object-name obj)))
-        (when (and (eq lowtag 'other-pointer-lowtag)
-                   (not (member widetag '(t nil))))
-          (setf (svref infos (symbol-value widetag))
-                (case name
-                 (fdefn  (make-room-info 0 name :fdefn))
-                 (symbol (make-room-info tiny-boxed-size-mask name :other))
-                 (t      (make-room-info default-size-mask name :other)))))))
+        (when (and (member lowtag '(other-pointer-lowtag fun-pointer-lowtag
+                                    instance-pointer-lowtag))
+                   (not (member widetag '(t nil simple-fun-widetag))))
+          (setf (svref infos (symbol-value widetag)) (make-room-info name)))))
 
-    (let ((info (make-room-info tiny-boxed-size-mask 'array-header :other)))
+    (let ((info (make-room-info 'array-header)))
       (dolist (code (list #+sb-unicode complex-character-string-widetag
                           complex-base-string-widetag simple-array-widetag
                           complex-bit-vector-widetag complex-vector-widetag
                           complex-array-widetag))
         (setf (svref infos code) info)))
 
-    (setf (svref infos bignum-widetag)
-          ;; Lose 1 more bit than n-widetag-bits because fullcgc robs 1 bit,
-          ;; not that this is expected to work concurrently with gc.
-          (make-room-info (ash most-positive-word (- (1+ n-widetag-bits)))
-                          'bignum :other))
-    (setf (svref infos filler-widetag)
-          (make-room-info (ash most-positive-word (- (1+ n-widetag-bits)))
-                          'filler :other))
-
-    (setf (svref infos closure-widetag)
-          (make-room-info short-header-max-words 'closure :closure))
+    (setf (svref infos filler-widetag) (make-room-info 'filler))
 
     (dotimes (i (length *specialized-array-element-type-properties*))
       (let ((saetp (aref *specialized-array-element-type-properties* i)))
-        (when (saetp-specifier saetp) ;; SIMPLE-ARRAY-NIL is a special case.
-          (setf (svref infos (saetp-typecode saetp)) saetp))))
+        (setf (svref infos (saetp-typecode saetp)) saetp)))
 
-    ;; This one is here for completeness only- LENGTH does not imply size.
-    (setf (svref infos simple-array-nil-widetag)
-          (make-room-info 0 'simple-array-nil :other))
-
-    (setf (svref infos code-header-widetag)
-          (make-room-info 0 'code :code))
-
-    (setf (svref infos instance-widetag)
-          (make-room-info 0 'instance :instance))
-
-    (setf (svref infos funcallable-instance-widetag)
-          (make-room-info short-header-max-words 'funcallable-instance :closure))
-
-    (let ((cons-info (make-room-info 0 'cons :list)))
+    (let ((cons-info (make-room-info 'cons)))
       ;; A cons consists of two words, both of which may be either a
       ;; pointer or immediate data.  According to the runtime this means
       ;; either a fixnum, a character, an unbound-marker, a single-float
@@ -146,7 +113,9 @@
 (defun %space-bounds (space)
   (macrolet ((bounds (a b) `(values (%make-lisp-obj ,a) (%make-lisp-obj ,b))))
     (ecase space
-      (:static
+      (:static ;; These bounds are appropriate for computing the space usage
+               ;; but NOT for computing iteration bounds, because there are
+               ;; "unformatted" words preceding the lowest addressable object.
        (bounds static-space-start
                (sap-int *static-space-free-pointer*)))
       (:read-only
@@ -188,65 +157,6 @@
                (ash header-word (- hash-slot-present-flag))
                1))))
 
-;;; * If symbols are 7 words incl header, as they are on 32-bit w/threads,
-;;;   then the part of NIL that manifests as symbol slots consumes 6 words
-;;;   (less the header) because NIL's symbol widetag precedes it by 1 word.
-;;;   So, there are 6 post-header words and 2 pre-header: one containing
-;;;   the widetag (not that it is ever read), and one 0 word.
-;;; * If symbols are 6 words incl header, as they are on 64-bit and
-;;;   32-bit w/o threads, then the symbol-like part of NIL is 5 words,
-;;;   which aligns up to 6, plus the two pre-header words.
-;;; So either way, NIL's alignment padding makes it come out to 8 words in total.
-(defconstant sizeof-nil-in-words (+ 2 (sb-int:align-up (1- sb-vm:symbol-size) 2)))
-
-;;; It's unclear to me whether reinventing sizetab[] in lisp is strictly
-;;; an improvement over doing the obvious:
-;;;  (WITH-PINNED-OBJECTS (object) (alien-funcall "ext_lispboj_size" ...))
-;;; It certainly doesn't feel that it's better, but it has significantly less overhead
-;;; at least on cheneygc if not also on the precise gencgc platforms.
-;;; But see WITH-PINNED-OBJECT-ITERATOR in 'target-hash-table' which offers
-;;; a way to mitigate the need to bind/unbind a special var when doing a massive
-;;; numbers of WITH-PINNED-OBJECT operations.
-(defun primitive-object-size (object)
-  "Return number of bytes of heap or stack directly consumed by OBJECT"
-  (unless (is-lisp-pointer (get-lisp-obj-address object))
-    (return-from primitive-object-size 0))
-  (let ((words
-         ;; This should pick off the most frequently-occurring things first.
-         ;; CONS and INSTANCE of course top the list.
-         (typecase object
-           (list (if object 2 sizeof-nil-in-words))
-           (instance (1+ (instance-length object)))
-           (function
-            (when (= (%fun-pointer-widetag object) simple-fun-widetag)
-              (return-from primitive-object-size
-                (code-object-size (fun-code-header (truly-the simple-fun object)))))
-            (1+ (get-closure-length object)))
-           ;; Must be an OTHER pointer now
-           (code-component
-            (return-from primitive-object-size (code-object-size object)))
-           (fdefn 4) ; constant length not stored in the header
-           ((satisfies array-header-p)
-            (+ array-dimensions-offset (array-rank object)))
-           ((simple-array nil (*)) 2) ; no payload
-           (bignum
-            #+bignum-assertions
-            (+ (* (sb-bignum:%bignum-length object) 2) 2)
-            #-bignum-assertions
-            ;; 64-bit machines might want to store the bignum length in the upper 4
-            ;; bytes of the header which would simplify %bignum-set-length.
-            (1+ (sb-bignum:%bignum-length object)))
-           (t
-            (let ((room-info (aref *room-info* (%other-pointer-widetag object))))
-              (if (typep room-info 'specialized-array-element-type-properties)
-                  (let ((n-data-octets (vector-n-data-octets object room-info)))
-                    (+ (ceiling n-data-octets n-word-bytes) ; N data words
-                       vector-data-offset))
-                  ;; GET-HEADER-DATA works for everything that's left
-                  (1+ (logand (get-header-data object)
-                              (room-info-mask room-info)))))))))
-        (* (align-up words 2) n-word-bytes)))
-
 ;;; Macros not needed after this file (and avoids a redefinition warning this way)
 (eval-when (:compile-toplevel)
 (defmacro widetag@baseptr (sap)
@@ -305,25 +215,6 @@
      #-sb-devel
      (aver (sap= start end)))))
 
-;;; Test the sizing function ASAP, because if it's broken, then MAP-ALLOCATED-OBJECTS
-;;; is too, and then creating the initial core will crash because of the various heap
-;;; traversals performed in SAVE-LISP-AND-DIE. Don't delay a crash.
-(dovector (saetp *specialized-array-element-type-properties*)
-  (let ((length 0) (et (saetp-specifier saetp)))
-    (loop (let* ((array (make-array length :element-type et))
-                 (size-from-lisp (primitive-object-size array))
-                 (size-from-c
-                  (with-pinned-objects (array)
-                    (alien-funcall (extern-alien "ext_lispobj_size" (function unsigned unsigned))
-                                   (logandc2 (get-lisp-obj-address array) lowtag-mask)))))
-            (unless (= size-from-lisp size-from-c)
-              (bug "size calculation mismatch on ~S" array))
-            ;; Stop after enough trials to hit all the edge case
-            (when (or (>= size-from-lisp (* 8 sb-vm:n-word-bytes))
-                      (and (eq et nil) (>= length 4))) ; always 2 words
-              (return))
-            (incf length)))))
-
 ;;; Access to the GENCGC page table for better precision in
 ;;; MAP-ALLOCATED-OBJECTS
 #+gencgc
@@ -338,13 +229,12 @@
               (start #+64-bit (unsigned 32) #-64-bit signed)
               ;; On platforms with small enough GC pages, this field
               ;; will be a short. On platforms with larger ones, it'll
-              ;; be an int.
-              ;; Measured in bytes; the low bit has to be masked off.
-              (bytes-used (unsigned
-                           #.(if (typep gencgc-card-bytes '(unsigned-byte 16))
+              ;; be an int. It should probably never be an int.
+              (words-used (unsigned
+                           #.(if (typep gencgc-page-words '(unsigned-byte 16))
                                  16
                                  32)))
-              (flags (unsigned 8))
+              (flags (unsigned 8)) ; in C this is {type, need_zerofill, pinned}
               (gen (signed 8))))
   #+immobile-space
   (progn
@@ -437,15 +327,10 @@ We could try a few things to mitigate this:
             (:static
              ;; Static space starts with NIL, which requires special
              ;; handling, as the header and alignment are slightly off.
-             (multiple-value-bind (start end) (%space-bounds space)
-               (declare (ignore start))
-               (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
-               (let ((start (+ (logandc2 sb-vm:nil-value sb-vm:lowtag-mask)
-                               (ash (- sizeof-nil-in-words 2) sb-vm:word-shift))))
-                 (map-objects-in-range fun
-                                       (ash start (- sb-vm:n-fixnum-tag-bits))
-                                       end))))
-
+             (funcall fun nil symbol-widetag (* sizeof-nil-in-words n-word-bytes))
+             (let ((start (%make-lisp-obj static-space-objects-start))
+                   (end (%make-lisp-obj (sap-int *static-space-free-pointer*))))
+               (map-objects-in-range fun start end)))
             ((:read-only #-gencgc :dynamic)
              ;; Read-only space (and dynamic space on cheneygc) is a block
              ;; of contiguous allocations.
@@ -490,7 +375,7 @@ We could try a few things to mitigate this:
        ;; in order to determine what regions contain objects.
 
        ;; We explicitly presume that any pages in an allocation region
-       ;; that are in-use have a BYTES-USED of GENCGC-CARD-BYTES
+       ;; that are in-use have a BYTES-USED of GENCGC-PAGE-BYTES
        ;; (indicating a full page) or an otherwise-valid BYTES-USED.
        ;; We also presume that the pages of an open allocation region
        ;; after the first page, and any pages that are unallocated,
@@ -498,7 +383,7 @@ We could try a few things to mitigate this:
 
        ;; Our procedure is to scan forward through the page table,
        ;; maintaining an "end pointer" until we reach a page where
-       ;; BYTES-USED is not GENCGC-CARD-BYTES or we reach
+       ;; BYTES-USED is not GENCGC-PAGE-BYTES or we reach
        ;; NEXT-FREE-PAGE.  We then MAP-OBJECTS-IN-RANGE if the range
        ;; is not empty, and proceed to the next page (unless we've hit
        ;; NEXT-FREE-PAGE).
@@ -508,7 +393,7 @@ We could try a few things to mitigate this:
        ;; prove to be an issue with concurrent systems, or with
        ;; spectacularly poor timing for closing an allocation region
        ;; in a single-threaded system.
-  (close-current-gc-region)
+  (close-thread-alloc-region)
   (do ((initial-next-free-page next-free-page)
        (base (int-sap (current-dynamic-space-start)))
        (start-page 0)
@@ -517,28 +402,27 @@ We could try a few things to mitigate this:
       ((> start-page initial-next-free-page))
          ;; The type constraint on page indices is probably too generous,
          ;; but it does its job of producing efficient code.
-    (declare (type (integer 0 (#.(/ (ash 1 n-machine-word-bits) gencgc-card-bytes)))
+    (declare (type (integer 0 (#.(/ (ash 1 n-machine-word-bits) gencgc-page-bytes)))
                    start-page end-page))
     (setq end-page start-page)
     (loop (setq end-page-bytes-used
-                ;; The low bit of bytes-used is the need-to-zero flag.
-                (logandc1 1 (slot (deref page-table end-page) 'bytes-used)))
+                (ash (slot (deref page-table end-page) 'words-used) word-shift))
           ;; See 'page_ends_contiguous_block_p' in gencgc.c
-          (when (or (< end-page-bytes-used gencgc-card-bytes)
+          (when (or (< end-page-bytes-used gencgc-page-bytes)
                     (= (slot (deref page-table (1+ end-page)) 'start) 0))
             (return))
           (incf end-page))
-    (let ((start (sap+ base (truly-the sb-vm:signed-word
-                                       (logand (* start-page gencgc-card-bytes)
+    (let ((start (sap+ base (truly-the signed-word
+                                       (logand (* start-page gencgc-page-bytes)
                                                most-positive-word))))
-          (end (sap+ base (truly-the sb-vm:signed-word
-                                     (logand (+ (* end-page gencgc-card-bytes)
+          (end (sap+ base (truly-the signed-word
+                                     (logand (+ (* end-page gencgc-page-bytes)
                                                 end-page-bytes-used)
                                              most-positive-word)))))
       (when (sap> end start)
-        ;; The bits in the 5-bit flag field have fixed positions,
+        ;; The bits in the 6-bit 'type' field have fixed positions,
         ;; but the position of the field itself depends on endianness.
-        (let ((flags (ldb #+little-endian (byte 5 0) #+big-endian (byte 5 3)
+        (let ((flags (ldb (byte 6 (+ #+big-endian 2))
                           (slot (deref page-table start-page) 'flags))))
           ;; The GEN slot is declared as (SIGNED 8) which does not satisfy the
           ;; type restriction on the first argument to LOGBITP.
@@ -553,27 +437,15 @@ We could try a few things to mitigate this:
                                   (< start-page initial-next-free-page))))))
     (setq start-page (1+ end-page))))
 
-;; Start with a Lisp rendition of ensure_region_closed() on the active
-;; thread's region, since users are often surprised to learn that a
-;; just-consed object can't necessarily be seen by MAP-ALLOCATED-OBJECTS.
+;; Users are often surprised to learn that a just-consed object can't
+;; necessarily be seen by MAP-ALLOCATED-OBJECTS, so close the region
+;; to update the page table.
 ;; Since we're in WITHOUT-GCING, there can be no interrupts.
-;; This is probably better than calling MAP-OBJECTS-IN-RANGE on the
-;; thread's region, because then we might visit some page twice
-;; by doing that.
-;; (And seeing small consing by other threads is hopeless either way)
-(defun close-current-gc-region ()
-  #+gencgc
-  (let ((region-struct
-          #+sb-thread (sap+ (sb-thread::current-thread-sap)
-                            (ash thread-boxed-tlab-slot word-shift))
-          ;; If no threads, the alloc_region struct is in static space.
-          #-sb-thread (int-sap boxed-region)))
-    ;; The 'start_addr' field is at word index 3 of the structure (see gencgc-alloc-region.h).
-    (unless (eql (sap-ref-word region-struct (ash 3 word-shift)) 0)
-      (alien-funcall (extern-alien "gc_close_region" (function void system-area-pointer int))
-                     region-struct
-                     1))
-    nil))
+;; Moreover it's probably not safe in the least to walk any thread's
+;; allocation region, unless the observer and observed aren't consing.
+(defun close-thread-alloc-region ()
+  #+gencgc (alien-funcall (extern-alien "close_current_thread_tlab" (function void)))
+  nil)
 
 ;;;; MEMORY-USAGE
 
@@ -847,6 +719,8 @@ We could try a few things to mitigate this:
            (type stream stream) (type spaces space)
            (type (or index null) type larger smaller count))
   (multiple-value-bind (start end) (%space-bounds space)
+    (when (eq space :static)
+      (setq start (%make-lisp-obj static-space-objects-start)))
     (let* ((space-start (ash start n-fixnum-tag-bits))
            (space-end (ash end n-fixnum-tag-bits))
            (space-size (- space-end space-start))
@@ -920,19 +794,6 @@ We could try a few things to mitigate this:
 
 ;;;; LIST-ALLOCATED-OBJECTS, LIST-REFERENCING-OBJECTS
 
-(defvar *ignore-after* nil)
-
-(defun valid-obj (space x)
-  (or (not (eq space :dynamic))
-      ;; this test looks bogus if the allocator doesn't work linearly,
-      ;; which I suspect is the case for GENCGC.  -- CSR, 2004-06-29
-      (< (get-lisp-obj-address x) (get-lisp-obj-address *ignore-after*))))
-
-(defun maybe-cons (space x stuff)
-  (if (valid-obj space x)
-      (cons x stuff)
-      stuff))
-
 (defun list-allocated-objects (space &key type larger smaller count
                                      test)
   (declare (type (or (eql :all) spaces) space)
@@ -1117,7 +978,7 @@ We could try a few things to mitigate this:
             ,.(make-case 'array)
             ,.(make-case* 'symbol
                `(,functoid (%primitive sb-c:fast-symbol-global-value ,obj) ,@more)
-               `(,functoid (symbol-info ,obj) ,@more)
+               `(,functoid (symbol-%info ,obj) ,@more)
                `(,functoid (symbol-name ,obj) ,@more)
                `(,functoid (symbol-package ,obj) ,@more)
                `(when (symbol-extra-slot-p ,obj)
@@ -1190,16 +1051,16 @@ We could try a few things to mitigate this:
 (defun map-referencing-objects (fun space object)
   (declare (type (or (eql :all) spaces) space))
   (declare (dynamic-extent fun))
-  (unless *ignore-after*
-    (setq *ignore-after* (cons 1 2)))
-  (let ((fun (%coerce-callable-to-fun fun)))
+  (let (list)
     (map-allocated-objects
      (lambda (referer widetag size)
        (declare (ignore widetag size))
-       (when (and (valid-obj space referer) ; semi-bogus!
+       ;; Don't count a self-reference as a reference
+       (when (and (neq referer object)
                   (references-p referer object))
-         (funcall fun referer)))
-     space)))
+         (push referer list)))
+     space)
+    (mapc (%coerce-callable-to-fun fun) list)))
 
 (defun list-referencing-objects (space object)
   (collect ((res))
@@ -1279,7 +1140,7 @@ We could try a few things to mitigate this:
                                                   :initial-element 0)))
   (flet ((dump-page (page-num)
            (format stream "~&Page ~D~%" page-num)
-           (let ((where (+ dynamic-space-start (* page-num gencgc-card-bytes)))
+           (let ((where (+ dynamic-space-start (* page-num gencgc-page-bytes)))
                  (seen-filler nil))
              (loop
                (let* ((obj (let ((sap (int-sap where)))
@@ -1298,7 +1159,7 @@ We could try a few things to mitigate this:
                  (loop for index from page-num to (find-page-index (1- where))
                        do (setf (sbit pages index) 1)))
                (let ((next-page (find-page-index where)))
-                 (cond ((= (logand where (1- gencgc-card-bytes)) 0)
+                 (cond ((= (logand where (1- gencgc-page-bytes)) 0)
                         (format stream "~&-- END OF PAGE --~%")
                         (return next-page))
                        ((eq next-page page-num))
@@ -1311,7 +1172,7 @@ We could try a few things to mitigate this:
                      (setq i (dump-page i))
                      (incf i)))))
     (let* ((n-pages (count 1 pages))
-           (tot (* n-pages gencgc-card-bytes))
+           (tot (* n-pages gencgc-page-bytes))
            (waste (- tot n-code-bytes)))
       (format t "~&Used-bytes=~D Pages=~D Waste=~D (~F%)~%"
               n-code-bytes n-pages waste
@@ -1336,17 +1197,18 @@ We could try a few things to mitigate this:
 
 #+gencgc
 (defun generation-of (object)
-  (let* ((addr (get-lisp-obj-address object))
-         (page (find-page-index addr)))
-    (cond ((>= page 0) (slot (deref page-table page) 'gen))
-          #+immobile-space
-          ((immobile-space-addr-p addr)
-           ;; SIMPLE-FUNs don't contain a generation byte
-           (when (simple-fun-p object)
-             (setq addr (get-lisp-obj-address (fun-code-header object))))
-           (let ((sap (int-sap (logandc2 addr lowtag-mask))))
-             (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
-                     #xF))))))
+  (with-pinned-objects (object)
+    (let* ((addr (get-lisp-obj-address object))
+           (page (find-page-index addr)))
+      (cond ((>= page 0) (slot (deref page-table page) 'gen))
+            #+immobile-space
+            ((immobile-space-addr-p addr)
+             ;; SIMPLE-FUNs don't contain a generation byte
+             (when (simple-fun-p object)
+               (setq addr (get-lisp-obj-address (fun-code-header object))))
+             (let ((sap (int-sap (logandc2 addr lowtag-mask))))
+               (logand (if (fdefn-p object) (sap-ref-8 sap 1) (sap-ref-8 sap 3))
+                       #xF)))))))
 
 ;;; Show objects in a much simpler way than print-allocated-objects.
 ;;; Probably don't use this for generation 0 of dynamic space. Other spaces are ok.
@@ -1364,7 +1226,10 @@ We could try a few things to mitigate this:
 ;;; Unfortunately this is a near total copy of the test in gc.impure.lisp
 (defun !ensure-genesis-code/data-separation ()
   #+gencgc
-  (let* ((n-bits (+ next-free-page 10))
+  (let* ((n-bits
+          (progn
+            (close-thread-alloc-region)
+            (+ next-free-page 50)))
          (code-bits (make-array n-bits :element-type 'bit :initial-element 0))
          (data-bits (make-array n-bits :element-type 'bit :initial-element 0))
          (total-code-size 0))
@@ -1382,6 +1247,20 @@ We could try a few things to mitigate this:
                                  data-bits)
                                 (t
                                  code-bits))))
+         ;; If Lisp allocates to cons pages, then this check is always valid.
+         ;; If not, then at first glance we could weaken the check to allow
+         ;; gen0 conses on MIXED pages, but even that is not enough- pinned conses
+         ;; will promote but keep their MIXED page type. So don't bother with this.
+         #+use-cons-region
+         (let* ((flags (slot (deref page-table (find-page-index obj-addr)) 'flags))
+                (type (ldb (byte 6 (+ #+big-endian 2)) flags))
+                (ok (if (consp obj)
+                        (or (= type #b101) ; PAGE_TYPE_CONS
+                            (and (eq (car obj) 0) (eq (cdr obj) 0)))
+                        (/= type #b101))))
+           (unless ok
+             (error "Object @ ~x (gen~D) is on page-type ~b~%"
+                    obj-addr (generation-of obj) type)))
          ;; This is not the most efficient way to update the bit arrays,
          ;; but the simplest and clearest for sure. (The loop could avoided
          ;; if the current page is the same as the previously seen page)
@@ -1392,7 +1271,7 @@ We could try a few things to mitigate this:
                do (cond ((= (sbit other-array index) 1)
                          (format t "~&broken on page index ~d base ~x~%"
                                  index
-                                 (+ dynamic-space-start (* index gencgc-card-bytes)))
+                                 (+ dynamic-space-start (* index gencgc-page-bytes)))
                          (alien-funcall (extern-alien "ldb_monitor" (function void))))
                         (t
                          (setf (sbit array index) 1))))))
@@ -1423,31 +1302,38 @@ We could try a few things to mitigate this:
     (with-pinned-objects (obj)
       (dotimes (i count)
         (let ((word (sap-ref-word (int-sap addr) (ash i word-shift))))
-          (multiple-value-bind (lispobj ok)
+          (multiple-value-bind (lispobj ok fmt)
               (cond ((and (typep thing 'code-component)
                           (< 1 i (code-header-words thing)))
                      (values (code-header-ref thing i) t))
+                    #+compact-symbol
+                    ((and (typep thing '(and symbol (not null)))
+                          (= i symbol-name-slot))
+                     (values (list (sb-impl::symbol-package-id thing)
+                                   (symbol-name thing))
+                             t
+                             "{~{~A,~S~}}"))
                     (decode
                      (make-lisp-obj word nil)))
             (let ((*print-lines* 1)
                   (*print-pretty* t))
-              (format t "~x: ~v,'0x~:[~; = ~S~]~%"
+              (format t "~x: ~v,'0x~:[~; = ~@?~]~%"
                       (+ addr (ash i word-shift))
                       (* 2 n-word-bytes)
-                      word ok lispobj))))))))
+                      word ok (or fmt "~S") lispobj))))))))
 #+sb-thread
 (defun show-tls-map ()
   (let ((list
-         (sort (sb-vm::list-allocated-objects
+         (sort (list-allocated-objects
                 :all
-                :type sb-vm:symbol-widetag
+                :type symbol-widetag
                 :test (lambda (x) (plusp (sb-kernel:symbol-tls-index x))))
                #'<
                :key #'sb-kernel:symbol-tls-index))
         (prev 0))
     (dolist (x list)
-      (let ((n  (ash (sb-kernel:symbol-tls-index x) (- sb-vm:word-shift))))
-        (when (and (> n sb-vm::primitive-thread-object-length)
+      (let ((n  (ash (sb-kernel:symbol-tls-index x) (- word-shift))))
+        (when (and (> n primitive-thread-object-length)
                    (> n (1+ prev)))
           (format t "(unused)~%"))
         (format t "~5d = ~s~%" n x)
@@ -1459,25 +1345,25 @@ We could try a few things to mitigate this:
          (let ((*print-level* 2) (*print-lines* 1))
            (format t "~x ~s~%" (get-lisp-obj-address obj) obj))))
 (defun print-all-code ()
-  (walk-dynamic-space #'print-it #x7f #b11 #b11))
+  (walk-dynamic-space #'print-it #x7f #b01111 #b00111))
 (defun print-large-code ()
-  (walk-dynamic-space #'print-it #x7f #b10011 #b10011))
+  (walk-dynamic-space #'print-it #x7f #b11111 #b10111))
 (defun print-large-unboxed ()
-  (walk-dynamic-space #'print-it #x7f #b10011 #b10010))
+  (walk-dynamic-space #'print-it #x7f #b11111 #b10010))
 ;;; Use this only if you know what you're doing. It can fail because a page
 ;;; that needs to continue onto the next page will cause the "overrun" check
 ;;; to fail.
 (defun print-page-contents (page)
   (let* ((start
-          (+ (current-dynamic-space-start) (* sb-vm:gencgc-card-bytes page)))
+          (+ (current-dynamic-space-start) (* gencgc-page-bytes page)))
          (end
-          (+ start sb-vm:gencgc-card-bytes)))
+          (+ start gencgc-page-bytes)))
     (map-objects-in-range #'print-it (%make-lisp-obj start) (%make-lisp-obj end)))))
 
 (defun map-code-objects (fun)
   (dx-flet ((filter (obj type size)
               (declare (ignore size))
-              (when (= type sb-vm:code-header-widetag)
+              (when (= type code-header-widetag)
                 (funcall fun obj)))
             (nofilter (obj type size)
               (declare (ignore type size))
@@ -1492,7 +1378,7 @@ We could try a few things to mitigate this:
       (alien-funcall (extern-alien "close_code_region" (function void)))
       (walk-dynamic-space #'nofilter
                           #b1111111 ; all generations
-                          3 3))))
+                          #b111 #b111)))) ; type mask and constraint
 
 (export 'code-from-serialno)
 (defun code-from-serialno (serial)
@@ -1502,7 +1388,7 @@ We could try a few things to mitigate this:
     (map-code-objects #'visit)))
 
 (defun show-all-layouts ()
-  (let ((l (sb-vm::list-allocated-objects :all :test #'sb-kernel::wrapper-p))
+  (let ((l (list-allocated-objects :all :test #'sb-kernel::wrapper-p))
         zero trailing-raw trailing-tagged vanilla)
     (dolist (x l)
       (let ((m (wrapper-bitmap x)))
@@ -1537,7 +1423,7 @@ We could try a few things to mitigate this:
 
 #+ubsan
 (defun find-poisoned-vectors (&aux result)
-  (dolist (v (sb-vm:list-allocated-objects :all :type sb-vm:simple-vector-widetag)
+  (dolist (v (list-allocated-objects :all :type simple-vector-widetag)
              result)
     (when (dotimes (i (length v))
             (declare (optimize (sb-c::aref-trapping 0)))
@@ -1552,17 +1438,22 @@ We could try a few things to mitigate this:
 
 (in-package "SB-C")
 ;;; As soon as practical in warm build it makes sense to add
-;;; cold-allocation-point-fixups into the weak hash-table.
+;;; cold-allocation-patch-points into the weak hash-table.
 ;;; FIXME: I suspect that this wants to be just a weak vector
 ;;; (all code objects that have any allocation profiling compiled in),
 ;;; and not a hash-table, and that the list of fixups in the component
 ;;; can be attached to the debug info (in the manner of debug funs).
 ;;; When this was first implemented, weak-vectors weren't a thing. Maybe?
-(defvar *!cold-allocation-point-fixups*)
-(let ((hash-table (make-hash-table :test 'eq)))
-  ;; Group by code component
-  (loop for (code . offset) across *!cold-allocation-point-fixups*
-        do (push offset (gethash code hash-table)))
-  (dohash ((code list) hash-table)
-    (setf (gethash code *allocation-point-fixups*)
-          (convert-alloc-point-fixups code list))))
+(defvar *!cold-allocation-patch-point*)
+(loop for (code . points) in *!cold-allocation-patch-point*
+      do (setf (gethash code *allocation-patch-points*) points))
+
+(defun gctablesize (heap-size-gb page-size cards-per-page)
+  (let* ((card-size (/ page-size cards-per-page))
+         (heap-size (* heap-size-gb (expt 1024 3)))
+         (npages (/ heap-size page-size))
+         (ncards (/ heap-size card-size))
+         (pte-nbytes (* 8 npages)))
+    (format t " PTE bytes: ~8D~%" pte-nbytes)
+    (format t "card bytes: ~8D~%" ncards)
+    (format t "     total: ~8D~%" (+ pte-nbytes ncards))))
