@@ -15,19 +15,11 @@
 ;;;; provided with absolutely no warranty. See the COPYING and CREDITS
 ;;;; files for more information.
 
-(defpackage "SB-LOCKLESS"
-  (:use "CL" "SB-EXT" "SB-INT" "SB-SYS" "SB-KERNEL")
-  (:shadow "ENDP"))
-
 (in-package "SB-LOCKLESS")
-(setf (system-package-p *package*) t)
 
 ;;; The changes to GC to support this code were as follows:
-;;; * One bit of the payload length in an instance header is reserved to signify
-;;;   that the instance has a special GC scavenge method. This avoids indirecting
-;;;   to the layout to see whether all instances of a type have a special method.
 ;;;
-;;; * If an instance has a header bit so indicating, then the first data slot
+;;; * If an instance is flagged as being a LFlist node, then the first data slot
 ;;;   is treated as an instance pointer even if it missing its tag bits.
 ;;;
 ;;; * Since you can't pin an object that you don't a-priori have a tagged pointer
@@ -45,15 +37,6 @@
 ;;; DO-REFERENCED-OBJECT can follow untagged pointers.
 ;;; This is potentially more of an annoyance than it is a bug.
 
-(eval-when (:compile-toplevel :load-toplevel :execute)
-  (import 'sb-kernel::list-node))
-
-(defstruct (list-node
-            (:conc-name nil)
-            (:constructor %make-sentinel-node ())
-            (:copier nil))
-  (%node-next nil))
-
 (defstruct (keyed-node
             (:conc-name nil)
             (:include list-node)
@@ -61,13 +44,6 @@
   (node-key 0 :read-only t)
   (node-data))
 
-(macrolet ((flags (x)
-             #+64-bit `(sb-kernel::layout-%bits ,x)
-             #-64-bit `(sb-kernel::layout-flags ,x)))
-  (let ((layout (sb-kernel:find-layout 'keyed-node)))
-    (setf (flags layout) (logior (flags layout)
-                                 sb-kernel::+custom-gc-scavenge-flag+))))
-
 (declaim (inline ptr-markedp node-markedp))
 (defun ptr-markedp (bits) (fixnump bits))
 (defun node-markedp (node) (fixnump (%node-next node)))
@@ -85,20 +61,6 @@
 
 (defmacro endp (node) `(eq ,node (load-time-value *tail-atom*)))
 
-;;; Specialized list variants will be created for
-;;;  fixnum, integer, real, string, generic "comparable"
-;;; but the node type and list type is the same regardless of key type.
-(defstruct (linked-list
-            (:constructor %make-lfl
-                          (head inserter deleter finder inequality equality))
-            (:conc-name list-))
-  (head       nil :type list-node :read-only t)
-  (inserter   nil :type function :read-only t)
-  (deleter    nil :type function :read-only t)
-  (finder     nil :type function :read-only t)
-  (inequality nil :type function :read-only t)
-  (equality   nil :type function :read-only t))
-
 (defconstant-eqx +predefined-key-types+
   #((fixnum  lfl-insert/fixnum  lfl-delete/fixnum  lfl-find/fixnum < =)
     (integer lfl-insert/integer lfl-delete/integer lfl-find/integer < =)
@@ -151,8 +113,11 @@
 
 (declaim (inline get-next))
 (defun get-next (node)
-  (with-pinned-objects (node) ; pinning a node also pins its 'next'
-    (let ((%next (%node-next node)))
+  ;; Storing NODE in *PINNED-OBJECTS* causes its successor to become pinned.
+  (#+cheneygc sb-sys:without-gcing #+gencgc progn
+   (let* ((sb-vm::*pinned-objects* (cons node sb-vm::*pinned-objects*))
+          (%next (%node-next node)))
+      (declare (truly-dynamic-extent sb-vm::*pinned-objects*))
       (values (truly-the list-node
                (%make-lisp-obj (logior (get-lisp-obj-address %next)
                                        sb-vm:instance-pointer-lowtag)))
@@ -371,6 +336,10 @@
     (do-lockfree-list (x list) (incf n))
     n))
 
+;;; This function is not really part of the API. It preserve the deletion bit,
+;;; which doesn't really make sense from an interface perspective.
+;;; However, when devising tests of the algorithms, it is useful to capture
+;;; a complete snapshot of the list.
 (defun copy-lfl (lfl)
   (labels ((copy-chain (node)
              (if (eq node *tail-atom*)
@@ -384,6 +353,6 @@
                                copy-of-next)))
                    copy))))
     (let ((new (copy-structure lfl)))
-      (setf (%instance-ref new (get-dsd-index linked-list head))
+      (%instance-set new (get-dsd-index linked-list head)
             (copy-chain (list-head new)))
       new)))
