@@ -190,10 +190,9 @@
                      (let* ((data (sap-ref-8 pc 1)) ; encodes dst register and size
                             (value (sb-vm:context-register context (ash data -2)))
                             (nbytes (ash 1 (logand data #b11)))
-                            ;; EMIT-SAP-REF always loads the EA into TEMP-REG-TN
-                            ;; which points to the shadow, not the user memory now.
-                            (ea (logxor (sb-vm:context-register
-                                         context (tn-offset sb-vm::temp-reg-tn))
+                            ;; EMIT-SAP-REF wires the EA to a predetermined register,
+                            ;; which now points to the shadow space, not the user memory.
+                            (ea (logxor (sb-vm:context-register context msan-temp-reg-number)
                                         msan-mem-to-shadow-xor-const)))
                        `(:raw ,ea ,nbytes ,value)))))
           (t
@@ -202,15 +201,9 @@
 
 #+immobile-space
 (defun alloc-immobile-fdefn ()
-  (or #+nil ; Avoid creating new objects in the text segment for now
-      (and (= (alien-funcall (extern-alien "lisp_code_in_elf" (function int))) 1)
-           (alloc-immobile-code (* fdefn-size n-word-bytes)
-                                  (logior (ash undefined-fdefn-header 16)
-                                          fdefn-widetag) ; word 0
-                                  0 other-pointer-lowtag nil)) ; word 1, lowtag, errorp
-      (alloc-immobile-fixedobj fdefn-size
-                               (logior (ash undefined-fdefn-header 16)
-                                       fdefn-widetag)))) ; word 0
+  (alloc-immobile-fixedobj fdefn-size
+                           (logior (ash undefined-fdefn-header 16)
+                                   fdefn-widetag))) ; word 0
 
 #+immobile-code
 (progn
@@ -303,7 +296,7 @@
               (truly-the word (+ addr insts-offs))
               (sap-ref-word sap insts-offs) #xFFFFFFE9058B48  ; MOV RAX,[RIP-23]
               (sap-ref-32 sap (+ insts-offs 7)) #x00FD60FF))) ; JMP [RAX-3]
-    (%set-funcallable-instance-info gf 0 slot-vector)
+    (setf (%funcallable-instance-info gf 0) slot-vector)
     gf))
 
 (defun fdefn-has-static-callers (fdefn)
@@ -393,11 +386,18 @@
       (return (loop (cond ((>= (incf i) len) (return t))
                           ((eq thing (svref things i)) (return nil))))))))
 
-;;; Allocate a code object.
-(defun alloc-dynamic-space-code (total-words)
-  (values (%primitive alloc-dynamic-space-code (the fixnum total-words))))
+(define-load-time-global *static-linker-lock* (sb-thread:make-mutex :name "static linker"))
 
-;;; Remove calls via fdefns from CODE when compiling into memory.
+(define-load-time-global *never-statically-link* '(find-package))
+;;; Remove calls via fdefns from CODE. This is called after compiling
+;;; to memory, or when saving a core.
+;;; Do not replace globally notinline functions, because notinline has
+;;; an extra connotation of ensuring that replacement of the function
+;;; under that name always works. It usually works to replace a statically
+;;; linked function, but with a caveat: un-statically-linking requires calling
+;;; MAP-OBJECTS-IN-RANGE, which is unreliable in the presence of
+;;; multiple threads. Unfortunately, some users dangerously redefine
+;;; builtin functions, and moreover, while there are multiple threads.
 (defun statically-link-code-obj (code fixups)
   (declare (ignorable code fixups))
   (unless (immobile-space-obj-p code)
@@ -420,7 +420,9 @@
       (let* ((fdefn (code-header-ref code (+ fdefns-start i)))
              (fun (when (fdefn-p fdefn) (fdefn-fun fdefn))))
         (when (and (immobile-space-obj-p fun)
-                   (not (fun-requires-simplifying-trampoline-p fun)))
+                   (not (fun-requires-simplifying-trampoline-p fun))
+                   (not (member (fdefn-name fdefn) *never-statically-link* :test 'equal))
+                   (neq (info :function :inlinep (fdefn-name fdefn)) 'notinline))
           (setf any-replacements t (aref replacements i) fun))))
     (dotimes (i fdefns-count)
       (when (and (aref replacements i)
@@ -453,7 +455,7 @@
             (setf (aref replacements fdefn-index) nil))))
       (let ((stored-locs (if any-ambiguous
                              (make-array fdefns-count :initial-element nil))))
-        (with-system-mutex (sb-c::*static-linker-lock*)
+        (with-system-mutex (*static-linker-lock*)
           (dolist (fixup fixups)
             (binding* ((fdefn-index (car fixup) :exit-if-null)
                        (offset (cdr fixup))
@@ -478,3 +480,13 @@
                 (setf (code-header-ref code wordindex)
                       (cons (code-header-ref code wordindex) locs)))))))))
   code)
+
+(sb-c::when-vop-existsp (:translate sb-c::unsigned+)
+  (defconstant cf-bit 0)
+  (defconstant sf-bit 7)
+  (defconstant of-bit 11)
+
+  (defun context-overflow-carry-flags (context)
+    (let ((flags (context-flags context)))
+      (values (logbitp of-bit flags)
+              (logbitp cf-bit flags)))))
