@@ -48,8 +48,7 @@
   (:save-p :compute-only)
   (:temporary (:scs (descriptor-reg) :from (:argument 0)) obj-temp))
 
-;;; With Symbol-Value, we check that the value isn't the trap object.  So
-;;; Symbol-Value of NIL is NIL.
+;;; With Symbol-Value, we check that the value isn't the trap object.
 ;;;
 (define-vop (symbol-value checked-cell-ref)
   (:translate symeval)
@@ -61,14 +60,12 @@
       (inst b :eq err-lab))))
 
 ;;; Like CHECKED-CELL-REF, only we are a predicate to see if the cell is bound.
-(define-vop (boundp-frob)
+(define-vop (boundp)
   (:args (object :scs (descriptor-reg)))
   (:conditional)
   (:info target not-p)
   (:policy :fast-safe)
-  (:temporary (:scs (descriptor-reg)) value))
-
-(define-vop (boundp boundp-frob)
+  (:temporary (:scs (descriptor-reg)) value)
   (:translate boundp)
   (:generator 9
     (loadw value object symbol-value-slot other-pointer-lowtag)
@@ -146,10 +143,9 @@
 (define-vop (fdefn-makunbound)
   (:policy :fast-safe)
   (:translate fdefn-makunbound)
-  (:args (fdefn :scs (descriptor-reg) :target result))
+  (:args (fdefn :scs (descriptor-reg)))
   (:temporary (:scs (non-descriptor-reg)) temp)
   (:temporary (:scs (interior-reg)) lip)
-  (:results (result :scs (descriptor-reg)))
   (:generator 38
     (let ((undefined-tramp-fixup (gen-label)))
       (assemble (:elsewhere)
@@ -157,9 +153,7 @@
         (inst word (make-fixup 'undefined-tramp :assembly-routine)))
       (storew null-tn fdefn fdefn-fun-slot other-pointer-lowtag)
       (inst load-from-label temp lip undefined-tramp-fixup)
-      (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag)
-      (move result fdefn))))
-
+      (storew temp fdefn fdefn-raw-addr-slot other-pointer-lowtag))))
 
 
 ;;;; Binding and Unbinding.
@@ -230,9 +224,9 @@
   closure-info-offset fun-pointer-lowtag
   (descriptor-reg any-reg) * %closure-index-ref)
 
-(define-full-setter set-funcallable-instance-info *
-  funcallable-instance-info-offset fun-pointer-lowtag
-  (descriptor-reg any-reg null) * %set-funcallable-instance-info)
+(define-full-setter %closure-index-set *
+  closure-info-offset fun-pointer-lowtag
+  (descriptor-reg any-reg null) * %closure-index-set)
 
 (define-full-reffer funcallable-instance-info *
   funcallable-instance-info-offset fun-pointer-lowtag
@@ -289,8 +283,49 @@
 (define-full-reffer code-header-ref * 0 other-pointer-lowtag
   (descriptor-reg any-reg) * code-header-ref)
 
-(define-full-setter code-header-set * 0 other-pointer-lowtag
-  (descriptor-reg any-reg null) * code-header-set)
+(define-vop (code-header-set)
+  (:translate code-header-set)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg))
+         (value :scs (any-reg descriptor-reg)))
+  (:arg-types * tagged-num *)
+  (:temporary (:scs (non-descriptor-reg)) temp #+gencgc card)
+  #+gencgc (:temporary (:scs (interior-reg)) lip)
+  #+gencgc (:temporary (:sc non-descriptor-reg) pa-flag)
+  (:generator 10
+    #+cheneygc
+    (progn  (inst sub temp index other-pointer-lowtag)
+            (inst str value (@ object temp)))
+    #+gencgc
+    (let ((mask-fixup-label (gen-label))
+          (table-fixup-label (gen-label)))
+      (inst load-from-label temp lip mask-fixup-label)
+      (inst ldr temp (@ temp))
+      (inst ldr temp (@ temp))
+      (pseudo-atomic (pa-flag)
+        ;; Compute card mark index
+        (inst mov card (lsr object gencgc-card-shift))
+        (inst and card card temp)
+        ;; Load mark table base
+        (inst load-from-label temp lip table-fixup-label)
+        (inst ldr temp (@ temp))
+        (inst ldr temp (@ temp))
+        ;; Touch the card mark byte.
+        (inst strb null-tn (@ temp card))
+        ;; set 'written' flag in the code header
+        ;; If two threads get here at the same time, they'll write the same byte.
+        (let ((byte (- #+little-endian 3 other-pointer-lowtag)))
+          (inst ldrb temp (@ object byte))
+          (inst orr temp temp #x40)
+          (inst strb temp (@ object byte)))
+        (inst sub temp index other-pointer-lowtag)
+        (inst str value (@ object temp)))
+      (assemble (:elsewhere)
+        (emit-label mask-fixup-label)
+        (inst word (make-fixup "gc_card_table_mask" :foreign-dataref))
+        (emit-label table-fixup-label)
+        (inst word (make-fixup "gc_card_mark" :foreign-dataref))))))
 
 ;;;; raw instance slot accessors
 
