@@ -128,8 +128,6 @@
 ;;;
 (defconstant-eqx +static-symbols+
  `#(,@+common-static-symbols+
-    *allocation-pointer*
-
      *control-stack-pointer*
      *binding-stack-pointer*
      *interrupted-control-stack-pointer*
@@ -155,5 +153,3 @@
 ;;; The number of bits per element in the assemblers code vector.
 ;;;
 (defparameter *assembly-unit-length* 8)
-
-(defvar *allocation-pointer*)
