@@ -21,8 +21,8 @@
          (rank :scs (any-reg)))
   (:arg-types tagged-num tagged-num)
   (:temporary (:scs (descriptor-reg) :to (:result 0) :target result) header)
-  (:temporary (:sc non-descriptor-reg) pa-flag ndescr)
-  (:temporary (:scs (interior-reg)) lip)
+  (:temporary (:sc non-descriptor-reg) ndescr)
+  (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:results (result :scs (descriptor-reg)))
   (:generator 5
     ;; Compute the allocation size.
@@ -30,40 +30,18 @@
     (inst add ndescr ndescr (+ (* array-dimensions-offset n-word-bytes)
                                lowtag-mask))
     (inst and ndescr ndescr (bic-mask lowtag-mask))
-    (pseudo-atomic (pa-flag)
-      (allocation nil ndescr other-pointer-lowtag header :flag-tn pa-flag :lip lip)
+    (pseudo-atomic (lr)
+      (allocation nil ndescr other-pointer-lowtag header :flag-tn lr)
       ;; Now that we have the space allocated, compute the header
       ;; value.
-      (inst lsl ndescr rank (- n-widetag-bits n-fixnum-tag-bits))
-      (inst add ndescr ndescr (ash (1- array-dimensions-offset) n-widetag-bits))
-      (inst orr ndescr ndescr (lsr type n-fixnum-tag-bits))
+      ;; See ENCODE-ARRAY-RANK.
+      (inst sub ndescr rank (fixnumize 1))
+      (inst and ndescr ndescr (fixnumize array-rank-mask))
+      (inst orr ndescr type (lsl ndescr array-rank-position))
+      (inst lsr ndescr ndescr n-fixnum-tag-bits)
       ;; And store the header value.
       (storew ndescr header 0 other-pointer-lowtag))
     (move result header)))
-
-(define-vop (make-array-header/c)
-  (:translate make-array-header)
-  (:policy :fast-safe)
-  (:arg-types (:constant t) (:constant t))
-  (:info type rank)
-  (:temporary (:scs (descriptor-reg) :to (:result 0) :target result) header)
-  (:temporary (:sc non-descriptor-reg) pa-flag)
-  (:temporary (:scs (interior-reg)) lip)
-  (:results (result :scs (descriptor-reg)))
-  (:generator 4
-    (let* ((header-size (+ rank
-                           (1- array-dimensions-offset)))
-           (bytes (logandc2 (+ (* (1+ header-size) n-word-bytes)
-                               lowtag-mask)
-                            lowtag-mask))
-           (header-bits (logior (ash header-size
-                                     n-widetag-bits)
-                                type)))
-      (pseudo-atomic (pa-flag)
-        (allocation nil bytes other-pointer-lowtag header :flag-tn pa-flag :lip lip)
-        (load-immediate-word pa-flag header-bits)
-        (storew pa-flag header 0 other-pointer-lowtag)))
-    (move result header)))
 
 ;;;; Additional accessors and setters for the array header.
 (define-full-reffer %array-dimension *
@@ -74,17 +52,50 @@
   array-dimensions-offset other-pointer-lowtag
   (any-reg) positive-fixnum sb-kernel:%set-array-dimension)
 
-(define-vop (array-rank-vop)
-  (:translate sb-kernel:%array-rank)
+(define-vop ()
+  (:translate %array-rank)
   (:policy :fast-safe)
   (:args (x :scs (descriptor-reg)))
-  (:temporary (:scs (non-descriptor-reg)) temp)
-  (:results (res :scs (any-reg descriptor-reg)))
+  (:results (res :scs (unsigned-reg)))
+  (:result-types positive-fixnum)
   (:generator 6
-    (loadw temp x 0 other-pointer-lowtag)
-    (inst asr temp temp n-widetag-bits)
-    (inst sub temp temp (1- array-dimensions-offset))
-    (inst lsl res temp n-fixnum-tag-bits)))
+    (inst ldrsb res (@ x (- (/ array-rank-position n-byte-bits)
+                          other-pointer-lowtag)))
+    (inst add res res 1)))
+
+(define-vop ()
+  (:translate %array-rank=)
+  (:policy :fast-safe)
+  (:args (array :scs (descriptor-reg)))
+  (:temporary (:scs (unsigned-reg)) x)
+  (:info rank)
+  (:arg-types * (:constant t))
+  (:conditional :eq)
+  (:generator 2
+    (inst ldrb x (@ array (- (/ array-rank-position n-byte-bits)
+                            other-pointer-lowtag)))
+    (inst cmp x (add-sub-immediate (encode-array-rank rank)))))
+
+(define-vop (simple-array-header-of-rank-p type-predicate)
+  (:translate sb-c::simple-array-header-of-rank-p)
+  (:info target not-p rank)
+  (:arg-types * (:constant t))
+  (:generator 2
+    (unless (other-pointer-tn-ref-p args)
+      (%test-lowtag value temp (if not-p
+                                   target
+                                   drop-through)
+                    t other-pointer-lowtag))
+    (inst ldrh temp (@ value (- other-pointer-lowtag)))
+    (inst cmp temp (add-sub-immediate
+                     (dpb (encode-array-rank rank)
+                       (byte 8 array-rank-position)
+                       simple-array-widetag)))
+    (inst b (if not-p
+                :ne
+                :eq)
+      target)
+    drop-through))
 
 ;;;; Bounds checking routine.
 (define-vop (check-bound)
@@ -140,6 +151,8 @@
              (cond ((integerp bound)
                     (inst cmp index bound)
                     (inst b :hs error))
+                   ((eql index 0)
+                    (inst cbz bound error))
                    (t
                     (inst cmp bound index)
                     (inst b :ls error))))))))
@@ -248,22 +261,23 @@
          (:translate data-vector-ref)
          (:policy :fast-safe)
          (:args (object :scs (descriptor-reg))
-                (index :scs (unsigned-reg)))
+                (index :scs (unsigned-reg) :to :save))
          (:arg-types ,type positive-fixnum)
-         (:results (value :scs (any-reg)))
+         (:results (result :scs (unsigned-reg)))
          (:result-types positive-fixnum)
-         (:temporary (:scs (interior-reg)) lip)
-         (:temporary (:scs (non-descriptor-reg) :to (:result 0)) temp result)
+         (:temporary (:scs (non-descriptor-reg)) temp)
          (:generator 20
            ;; Compute the offset for the word we're interested in.
            (inst lsr temp index ,bit-shift)
            ;; Load the word in question.
-           (inst add lip object (lsl temp word-shift))
-           (inst ldr result (@ lip
+           (inst add temp object (lsl temp word-shift))
+           (inst ldr result (@ temp
                                (- (* vector-data-offset n-word-bytes)
                                   other-pointer-lowtag)))
            ;; Compute the position of the bitfield we need.
-           (inst and temp index ,(1- elements-per-word))
+           ,(if (= bits 1)
+                `(setf temp index)
+                `(inst and temp index ,(1- elements-per-word)))
            ,@(when (eq *backend-byte-order* :big-endian)
                `((inst eor temp temp ,(1- elements-per-word))))
            ,@(unless (= bits 1)
@@ -271,9 +285,7 @@
            ;; Shift the field we need to the low bits of RESULT.
            (inst lsr result result temp)
            ;; Mask out the field we're interested in.
-           (inst and result result ,(1- (ash 1 bits)))
-           ;; And fixnum-tag the result.
-           (inst lsl value result n-fixnum-tag-bits)))
+           (inst and result result ,(1- (ash 1 bits)))))
        (define-vop (,(symbolicate "DATA-VECTOR-REF/" type "-C"))
          (:note "inline array access")
          (:translate data-vector-ref)
@@ -281,9 +293,8 @@
          (:args (object :scs (descriptor-reg)))
          (:info index)
          (:arg-types ,type (:constant index))
-         (:results (value :scs (any-reg)))
+         (:results (result :scs (unsigned-reg)))
          (:result-types positive-fixnum)
-         (:temporary (:scs (non-descriptor-reg) :to (:result 0)) result)
          (:generator 15
            (multiple-value-bind (index bit) (floor index ,elements-per-word)
              (inst ldr result (@ object
@@ -291,63 +302,120 @@
                                   (+ (* index n-word-bytes)
                                      (- (* vector-data-offset n-word-bytes)
                                         other-pointer-lowtag)))))
-             (inst ubfm result result (* bit ,bits) (+ (* bit ,bits) (1- ,bits)))
-             (inst lsl value result n-fixnum-tag-bits))))
+             (inst ubfm result result (* bit ,bits) (+ (* bit ,bits) (1- ,bits))))))
+       (define-vop (,(symbolicate "DATA-VECTOR-SET/" type "-C"))
+         (:note "inline array store")
+         (:translate data-vector-set)
+         (:policy :fast-safe)
+         (:args (object :scs (descriptor-reg))
+                (value :scs (unsigned-reg)
+                       :load-if (not (and (sc-is value immediate)
+                                          (memq (tn-value value)
+                                                '(0 ,(1- (ash 1 bits))))))))
+         (:info index)
+         (:arg-types ,type (:constant index) positive-fixnum)
+         (:temporary (:scs (non-descriptor-reg)) old)
+         (:generator 20
+            (multiple-value-bind (index bit) (floor index ,elements-per-word)
+              (inst ldr old (@ object
+                               (load-store-offset
+                                (+ (* index n-word-bytes)
+                                   (- (* vector-data-offset n-word-bytes)
+                                      other-pointer-lowtag)))))
+              (cond ((not (sc-is value immediate))
+                     (inst bfm old value (mod (- n-word-bits (* bit ,bits)) n-word-bits) (1- ,bits)))
+                    ((zerop (tn-value value))
+                     (inst and old old (lognot (ash (1- (ash 1 ,bits)) (* bit ,bits)))))
+                    (t
+                     (inst orr old old (ash (1- (ash 1 ,bits)) (* bit ,bits)))))
+              (inst str old (@ object
+                               (load-store-offset
+                                (+ (* index n-word-bytes)
+                                   (- (* vector-data-offset n-word-bytes)
+                                      other-pointer-lowtag))))))))
        (define-vop (,(symbolicate "DATA-VECTOR-SET/" type))
          (:note "inline array store")
          (:translate data-vector-set)
          (:policy :fast-safe)
          (:args (object :scs (descriptor-reg))
-                (index :scs (unsigned-reg) :target shift)
-                (value :scs (unsigned-reg immediate) :target result))
+                (index :scs (unsigned-reg)
+                       ,@(unless (= bits 1)
+                           '(:target shift)))
+                (value :scs (unsigned-reg immediate)))
          (:arg-types ,type positive-fixnum positive-fixnum)
-         (:results (result :scs (unsigned-reg)))
-         (:result-types positive-fixnum)
-         (:temporary (:scs (interior-reg)) lip)
+         (:temporary (:scs (non-descriptor-reg)) lip)
          (:temporary (:scs (non-descriptor-reg)) temp old)
-         (:temporary (:scs (non-descriptor-reg) :from (:argument 1)) shift)
+         ,@(unless (= bits 1)
+             '((:temporary (:scs (non-descriptor-reg) :from (:argument 1)) shift)))
          (:generator 25
-           ;; Compute the offset for the word we're interested in.
-           (inst lsr temp index ,bit-shift)
-           ;; Load the word in question.
-           (inst add lip object (lsl temp word-shift))
-           (inst ldr old (@ lip
-                            (- (* vector-data-offset n-word-bytes)
-                               other-pointer-lowtag)))
-           ;; Compute the position of the bitfield we need.
-           (inst and shift index ,(1- elements-per-word))
-           ,@(when (eq *backend-byte-order* :big-endian)
-               `((inst eor shift ,(1- elements-per-word))))
-           ,@(unless (= bits 1)
-               `((inst lsl shift shift ,(1- (integer-length bits)))))
-           ;; Clear the target bitfield.
-           (unless (and (sc-is value immediate)
-                        (= (tn-value value) ,(1- (ash 1 bits))))
-             (inst mov temp ,(1- (ash 1 bits)))
-             (inst lsl temp temp shift)
-             (inst bic old old temp))
-           ;; LOGIOR in the new value (shifted appropriatly).
-           (sc-case value
-             (immediate
-              (inst mov temp (logand (tn-value value) ,(1- (ash 1 bits)))))
-             (unsigned-reg
-              (inst and temp value ,(1- (ash 1 bits)))))
-           (inst lsl temp temp shift)
-           (inst orr old old temp)
-           ;; Write the altered word back to the array.
-           (inst str old (@ lip
-                            (- (* vector-data-offset n-word-bytes)
-                               other-pointer-lowtag)))
-           ;; And present the result properly.
-           (sc-case value
-             (immediate
-              (inst mov result (tn-value value)))
-             (unsigned-reg
-              (move result value)))))))))
+           (let ((shift ,(if (= bits 1)
+                             'index
+                             'shift)))
+             ;; Compute the offset for the word we're interested in.
+             (inst lsr temp index ,bit-shift)
+             ;; Load the word in question.
+             (inst add lip object (lsl temp word-shift))
+             (inst ldr old (@ lip
+                              (- (* vector-data-offset n-word-bytes)
+                                 other-pointer-lowtag)))
+             ;; Compute the position of the bitfield we need.
+             ,@(unless (= bits 1)
+                 `((inst and shift index ,(1- elements-per-word))))
+             ,@(when (eq *backend-byte-order* :big-endian)
+                 `((inst eor shift ,(1- elements-per-word))))
+             ,@(unless (= bits 1)
+                 `((inst lsl shift shift ,(1- (integer-length bits)))))
+             ;; Clear the target bitfield.
+             (unless (and (sc-is value immediate)
+                          (= (tn-value value) ,(1- (ash 1 bits))))
+               (inst mov temp ,(1- (ash 1 bits)))
+               (inst lsl temp temp shift)
+               (inst bic old old temp))
+             (unless (and (sc-is value immediate)
+                          (= (tn-value value) 0))
+               ;; LOGIOR in the new value (shifted appropriately).
+               (sc-case value
+                 (immediate
+                  (inst mov temp (logand (tn-value value) ,(1- (ash 1 bits))))
+                  (inst lsl temp temp shift))
+                 (unsigned-reg
+                  (inst lsl temp value shift)))
+               (inst orr old old temp))
+             ;; Write the altered word back to the array.
+             (inst str old (@ lip
+                              (- (* vector-data-offset n-word-bytes)
+                                 other-pointer-lowtag))))))))))
   (def-small-data-vector-frobs simple-bit-vector 1)
   (def-small-data-vector-frobs simple-array-unsigned-byte-2 2)
   (def-small-data-vector-frobs simple-array-unsigned-byte-4 4))
 
+(define-vop (data-vector-ref/simple-bit-vector-eq)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg)) (index :scs (unsigned-reg)))
+  (:arg-types simple-bit-vector positive-fixnum)
+  (:conditional :eq)
+  (:temporary (:scs (non-descriptor-reg)) temp x)
+  (:generator 20
+    (inst lsr temp index 6)
+    (inst add temp object (lsl temp word-shift))
+    (inst ldr x (@ temp (- (* vector-data-offset n-word-bytes) other-pointer-lowtag)))
+    (inst lsr x x index)
+    (inst tst x 1)))
+
+(define-vop (data-vector-ref/simple-bit-vector-c-eq)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg)))
+  (:info index)
+  (:arg-types simple-bit-vector (:constant index))
+  (:conditional :eq)
+  (:temporary (:scs (non-descriptor-reg)) x)
+  (:generator 15
+    (multiple-value-bind (index bit)
+        (floor index 64)
+      (inst ldr x (@ object (load-store-offset (+ (* index n-word-bytes)
+                                                  (- (* vector-data-offset n-word-bytes) other-pointer-lowtag)))))
+      (inst tst x (ash 1 bit)))))
+
 ;;; And the float variants.
 (define-vop (data-vector-ref/simple-array-single-float)
   (:note "inline array access")
@@ -371,18 +439,14 @@
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (single-reg) :target result))
+         (value :scs (single-reg)))
   (:arg-types simple-array-single-float positive-fixnum single-float)
-  (:results (result :scs (single-reg)))
-  (:result-types single-float)
   (:temporary (:scs (non-descriptor-reg) :from (:argument 1)) offset)
   (:generator 5
     (inst lsl offset index (1- (- word-shift n-fixnum-tag-bits)))
     (inst add offset offset (- (* vector-data-offset n-word-bytes)
                                other-pointer-lowtag))
-    (inst str value (@ object offset))
-    (unless (location= result value)
-      (inst fmov result value))))
+    (inst str value (@ object offset))))
 
 (define-vop (data-vector-ref/simple-array-double-float)
   (:note "inline array access")
@@ -406,18 +470,14 @@
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg))
          (index :scs (any-reg))
-         (value :scs (double-reg) :target result))
+         (value :scs (double-reg)))
   (:arg-types simple-array-double-float positive-fixnum double-float)
-  (:results (result :scs (double-reg)))
-  (:result-types double-float)
   (:temporary (:scs (non-descriptor-reg) :from (:argument 1)) offset)
   (:generator 20
     (inst lsl offset index (- word-shift n-fixnum-tag-bits))
     (inst add offset offset (- (* vector-data-offset n-word-bytes)
                                other-pointer-lowtag))
-    (inst str value (@ object offset))
-    (unless (location= result value)
-      (inst fmov result value))))
+    (inst str value (@ object offset))))
 
 ;;; Complex float arrays.
 
@@ -443,19 +503,15 @@
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg) :to :result)
          (index :scs (any-reg))
-         (value :scs (complex-single-reg) :target result))
+         (value :scs (complex-single-reg)))
   (:arg-types simple-array-complex-single-float positive-fixnum
               complex-single-float)
-  (:results (result :scs (complex-single-reg)))
-  (:result-types complex-single-float)
   (:temporary (:scs (non-descriptor-reg) :from (:argument 1)) offset)
   (:generator 5
     (inst lsl offset index (- word-shift n-fixnum-tag-bits))
     (inst add offset offset (- (* vector-data-offset n-word-bytes)
                                other-pointer-lowtag))
-    (inst str value (@ object offset))
-    (unless (location= result value)
-      (inst fmov result value))))
+    (inst str value (@ object offset))))
 
 (define-vop (data-vector-ref/simple-array-complex-double-float)
   (:note "inline array access")
@@ -479,19 +535,15 @@
   (:policy :fast-safe)
   (:args (object :scs (descriptor-reg) :to :result)
          (index :scs (any-reg))
-         (value :scs (complex-double-reg) :target result))
+         (value :scs (complex-double-reg)))
   (:arg-types simple-array-complex-double-float positive-fixnum
               complex-double-float)
-  (:results (result :scs (complex-double-reg)))
-  (:result-types complex-double-float)
   (:temporary (:scs (non-descriptor-reg) :from (:argument 1)) offset)
   (:generator 5
     (inst lsl offset index (1+ (- word-shift n-fixnum-tag-bits)))
     (inst add offset offset (- (* vector-data-offset n-word-bytes)
                                other-pointer-lowtag))
-    (inst str value (@ object offset))
-    (unless (location= result value)
-      (inst s-mov result value))))
+    (inst str value (@ object offset))))
 
 ;;; These vops are useful for accessing the bits of a vector irrespective of
 ;;; what type of vector it is.
@@ -518,7 +570,7 @@
   (:result-types unsigned-num)
   (:temporary (:sc unsigned-reg) offset)
   (:temporary (:sc non-descriptor-reg) sum)
-  (:temporary (:sc interior-reg) lip)
+  (:temporary (:sc non-descriptor-reg) lip)
   (:generator 4
     (inst lsl offset index (- word-shift n-fixnum-tag-bits))
     (inst add offset offset (- (* vector-data-offset n-word-bytes)
@@ -532,3 +584,22 @@
     (inst stlxr tmp-tn sum lip)
     (inst cbnz tmp-tn LOOP)
     (inst dmb)))
+
+(define-vop (array-atomic-incf/word-v8.1)
+  (:translate %array-atomic-incf/word)
+  (:policy :fast-safe)
+  (:args (object :scs (descriptor-reg))
+         (index :scs (any-reg) :target offset)
+         (diff :scs (unsigned-reg)))
+  (:arg-types * positive-fixnum unsigned-num)
+  (:results (result :scs (unsigned-reg) :from :load))
+  (:result-types unsigned-num)
+  (:temporary (:sc unsigned-reg) offset)
+  (:temporary (:sc non-descriptor-reg) lip)
+  (:guard (member :arm-v8.1 *backend-subfeatures*))
+  (:generator 3
+    (inst lsl offset index (- word-shift n-fixnum-tag-bits))
+    (inst add offset offset (- (* vector-data-offset n-word-bytes)
+                               other-pointer-lowtag))
+    (inst add lip object offset)
+    (inst ldaddal diff result lip)))
