@@ -66,7 +66,8 @@
     ;; Like closure, but these can also have a layout pointer in the high header bytes.
     (funcallable-instance "funinstance" "lose" "short_boxed")
     ;; These have a scav and trans function, but no size function.
-    #-(or x86 x86-64 arm64) (return-pc "return_pc_header" "return_pc_header" "lose")
+    #-(or x86 x86-64 arm64 riscv)
+    (return-pc "return_pc_header" "return_pc_header" "lose")
 
     (value-cell "boxed")
     (symbol "symbol"
@@ -76,7 +77,7 @@
     (character "immediate")
     (sap "unboxed")
     (unbound-marker "immediate")
-    (weak-pointer "weak_pointer" "tiny_mixed" "boxed")
+    (weak-pointer "weakptr")
     (instance "instance" "lose" "instance")
     (fdefn "fdefn")
 
