@@ -97,11 +97,11 @@
 (defknown %sp-string-compare
   (simple-string index (or null index) simple-string index (or null index))
   (values index fixnum)
-  (foldable flushable))
+  (foldable flushable no-verify-arg-count))
 (defknown %sp-string=
   (simple-string index (or null index) simple-string index (or null index))
   boolean
-  (foldable flushable))
+  (foldable flushable no-verify-arg-count))
 
 (defknown sb-impl::number-sxhash (number) hash-code (foldable flushable))
 (defknown %sxhash-string (string) hash-code (foldable flushable))
@@ -143,7 +143,7 @@
   :result-arg 0)
 
 (defknown (vector-fill* vector-fill/t) (t t t t) vector
-  ()
+  (no-verify-arg-count)
   :result-arg 0)
 
 ;;; Return the length of VECTOR.
@@ -202,10 +202,10 @@
   (flushable))
 (defknown sb-kernel::check-array-shape (simple-array list)
   (simple-array)
-  (flushable)
+  (flushable no-verify-arg-count)
   :result-arg 0)
 
-(defknown %make-instance (index) instance
+(defknown (%make-instance %make-instance/mixed) (index) instance
   (flushable))
 (defknown %make-structure-instance (defstruct-description list &rest t) instance
   (flushable always-translatable))
@@ -247,18 +247,20 @@
   sb-vm:signed-word ())
 (defknown %raw-instance-xchg/word (instance index sb-vm:word) sb-vm:word ())
 
-#.`(progn
-     ,@(map 'list
-            (lambda (rsd)
-              (let* ((reader (sb-kernel::raw-slot-data-reader-name rsd))
-                     (writer (sb-kernel::raw-slot-data-writer-name rsd))
-                     (type (sb-kernel::raw-slot-data-raw-type rsd)))
-                `(progn
-                   (defknown ,reader (instance index) ,type
-                     (flushable always-translatable))
-                   (defknown ,writer (instance index ,type) (values)
-                     (always-translatable)))))
-            sb-kernel::*raw-slot-data*))
+(macrolet ((define-raw-slot-defknowns ()
+             `(progn
+                ,@(map 'list
+                       (lambda (rsd)
+                         (let* ((reader (sb-kernel::raw-slot-data-reader-name rsd))
+                                (writer (sb-kernel::raw-slot-data-writer-name rsd))
+                                (type (sb-kernel::raw-slot-data-raw-type rsd)))
+                           `(progn
+                              (defknown ,reader (instance index) ,type
+                                  (flushable always-translatable))
+                              (defknown ,writer (instance index ,type) (values)
+                                  (always-translatable)))))
+                       sb-kernel::*raw-slot-data*))))
+  (define-raw-slot-defknowns))
 
 #+compare-and-swap-vops
 (defknown %raw-instance-atomic-incf/word (instance index sb-vm:word) sb-vm:word
@@ -341,14 +343,30 @@
       (flushable movable foldable))
   (defknown %make-simd-pack-ub32 ((unsigned-byte 32) (unsigned-byte 32)
                                   (unsigned-byte 32) (unsigned-byte 32))
-      (simd-pack integer)
+      (simd-pack (unsigned-byte 32))
       (flushable movable #-sb-xc-host foldable))
   (defknown %make-simd-pack-ub64 ((unsigned-byte 64) (unsigned-byte 64))
-      (simd-pack integer)
+      (simd-pack (unsigned-byte 64))
       (flushable movable #-sb-xc-host foldable))
   (defknown (%simd-pack-low %simd-pack-high) (simd-pack)
       (unsigned-byte 64)
       (flushable movable foldable))
+  (defknown %simd-pack-ub8s (simd-pack)
+      (values (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-ub16s (simd-pack)
+      (values (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16))
+      (flushable movable foldable))
   (defknown %simd-pack-ub32s (simd-pack)
       (values (unsigned-byte 32) (unsigned-byte 32)
               (unsigned-byte 32) (unsigned-byte 32))
@@ -356,6 +374,29 @@
   (defknown %simd-pack-ub64s (simd-pack)
       (values (unsigned-byte 64) (unsigned-byte 64))
       (flushable movable foldable))
+  (defknown %simd-pack-sb8s (simd-pack)
+      (values (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-sb16s (simd-pack)
+      (values (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16))
+      (flushable movable foldable))
+  (defknown %simd-pack-sb32s (simd-pack)
+      (values (signed-byte 32) (signed-byte 32)
+              (signed-byte 32) (signed-byte 32))
+      (flushable movable foldable))
+  (defknown %simd-pack-sb64s (simd-pack)
+      (values (signed-byte 64) (signed-byte 64))
+      (flushable movable foldable))
   (defknown %simd-pack-singles (simd-pack)
       (values single-float single-float single-float single-float)
       (flushable movable foldable))
@@ -380,14 +421,30 @@
       (flushable movable foldable))
   (defknown %make-simd-pack-256-ub32 ((unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32)
                                       (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32))
-      (simd-pack-256 integer)
+      (simd-pack-256 (unsigned-byte 32))
       (flushable movable foldable))
   (defknown %make-simd-pack-256-ub64 ((unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64))
-      (simd-pack-256 integer)
+      (simd-pack-256 (unsigned-byte 64))
       (flushable movable foldable))
   (defknown (%simd-pack-256-0 %simd-pack-256-1 %simd-pack-256-2 %simd-pack-256-3) (simd-pack-256)
       (unsigned-byte 64)
       (flushable movable foldable))
+  (defknown %simd-pack-256-ub8s (simd-pack-256)
+      (values (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8)
+              (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8) (unsigned-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-ub16s (simd-pack-256)
+      (values (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16)
+              (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16) (unsigned-byte 16))
+      (flushable movable foldable))
   (defknown %simd-pack-256-ub32s (simd-pack-256)
       (values (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32)
               (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32) (unsigned-byte 32))
@@ -395,6 +452,29 @@
   (defknown %simd-pack-256-ub64s (simd-pack-256)
       (values (unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64) (unsigned-byte 64))
       (flushable movable foldable))
+  (defknown %simd-pack-256-sb8s (simd-pack-256)
+      (values (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8)
+              (signed-byte 8) (signed-byte 8) (signed-byte 8) (signed-byte 8))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-sb16s (simd-pack-256)
+      (values (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16)
+              (signed-byte 16) (signed-byte 16) (signed-byte 16) (signed-byte 16))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-sb32s (simd-pack-256)
+      (values (signed-byte 32) (signed-byte 32) (signed-byte 32) (signed-byte 32)
+              (signed-byte 32) (signed-byte 32) (signed-byte 32) (signed-byte 32))
+      (flushable movable foldable))
+  (defknown %simd-pack-256-sb64s (simd-pack-256)
+      (values (signed-byte 64) (signed-byte 64) (signed-byte 64) (signed-byte 64))
+      (flushable movable foldable))
   (defknown %simd-pack-256-singles (simd-pack-256)
       (values single-float single-float single-float single-float
               single-float single-float single-float single-float)
@@ -457,7 +537,7 @@
 (defknown %allocate-bignum (bignum-length) bignum
   (flushable #-bignum-assertions always-translatable))
 
-(defknown %bignum-length (bignum) bignum-length
+(defknown %bignum-length (bignum) (and (integer 1) bignum-length)
   (foldable flushable always-translatable))
 
 ;;; Change the length of bignum to be newlen. Newlen must be the same or
@@ -634,11 +714,15 @@
 
 ;;; formerly in 'float-tran'
 
-(defknown %single-float (real) single-float (movable foldable))
-(defknown %double-float (real) double-float (movable foldable))
+(defknown %single-float (real) single-float
+  (movable foldable no-verify-arg-count))
+(defknown %double-float (real) double-float
+  (movable foldable no-verify-arg-count))
 
-(defknown bignum-to-float (bignum symbol) float (movable foldable))
-(defknown sb-kernel::float-ratio (ratio symbol) float (movable foldable))
+(defknown bignum-to-float (bignum symbol) float
+  (movable foldable no-verify-arg-count))
+(defknown sb-kernel::float-ratio (ratio symbol) float
+  (movable foldable no-verify-arg-count))
 
 (defknown make-single-float ((signed-byte 32)) single-float
   (movable flushable))
@@ -719,4 +803,4 @@
   (movable foldable flushable))
 
 (defknown (%unary-truncate %unary-round) (real) integer
-  (movable foldable flushable))
+  (movable foldable flushable no-verify-arg-count))
