@@ -12,6 +12,8 @@
 (def-alloc '%make-structure-instance 1 :structure-alloc
            sb-vm:instance-widetag sb-vm:instance-pointer-lowtag
            nil)
+(def-alloc '%make-instance/mixed 1 :var-alloc
+           sb-vm:instance-widetag sb-vm:instance-pointer-lowtag nil)
 
 (defoptimizer (%make-structure-instance stack-allocate-result)
     ((defstruct-description &rest args) node dx)
@@ -21,18 +23,18 @@
   ;; slots, or if we're on a target with a conservatively-scavenged
   ;; stack.  We have no reader conditional for stack conservation, but
   ;; it turns out that the only time stack conservation is in play is
-  ;; when we're on GENCGC (since CHENEYGC doesn't have conservation)
-  ;; and C-STACK-IS-CONTROL-STACK (otherwise, the C stack is the
+  ;; C-STACK-IS-CONTROL-STACK (otherwise, the C stack is the
   ;; number stack, and we precisely-scavenge the control stack).
-  #-(and :gencgc :c-stack-is-control-stack)
-  (every (lambda (x) (eq (dsd-raw-type x) t))
-         (dd-slots (lvar-value defstruct-description)))
-  #+(and :gencgc :c-stack-is-control-stack)
-  t)
+  (or #+c-stack-is-control-stack t
+      (not (dd-has-raw-slot-p (lvar-value defstruct-description)))))
 
 (defoptimizer (%make-instance stack-allocate-result) ((n) node dx)
   (declare (ignore n))
   (eq dx 'truly-dynamic-extent))
+#+(and gencgc c-stack-is-control-stack)
+(defoptimizer (%make-instance/mixed stack-allocate-result) ((n) node dx)
+  (declare (ignore n))
+  (eq dx 'truly-dynamic-extent))
 (defoptimizer (%make-funcallable-instance stack-allocate-result) ((n) node dx)
   (declare (ignore n))
   (eq dx 'truly-dynamic-extent))
@@ -75,7 +77,7 @@
          res)
     (move-lvar-result node block locs lvar)))
 
-(defun emit-inits (node block name object lowtag inits args)
+(defun emit-inits (node block object lowtag inits args)
   (let ((unbound-marker-tn nil)
         (funcallable-instance-tramp-tn nil)
         (dx-p (node-stack-allocate-p node)))
@@ -100,7 +102,7 @@
                          `(ecase raw-type
                             ((t)
                              (vop set-slot node block object arg-tn
-                                  name (+ sb-vm:instance-slots-offset slot) lowtag))
+                                  :allocator (+ sb-vm:instance-slots-offset slot) lowtag))
                             ,@(map 'list
                                (lambda (rsd)
                                  `(,(sb-kernel::raw-slot-data-raw-type rsd)
@@ -114,7 +116,7 @@
            (:dd
             (vop set-slot node block object
                  (emit-constant (sb-kernel::dd-layout-or-lose slot))
-                 name sb-vm:instance-slots-offset lowtag))
+                 :allocator sb-vm:instance-slots-offset lowtag))
            (otherwise
             (if (and (eq kind :arg)
                      (zero-init-p (car args)))
@@ -146,16 +148,25 @@
                                              nil sb-vm:any-reg-sc-number)))
                                     (vop make-funcallable-instance-tramp node block tn)
                                     tn)))))
-                     name slot lowtag))))))))
+                     :allocator slot lowtag))))))))
   (unless (null args)
     (bug "Leftover args: ~S" args)))
 
+(defun unbound-marker-tn-p (tn)
+  (let ((writes (tn-writes tn)))
+    (and writes
+         (not (tn-ref-next writes)) ; is never changed
+         (let ((vop (tn-ref-vop writes)))
+           (and vop (eq (vop-name vop) 'make-unbound-marker))))))
+
 (defun emit-fixed-alloc (node block name words type lowtag result lvar)
   (let ((stack-allocate-p (and lvar (lvar-dynamic-extent lvar))))
-    (when stack-allocate-p
-      (vop current-stack-pointer node block
-           (ir2-lvar-stack-pointer (lvar-info lvar))))
-    (vop fixed-alloc node block name words type lowtag stack-allocate-p result)))
+    (cond (stack-allocate-p
+           (vop current-stack-pointer node block
+                (ir2-lvar-stack-pointer (lvar-info lvar)))
+           (vop fixed-alloc-to-stack node block name words type lowtag t result))
+          (t
+           (vop fixed-alloc node block name words type lowtag nil result)))))
 
 (defoptimizer ir2-convert-fixed-allocation
               ((&rest args) node block name words type lowtag inits)
@@ -163,7 +174,7 @@
          (locs (lvar-result-tns lvar (list *universal-type*)))
          (result (first locs)))
     (emit-fixed-alloc node block name words type lowtag result lvar)
-    (emit-inits node block name result lowtag inits args)
+    (emit-inits node block result lowtag inits args)
     (move-lvar-result node block locs lvar)))
 
 (defoptimizer ir2-convert-variable-allocation
@@ -180,7 +191,7 @@
                  (ir2-lvar-stack-pointer (lvar-info lvar))))
           (vop var-alloc node block (lvar-tn node block extra) name words
                type lowtag stack-allocate-p result)))
-    (emit-inits node block name result lowtag inits args)
+    (emit-inits node block result lowtag inits args)
     (move-lvar-result node block locs lvar)))
 
 (defoptimizer ir2-convert-structure-allocation
@@ -190,17 +201,24 @@
          (locs (lvar-result-tns lvar (list *universal-type*)))
          (result (first locs)))
     (aver (and (constant-lvar-p dd) (constant-lvar-p slot-specs) (= words 1)))
+    (aver (= type sb-vm:instance-widetag))
     (let* ((c-dd (lvar-value dd))
            (c-slot-specs (lvar-value slot-specs))
            (words (+ (dd-length c-dd) words)))
-      #+compact-instance-header
-      (progn (aver (= type sb-vm:instance-widetag))
-             (emit-constant (setq type (sb-kernel::dd-layout-or-lose c-dd))))
-      (emit-fixed-alloc node block name words type lowtag result lvar)
-      (emit-inits node block name result lowtag
+      (let ((metadata #+compact-instance-header
+                      (let ((layout (sb-kernel::dd-layout-or-lose c-dd)))
+                        (emit-constant layout)
+                        layout)
+                      #-compact-instance-header
+                      c-dd))
+        (emit-fixed-alloc node block name words metadata lowtag result lvar))
+      (emit-inits node block result lowtag
                   `(#-compact-instance-header (:dd . ,c-dd) ,@c-slot-specs) args)
       (move-lvar-result node block locs lvar))))
 
+;;; FIXME: this causes emission of GC store barriers, but it should not.
+;;; The vector is freshly consed, so anything being stored into it
+;;; is at least as old.
 (defoptimizer (initialize-vector ir2-convert)
     ((vector &rest initial-contents) node block)
   (let* ((vector-ctype (lvar-type vector))
@@ -209,13 +227,14 @@
                         (bug "Unknown vector type in IR2 conversion for ~S."
                              'initialize-vector)))
          (bit-vector-p (type= elt-ctype (specifier-type 'bit)))
+         (simple-vector-p (type= elt-ctype (specifier-type 't)))
          (saetp (find-saetp-by-ctype elt-ctype))
          (lvar (node-lvar node))
          (locs (lvar-result-tns lvar (list vector-ctype)))
          (result (first locs))
          (elt-ptype (primitive-type elt-ctype))
          (tmp (make-normal-tn elt-ptype)))
-    (declare (ignorable bit-vector-p))
+    (declare (ignorable bit-vector-p simple-vector-p))
     (emit-move node block (lvar-tn node block vector) result)
     (flet ((compute-setter ()
              ;; Such cringe. I had no idea why all the "-C" vops were mandatory.
@@ -278,9 +297,14 @@
               ;; Nonetheless it's far better than it was. In all other scenarios, don't pass
               ;; a constant TN, because we don't know that generated code is better.
               (cond #+x86-64 ; still moar cringe
-                    ((and bit-vector-p (constant-lvar-p value))
+                    ((and (or bit-vector-p simple-vector-p) (constant-lvar-p value))
                      (funcall setter (tnify i) (emit-constant (lvar-value value))))
                     (t
+                     ;; FIXME: for simple-vector, fixnums should get stored via an ANY-REG
+                     ;; so that data-vector-set doesn't emit a store barrier.
+                     ;; (TMP is a descriptor-reg because ELT-CTYPE is T)
+                     ;; Or just fix this optimizer - as commented above - to somehow elide
+                     ;; the barrier for all elements.
                      (emit-move node block (lvar-tn node block value) tmp)
                      (funcall setter (tnify i) tmp))))))))
     (move-lvar-result node block locs lvar)))
@@ -466,10 +490,3 @@
                     (primitive-object-slots (primitive-object 'array)))
                '(fill-pointer elements data
                  displacement displaced-p displaced-from dimensions)))
-
-(defun emit-code-page-write-barrier-p (fun-name)
-  (and (listp fun-name)
-       (eq (car fun-name) 'setf)
-       (member (cadr fun-name)
-               '(%code-debug-info %code-fixups))
-       t))
