@@ -20,15 +20,14 @@
 ;;;; Note: Take a look at the compiler-overview.tex section on "Hairy
 ;;;; function representation" before you seriously mess with this
 ;;;; stuff.
-#-sb-devel
 (declaim (start-block ir1-convert-lambda ir1-convert-lambda-body
                       ir1-convert-aux-bindings varify-lambda-arg
                       ir1-convert-lambdalike))
 
 ;;; Return a VAR structure for NAME, filling in info if it is globally
 ;;; special. If it is losing, we punt with a COMPILER-ERROR.
-(declaim (ftype (sfunction (symbol) lambda-var) varify-lambda-arg))
-(defun varify-lambda-arg (name)
+(declaim (ftype (sfunction (symbol &optional t) lambda-var) varify-lambda-arg))
+(defun varify-lambda-arg (name &optional source-form)
   (case (info :variable :kind name)
     (:special
      (let ((variable (find-free-var name)))
@@ -37,7 +36,8 @@
                         :where-from (leaf-where-from variable)
                         :specvar variable)))
     (t
-     (make-lambda-var :%source-name name))))
+     (make-lambda-var :%source-name name
+                      :source-form source-form))))
 
 ;;; Parse a lambda list into a list of VAR structures, stripping off
 ;;; any &AUX bindings. Each arg name is checked for legality, and
@@ -67,7 +67,8 @@
              (add-info (var kind &key (default nil defaultp) suppliedp-var key)
                (let ((info (make-arg-info :kind kind)))
                  (when defaultp
-                   (setf (arg-info-default info) default))
+                   (setf (arg-info-default info) default
+                         (arg-info-default-p info) t))
                  (when suppliedp-var
                    (setf (arg-info-supplied-p info)
                          (varify-lambda-arg suppliedp-var)))
@@ -164,7 +165,7 @@
           (bind-ctran (make-ctran))
           (cleanup-ctran (make-ctran)))
       (ir1-convert start bind-ctran nil
-                   `(%special-bind ',(leaf-source-name (lambda-var-specvar var)) ,var))
+                   `(%special-bind ',(lambda-var-specvar var) ,var))
       (setf (cleanup-mess-up cleanup) (ctran-use bind-ctran))
       (let ((*lexenv* (make-lexenv :cleanup cleanup)))
         (ir1-convert bind-ctran cleanup-ctran nil '(%cleanup-point))
@@ -200,7 +201,8 @@
                                 debug-name
                                 (note-lexical-bindings t)
                                 post-binding-lexenv
-                                system-lambda)
+                                system-lambda
+                                local-policy)
   (declare (list body vars aux-vars aux-vals))
 
   ;; We're about to try to put new blocks into *CURRENT-COMPONENT*.
@@ -208,10 +210,13 @@
 
   (let* ((bind (make-bind))
          (lambda (make-lambda :vars vars
-                  :bind bind
-                  :%source-name source-name
-                  :%debug-name debug-name
-                  :system-lambda-p system-lambda))
+                              :bind bind
+                              :%source-name source-name
+                              :%debug-name debug-name
+                              :system-lambda-p system-lambda
+                              :lexenv (if local-policy
+                                          (make-lexenv :policy local-policy)
+                                          *lexenv*)))
          (result-ctran (make-ctran))
          (result-lvar (make-lvar)))
     ;; just to check: This function should fail internal assertions if
@@ -304,7 +309,7 @@
          (fun (collect ((default-bindings)
                         (default-vals))
                 (dolist (default defaults)
-                  (if (sb-xc:constantp default)
+                  (if (constantp default)
                       (default-vals default)
                       (let ((var (sb-xc:gensym)))
                         (default-bindings `(,var ,default))
@@ -358,7 +363,7 @@
          (default (arg-info-default info))
          (supplied-p (arg-info-supplied-p info))
          (force (or force
-                    (not (sb-xc:constantp (arg-info-default info)))))
+                    (not (constantp (arg-info-default info)))))
          (ep (if supplied-p
                  (ir1-convert-hairy-args
                   res
@@ -479,7 +484,9 @@
               (found-allow-p nil))
 
           (temps #-stack-grows-downward-not-upward
-                 `(,n-index (1- ,n-count))
+                 `(,n-index (+ ,n-count ,(if (vop-existsp :translate %more-keyword-pair)
+                                             0
+                                             -1)))
                  #+stack-grows-downward-not-upward
                  `(,n-index (- (1- ,n-count)))
                  #-stack-grows-downward-not-upward n-value-temp
@@ -535,7 +542,9 @@
                        (setq ,n-lose ,n-key))))
 
             (body
-             `(when (oddp ,n-count)
+             `(when (oddp ,(if (vop-existsp :translate %more-keyword-pair)
+                               n-index
+                               n-count))
                 (%odd-key-args-error)))
 
             (body
@@ -543,12 +552,19 @@
              `(locally
                 (declare (optimize (safety 0)))
                 (loop
-                  (when (minusp ,n-index) (return))
-                  (setf ,n-value-temp (%more-arg ,n-context ,n-index))
-                  (decf ,n-index)
-                  (setq ,n-key (%more-arg ,n-context ,n-index))
-                  (decf ,n-index)
-                  (case ,n-key ,@(tests))))
+                 ,@(cond ((vop-existsp :translate %more-keyword-pair)
+                          `((when (zerop ,n-index) (return))
+                            (decf ,n-index 2)
+                            (multiple-value-bind (key value)
+                                (%more-keyword-pair ,n-context ,n-index)
+                              (setf ,n-value-temp value ,n-key key))))
+                         (t
+                          `((when (minusp ,n-index) (return))
+                            (setf ,n-value-temp (%more-arg ,n-context ,n-index))
+                            (decf ,n-index)
+                            (setq ,n-key (%more-arg ,n-context ,n-index))
+                            (decf ,n-index))))
+                 (case ,n-key ,@(tests))))
              #+stack-grows-downward-not-upward
              `(locally (declare (optimize (safety 0)))
                 (loop
@@ -560,7 +576,7 @@
                     (case ,n-key ,@(tests))))))
 
             (unless allowp
-              (let ((location (opaquely-quote (make-restart-location))))
+              (let ((location (make-restart-location)))
                 (body `(if (and (not (unbound-marker-p ,n-lose))
                                 (not ,n-allowp))
                            (%unknown-key-arg-error ,n-lose ,location)
@@ -636,7 +652,7 @@
     (dolist (key keys)
       (let* ((info (lambda-var-arg-info key))
              (default (arg-info-default info))
-             (hairy-default (not (sb-xc:constantp default)))
+             (hairy-default (not (constantp default)))
              (supplied-p (arg-info-supplied-p info))
              ;; was: (format nil "~A-DEFAULTING-TEMP" (leaf-source-name key))
              (n-val (make-symbol ".DEFAULTING-TEMP."))
@@ -878,56 +894,61 @@
     (bug "Both SYSTEM-LAMBDA and MAYBE-ADD-DEBUG-CATCH specified"))
   (unless (or debug-name (neq '.anonymous. source-name))
     (setf debug-name (name-lambdalike form)))
-  (multiple-value-bind (forms decls doc) (parse-body (cddr form) t)
-    (let ((*lexenv* (process-muffle-decls decls *lexenv*)))
-      (multiple-value-bind (vars keyp allow-other-keys aux-vars aux-vals)
-          (make-lambda-vars (cadr form))
-        (binding* (((*lexenv* result-type post-binding-lexenv
-                              lambda-list explicit-check source-form)
-                    (process-decls decls (append aux-vars vars) nil
-                                   :binding-form-p t :allow-lambda-list t))
-                   (debug-catch-p (and maybe-add-debug-catch
-                                       *allow-instrumenting*
-                                       (policy *lexenv*
-                                           (>= insert-debug-catch 2))))
-                   (forms (if debug-catch-p
-                              (wrap-forms-in-debug-catch forms)
-                              forms))
-                   (forms (if (eq result-type *wild-type*)
-                              forms
-                              `((the ,(type-specifier result-type) (progn ,@forms)))))
-                   (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
-                   (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
-                               (ir1-convert-hairy-lambda forms vars keyp
-                                                         allow-other-keys
-                                                         aux-vars aux-vals
-                                                         :post-binding-lexenv post-binding-lexenv
-                                                         :source-name source-name
-                                                         :debug-name debug-name
-                                                         :system-lambda system-lambda))
-                              (t
-                               (ir1-convert-lambda-body forms vars
-                                                        :aux-vars aux-vars
-                                                        :aux-vals aux-vals
-                                                        :post-binding-lexenv post-binding-lexenv
-                                                        :source-name source-name
-                                                        :debug-name debug-name
-                                                        :system-lambda system-lambda)))))
-          (when explicit-check
-            (setf (getf (functional-plist res) 'explicit-check) explicit-check))
-          (setf (functional-inline-expansion res) (or source-form form))
-          (setf (functional-arg-documentation res)
-                (if (eq lambda-list :unspecified)
-                    (strip-lambda-list (cadr form) :arglist)
-                    lambda-list))
-          (setf (functional-documentation res) doc)
-          (when (boundp '*lambda-conversions*)
-            ;; KLUDGE: Not counting TL-XEPs is a lie, of course, but
-            ;; keeps things less confusing to users of TIME, where this
-            ;; count gets used.
-            (unless (and (consp debug-name) (eq 'tl-xep (car debug-name)))
-              (incf *lambda-conversions*)))
-          res)))))
+  (binding* (((forms decls doc) (parse-body (cddr form) t))
+             ((*lexenv* source-form) (process-muffle-decls decls *lexenv*))
+             (*current-path* (or (and source-form
+                                      (get-source-path source-form))
+                                 *current-path*))
+             ((vars keyp allow-other-keys aux-vars aux-vals)
+              (make-lambda-vars (cadr form)))
+             ((*lexenv* result-type post-binding-lexenv
+                        lambda-list explicit-check source-form
+                        local-policy)
+              (process-decls decls (append aux-vars vars) nil
+                             :binding-form-p t :allow-lambda-list t))
+             (debug-catch-p (and maybe-add-debug-catch
+                                 *allow-instrumenting*
+                                 (policy *lexenv*
+                                     (>= insert-debug-catch 2))))
+             (forms (if debug-catch-p
+                        (wrap-forms-in-debug-catch forms)
+                        forms))
+             (forms (if (eq result-type *wild-type*)
+                        forms
+                        `((the ,(type-specifier result-type) (progn ,@forms)))))
+             (*allow-instrumenting* (and (not system-lambda) *allow-instrumenting*))
+             (res (cond ((or (find-if #'lambda-var-arg-info vars) keyp)
+                         (ir1-convert-hairy-lambda forms vars keyp
+                                                   allow-other-keys
+                                                   aux-vars aux-vals
+                                                   :post-binding-lexenv post-binding-lexenv
+                                                   :source-name source-name
+                                                   :debug-name debug-name
+                                                   :system-lambda system-lambda))
+                        (t
+                         (ir1-convert-lambda-body forms vars
+                                                  :aux-vars aux-vars
+                                                  :aux-vals aux-vals
+                                                  :post-binding-lexenv post-binding-lexenv
+                                                  :source-name source-name
+                                                  :debug-name debug-name
+                                                  :system-lambda system-lambda
+                                                  :local-policy local-policy)))))
+    (when explicit-check
+      (setf (getf (functional-plist res) 'explicit-check) explicit-check))
+    (setf (functional-inline-expansion res) (or source-form form))
+    (setf (functional-arg-documentation res)
+          (if (eq lambda-list :unspecified)
+              (strip-lambda-list (cadr form) :arglist)
+              lambda-list))
+    (setf (functional-documentation res) doc)
+    (when (boundp '*lambda-conversions*)
+      ;; KLUDGE: Not counting TL-XEPs is a lie, of course, but
+      ;; keeps things less confusing to users of TIME, where this
+      ;; count gets used.
+      (unless (and (consp debug-name) (eq 'tl-xep (car debug-name)))
+        (incf *lambda-conversions*)))
+    res))
 
 (defun wrap-forms-in-debug-catch (forms)
   #+unwind-to-frame-and-call-vop
@@ -1031,8 +1052,8 @@
 
 (declaim (end-block))
 
-
 ;;;; defining global functions
+
 ;;; Given a lambda-list, return a FUN-TYPE object representing the signature:
 ;;; return type is *, and each individual arguments type is T -- but we get
 ;;; the argument counts and keywords.
@@ -1054,7 +1075,70 @@
             (allow (when (ll-kwds-allowp llks) '(&allow-other-keys))))
         (careful-specifier-type `(function (,@reqs ,@opts ,@rest ,@keys ,@allow) *))))))
 
-#-sb-devel
+(declaim (start-block maybe-inline-syntactic-closure))
+
+;;; Return a lambda form that has been "closed" with respect to
+;;; LEXENV, returning a LAMBDA-WITH-LEXENV if there are interesting
+;;; declarations. To handle local macros, rather than closing over
+;;; definitions in the environment, expand all macros in the body of
+;;; LAMBDA, so that nothing in the syntactic environment is needed in
+;;; the expansion. If there is something too complex in the lexical
+;;; environment (like a lexical variable), then we return NIL.
+(defun maybe-inline-syntactic-closure (lambda lexenv)
+  (declare (type list lambda) (type lexenv-designator lexenv))
+  (aver (eql (first lambda) 'lambda))
+  (typecase lexenv
+   (lexenv
+    (let ((vars (lexenv-vars lexenv))
+          (funs (lexenv-funs lexenv))
+          (decls ()))
+      (cond ((or (lexenv-blocks lexenv) (lexenv-tags lexenv)) nil)
+            ((and (null vars) (null funs)) lambda)
+            ((dolist (x vars nil)
+               (let ((name (car x))
+                     (what (cdr x)))
+                 (when (eq x (assoc name vars :test #'eq))
+                   (typecase what
+                     (cons
+                      (aver (eq (car what) 'macro)))
+                     (global-var
+                      (aver (eq (global-var-kind what) :special))
+                      (push `(special ,name) decls))
+                     (t (return t))))))
+             nil)
+            ((dolist (x funs nil)
+               (let ((name (car x))
+                     (what (cdr x)))
+                 (when (eq x (assoc name funs :test #'equal))
+                   (typecase what
+                     (cons)
+                     ;; FIXME: Is there a good reason for this not to be
+                     ;; DEFINED-FUN (which :INCLUDEs GLOBAL-VAR, in case
+                     ;; you're wondering how this ever worked :-)? Maybe
+                     ;; in conjunction with an AVERrance that it's not an
+                     ;; (AND GLOBAL-VAR (NOT GLOBAL-FUN))? -- CSR,
+                     ;; 2002-07-08
+                     (global-var
+                      (when (defined-fun-p what)
+                        (push `(,(car (defined-fun-inlinep what))
+                                ,name)
+                              decls)))
+                     (t (return t))))))
+             nil)
+            (t
+             (let ((expansion (sb-walker:macroexpand-all lambda lexenv)))
+               (if decls
+                   `(lambda-with-lexenv (:declare ,decls) ,@(cdr expansion))
+                   expansion))))))
+   #+(and sb-fasteval (not sb-xc-host))
+   (sb-interpreter:basic-env
+    (awhen (sb-interpreter::reconstruct-syntactic-closure-env lexenv)
+      `(lambda-with-lexenv ,it ,@(cdr lambda))))
+   #+sb-fasteval
+   (null lambda))) ; trivial case. Never occurs in the compiler.
+
+(declaim (end-block))
+
 (declaim (start-block ir1-convert-inline-lambda))
 
 ;;; Convert the forms produced by RECONSTRUCT-LEXENV to LEXENV
@@ -1123,6 +1207,7 @@
                 (lexenv-flushable *lexenv*)
                 lexenv-lambda
                 *lexenv*)))
+         (*inlining* (1+ *inlining*))
          (clambda (ir1-convert-lambda body
                                       :source-name source-name
                                       :debug-name debug-name
@@ -1137,39 +1222,38 @@
 ;;; previous references.
 (defun get-defined-fun (name &optional (lambda-list nil lp))
   (proclaim-as-fun-name name)
-  (when #+sb-xc-host (not *compile-time-eval*)
-        #-sb-xc-host (boundp '*ir1-namespace*)
-    (let ((found (find-free-fun name "shouldn't happen! (defined-fun)"))
-          (free-funs (free-funs *ir1-namespace*)))
-      (note-name-defined name :function)
-      (cond ((not (defined-fun-p found))
-             ;; This assertion is wrong in block compilation mode, for
-             ;; instance
-             ;;
-             ;; (defun foo (x) (bar x))
-             ;; (declaim (inline bar))
-             ;; (defun bar (x) x)
-             (aver (or (block-compile *compilation*)
-                       (not (info :function :inlinep name))))
-             (let* ((where-from (leaf-where-from found))
-                    (res (make-defined-fun
-                          :%source-name name
-                          :where-from (if (eq where-from :declared)
-                                          :declared
-                                          :defined-here)
-                          :type (if (eq :declared where-from)
-                                    (leaf-type found)
-                                    (or (and lp
-                                             (ftype-from-lambda-list lambda-list))
-                                        (specifier-type 'function))))))
-               (substitute-leaf res found)
-               (setf (gethash name free-funs) res)))
-            ;; If FREE-FUNS has a previously converted definition
-            ;; for this name, then blow it away and try again.
-            ((defined-fun-functionals found)
-             (remhash name free-funs)
-             (get-defined-fun name lambda-list))
-            (t found)))))
+  (let ((found (find-free-fun name "shouldn't happen! (defined-fun)"))
+        (free-funs (free-funs *ir1-namespace*)))
+    (note-name-defined name :function)
+    (cond ((not (defined-fun-p found))
+           ;; This assertion is wrong in block compilation mode, for
+           ;; instance
+           ;;
+           ;; (defun foo (x) (bar x))
+           ;; (declaim (inline bar))
+           ;; (defun bar (x) x)
+           (aver (or (block-compile *compilation*)
+                     (not (info :function :inlinep name))))
+           (let* ((where-from (leaf-where-from found))
+                  (res (make-defined-fun
+                        :%source-name name
+                        :where-from (if (eq where-from :declared)
+                                        :declared
+                                        :defined-here)
+                        :type (if (eq :declared where-from)
+                                  (leaf-type found)
+                                  (or (and lp
+                                           (ignore-errors
+                                            (ftype-from-lambda-list lambda-list)))
+                                      (specifier-type 'function))))))
+             (substitute-leaf res found)
+             (setf (gethash name free-funs) res)))
+          ;; If FREE-FUNS has a previously converted definition
+          ;; for this name, then blow it away and try again.
+          ((defined-fun-functionals found)
+           (remhash name free-funs)
+           (get-defined-fun name lambda-list))
+          (t found))))
 
 ;;; Check a new global function definition for consistency with
 ;;; previous declaration or definition, and assert argument/result
@@ -1243,7 +1327,7 @@
         (note-inlining (optional-dispatch-more-entry fun))
         (mapc #'note-inlining (optional-dispatch-entry-points fun))))
     ;; substitute for any old references
-    (unless (or (eq (defined-fun-inlinep var) :notinline)
+    (unless (or (eq (defined-fun-inlinep var) 'notinline)
                 (not (block-compile *compilation*))
                 (and info
                      (or (fun-info-transforms info)
@@ -1373,27 +1457,3 @@ is potentially harmful to any already-compiled callers using (SAFETY 0)."
             (specifier-type 'function))))
 
   (values))
-
-;; Similar to above, detect duplicate definitions within a file,
-;; but the package lock check is unnecessary - it's handled elsewhere.
-;;
-;; Additionally, this is a STYLE-WARNING, not a WARNING, because there is
-;; meaningful behavior that can be ascribed to some redefinitions, e.g.
-;;  (defmacro foo () first-definition)
-;;  (defun f () (use-it (foo )))
-;;  (defmacro foo () other-definition)
-;; will use the first definition when compiling F, but make the second available
-;; in the loaded fasl. In this usage it would have made sense to wrap the
-;; respective definitions with EVAL-WHEN for different situations,
-;; but as long as the compile-time behavior is deterministic, it's just bad style
-;; and not flat-out wrong, though there is indeed some waste in the fasl.
-;;
-;; KIND is the globaldb KIND of this NAME
-(defun %compiler-defmacro (kind name)
-  (let ((name-key `(,kind ,name)))
-    (when (boundp '*lexenv*)
-      (aver (producing-fasl-file))
-      ;; a slight OAOO issue here wrt %COMPILER-DEFUN
-      (if (member name-key (fun-names-in-this-file *compilation*) :test #'equal)
-          (compiler-style-warn 'same-file-redefinition-warning :name name)
-          (push name-key (fun-names-in-this-file *compilation*))))))
