@@ -374,8 +374,6 @@
                        (type (type-specifier (info :variable :type name))))
                    `(macro . (the ,type ,expansion))))
                 (:constant
-                 (when (producing-fasl-file)
-                   (emit-load-time-constant-reference name))
                  (find-constant (symbol-value name) name))
                 (t
                  (make-global-var :kind kind
@@ -412,7 +410,9 @@
   (dx-let ((things-processed (alloc-xset)))
     (named-let grovel ((value constant))
       (unless (or (dumpable-leaflike-p value)
-                  (xset-member-p value things-processed))
+                  (xset-member-p value things-processed)
+                  #-sb-xc-host
+                  (unbound-marker-p value))
         (add-to-xset value things-processed)
         ;; FIXME: shouldn't this be something like SB-XC:TYPECASE ?
         (typecase value
@@ -435,13 +435,7 @@
           ((array t)
            (dotimes (i (array-total-size value))
              (grovel (row-major-aref value i))))
-          (opaque-box)
           (instance
-           ;; In the target SBCL, we can dump any instance, but
-           ;; in the cross-compilation host, %INSTANCE-FOO
-           ;; functions don't work on general instances, only on
-           ;; STRUCTURE!OBJECTs.
-           ;;
            ;; Behold the wonderfully clear sense of this-
            ;;  WHEN (EMIT-MAKE-LOAD-FORM VALUE)
            ;; meaning "when you're _NOT_ using a custom load-form"
@@ -1143,13 +1137,15 @@
            (type (or lvar null) result)
            (list form)
            (type global-var var))
-  (let ((info (info :function :info (leaf-source-name var))))
-    (if (and info
-             (ir1-attributep (fun-info-attributes info) predicate)
-             (not (if-p (and result (lvar-dest result)))))
-        (let ((*instrument-if-for-code-coverage* nil))
-          (ir1-convert start next result `(if ,form t nil)))
-        (ir1-convert-combination-checking-type start next result form var))))
+  (if (vop-existsp :named sb-vm::move-conditional-result)
+      (ir1-convert-combination-checking-type start next result form var)
+      (let ((info (info :function :info (leaf-source-name var))))
+        (if (and info
+                 (ir1-attributep (fun-info-attributes info) predicate)
+                 (not (if-p (and result (lvar-dest result)))))
+            (let ((*instrument-if-for-code-coverage* nil))
+              (ir1-convert start next result `(if ,form t nil)))
+            (ir1-convert-combination-checking-type start next result form var)))))
 
 ;;; Actually really convert a global function call that we are allowed
 ;;; to early-bind.
@@ -1685,15 +1681,18 @@
      optimize-qualities)))
 
 (defun process-muffle-decls (decls lexenv)
-  (flet ((process-it (spec)
-           (cond ((atom spec))
-                 ((member (car spec) '(muffle-conditions unmuffle-conditions))
-                  (setq lexenv
-                        (process-1-decl spec lexenv nil nil nil nil))))))
-    (dolist (decl decls)
-      (dolist (spec (rest decl))
-        (process-it spec))))
-  lexenv)
+  (let (source-form)
+   (flet ((process-it (spec)
+            (cond ((atom spec))
+                  ((member (car spec) '(muffle-conditions unmuffle-conditions))
+                   (setq lexenv
+                         (process-1-decl spec lexenv nil nil nil nil)))
+                  ((eq (car spec) 'source-form)
+                   (setf source-form (cadr spec))))))
+     (dolist (decl decls)
+       (dolist (spec (rest decl))
+         (process-it spec))))
+    (values lexenv source-form)))
 
 ;;; Use a list of DECLARE forms to annotate the lists of LAMBDA-VAR
 ;;; and FUNCTIONAL structures which are being bound. In addition to
@@ -1763,8 +1762,7 @@
                    ;; E.g. suppressing argument checking without doing
                    ;; so in all the subforms.
                    ((typep spec '(cons (eql local-optimize)))
-                    (setf local-optimize spec)
-                    (setf source-form (cadr spec)))
+                    (setf local-optimize spec))
                    (t
                     (multiple-value-bind (new-env new-qualities)
                         (process-1-decl spec lexenv vars fvars
