@@ -164,8 +164,9 @@
   ;; the basic block this continuation is in. This is null only in
   ;; :UNUSED continuations.
   (block nil :type (or cblock null))
-  ;; Entries created by the BLOCK special operator
-  (entries nil :type list))
+  ;; Use for reporting notes for the following node,
+  ;; which can be transformed and lose its original source code.
+  (source-path nil :type list))
 
 (defmethod print-object ((x ctran) stream)
   (print-unreadable-object (x stream :type t :identity t)
@@ -197,6 +198,7 @@
   (dependent-nodes nil)
   (annotations nil)
   (dependent-annotations nil))
+(!set-load-form-method lvar (:xc :target) :ignore-it)
 
 ;;; These are used for annotating a LVAR with information that can't
 ;;; be expressed using types or if the CAST semantics are undesirable
@@ -323,9 +325,9 @@
   ;; top level form containing the original source.
   (source-path *current-path* :type list)
   ;; If this node is in a tail-recursive position, then this is set to
-  ;; T. At the end of IR1 (in physical environment analysis) this is
-  ;; computed for all nodes (after cleanup code has been emitted).
-  ;; Before then, a non-null value indicates that IR1 optimization has
+  ;; T. At the end of IR1 (in environment analysis) this is computed
+  ;; for all nodes (after cleanup code has been emitted).  Before
+  ;; then, a non-null value indicates that IR1 optimization has
   ;; converted a tail local call to a direct transfer.
   ;;
   ;; If the back-end breaks tail-recursion for some reason, then it
@@ -496,9 +498,9 @@
   ;; what macroexpansions and source transforms happened "in" this block, used
   ;; for xref
   (xrefs nil :type list)
-  ;; Cache the physenv of a block during lifetime analysis. :NONE if
-  ;; no cached value has been stored yet.
-  (physenv-cache :none :type (or null physenv (member :none))))
+  ;; Cache the environment of a block during lifetime analysis. :NONE
+  ;; if no cached value has been stored yet.
+  (environment-cache :none :type (or null environment (member :none))))
 (defmethod print-object ((cblock cblock) stream)
   (if (boundp '*compilation*)
       (print-unreadable-object (cblock stream :type t :identity t)
@@ -577,9 +579,9 @@
   ;; Entry/exit points have these blocks as their
   ;; predecessors/successors. The start and return from each
   ;; non-deleted function is linked to the component head and
-  ;; tail. Until physical environment analysis links NLX entry stubs
-  ;; to the component head, every successor of the head is a function
-  ;; start (i.e. begins with a BIND node.)
+  ;; tail. Until environment analysis links NLX entry stubs to the
+  ;; component head, every successor of the head is a function start
+  ;; (i.e. begins with a BIND node.)
   (head (missing-arg) :type cblock)
   (tail (missing-arg) :type cblock)
   ;; New blocks are inserted before this.
@@ -630,13 +632,13 @@
   ;; from COMPONENT-LAMBDAS.
   (reanalyze-functionals nil :type list)
   (delete-blocks nil :type list)
-  (nlx-info-generated-p nil :type boolean)
-  ;; this is filled by physical environment analysis
+  ;; this is filled by environment analysis
   (dx-lvars nil :type list)
   ;; The default LOOP in the component.
   (outer-loop (missing-arg) :type cloop)
-  ;; The current sset index
-  (sset-number 0 :type fixnum))
+  (max-block-number 0 :type fixnum)
+  (dominators-computed nil))
+
 (defprinter (component :identity t)
   name
   (reanalyze :test reanalyze))
@@ -683,59 +685,37 @@
   (mess-up nil :type (or node null))
   ;; For all kinds, except :DYNAMIC-EXTENT: a list of all the NLX-INFO
   ;; structures whose NLX-INFO-CLEANUP is this cleanup. This is filled
-  ;; in by physical environment analysis.
+  ;; in by environment analysis.
   ;;
   ;; For :DYNAMIC-EXTENT: a list of all DX LVARs, preserved by this
   ;; cleanup. This is filled when the cleanup is created (now by
-  ;; locall call analysis) and is rechecked by physical environment
-  ;; analysis. (For closures this is a list of the enclose node during
-  ;; IR1, and a list of the LVAR of the enclose after physical
-  ;; environment analysis.)
-  (info nil :type list))
+  ;; locall call analysis) and is rechecked by environment
+  ;; analysis. (For closures this is a list of the LVAR of the enclose
+  ;; after environment analysis.)
+  (nlx-info nil :type list))
 (defprinter (cleanup :identity t)
   kind
   mess-up
-  (info :test info))
+  (nlx-info :test nlx-info))
 
-;;; A PHYSENV represents the result of physical environment analysis.
-;;;
-;;; As far as I can tell from reverse engineering, this IR1 structure
-;;; represents the physical environment (which is probably not the
-;;; standard Lispy term for this concept, but I dunno what is the
-;;; standard term): those things in the lexical environment which a
-;;; LAMBDA actually interacts with. Thus in
-;;;   (DEFUN FROB-THINGS (THINGS)
-;;;     (DOLIST (THING THINGS)
-;;;       (BLOCK FROBBING-ONE-THING
-;;;         (MAPCAR (LAMBDA (PATTERN)
-;;;                   (WHEN (FITS-P THING PATTERN)
-;;;                     (RETURN-FROM FROB-THINGS (LIST :FIT THING PATTERN))))
-;;;                 *PATTERNS*))))
-;;; the variables THINGS, THING, and PATTERN and the block names
-;;; FROB-THINGS and FROBBING-ONE-THING are all in the inner LAMBDA's
-;;; lexical environment, but of those only THING, PATTERN, and
-;;; FROB-THINGS are in its physical environment. In IR1, we largely
-;;; just collect the names of these things; in IR2 an IR2-PHYSENV
-;;; structure is attached to INFO and used to keep track of
-;;; associations between these names and less-abstract things (like
-;;; TNs, or eventually stack slots and registers). -- WHN 2001-09-29
-(defstruct (physenv (:copier nil))
-  ;; the function that allocates this physical environment
+;;; The ENVIRONMENT structure represents the result of environment analysis.
+(defstruct (environment (:copier nil))
+  ;; the function that allocates this environment
   (lambda (missing-arg) :type clambda :read-only t)
   ;; This ultimately converges to a list of all the LAMBDA-VARs and
   ;; NLX-INFOs needed from enclosing environments by code in this
-  ;; physical environment. In the meantime, it may be
+  ;; environment. In the meantime, it may be
   ;;   * NIL at object creation time
   ;;   * a superset of the correct result, generated somewhat later
   ;;   * smaller and smaller sets converging to the correct result as
   ;;     we notice and delete unused elements in the superset
   (closure nil :type list)
   ;; a list of NLX-INFO structures describing all the non-local exits
-  ;; into this physical environment
+  ;; into this environment
   (nlx-info nil :type list)
   ;; some kind of info used by the back end
   (info nil))
-(defprinter (physenv :identity t)
+(defprinter (environment :identity t)
   lambda
   (closure :test closure)
   (nlx-info :test nlx-info))
@@ -770,11 +750,10 @@
 ;;; An NLX-INFO structure is used to collect various information about
 ;;; non-local exits. This is effectively an annotation on the
 ;;; continuation, although it is accessed by searching in the
-;;; PHYSENV-NLX-INFO.
+;;; ENVIRONMENT-NLX-INFO.
 (defstruct (nlx-info
             (:copier nil)
-            (:constructor make-nlx-info
-                (cleanup exit &aux (block (first (block-succ (node-block exit)))))))
+            (:constructor make-nlx-info (cleanup block)))
   ;; the cleanup associated with this exit. In a catch or
   ;; unwind-protect, this is the :CATCH or :UNWIND-PROTECT cleanup,
   ;; and not the cleanup for the escape block. The CLEANUP-KIND of
@@ -783,7 +762,7 @@
   (cleanup (missing-arg) :type cleanup)
   ;; the ``continuation'' exited to (the block, succeeding the EXIT
   ;; nodes). If this exit is from an escape function (CATCH or
-  ;; UNWIND-PROTECT), then physical environment analysis deletes the
+  ;; UNWIND-PROTECT), then environment analysis deletes the
   ;; escape function and instead has the %NLX-ENTRY use this
   ;; continuation.
   ;;
@@ -792,9 +771,9 @@
   ;; ENTRY must also be used to disambiguate, since exits to different
   ;; places may deliver their result to the same continuation.
   (block (missing-arg) :type cblock)
-  ;; the entry stub inserted by physical environment analysis. This is
-  ;; a block containing a call to the %NLX-ENTRY funny function that
-  ;; has the original exit destination as its successor. Null only
+  ;; the entry stub inserted by environment analysis. This is a block
+  ;; containing a call to the %NLX-ENTRY funny function that has the
+  ;; original exit destination as its successor. Null only
   ;; temporarily.
   (target nil :type (or cblock null))
   ;; for a lexical exit it determines whether tag existence check is
@@ -806,6 +785,8 @@
   block
   target
   info)
+(!set-load-form-method nlx-info (:xc :target) :ignore-it)
+
 
 ;;;; LEAF structures
 
@@ -868,6 +849,7 @@
   (extent nil :type (member nil truly-dynamic-extent dynamic-extent indefinite-extent))
   ;; some kind of info used by the back end
   (info nil))
+(!set-load-form-method leaf (:xc :target) :ignore-it)
 
 (defun leaf-dynamic-extent (leaf)
   (let ((extent (leaf-extent leaf)))
@@ -1028,8 +1010,8 @@
   ;;
   ;;    :ASSIGNMENT
   ;;    similar to a LET (as per FUNCTIONAL-SOMEWHAT-LETLIKE-P), but
-  ;;    can have other than one call as long as there is at most
-  ;;    one non-tail call.
+  ;;    can have more than one call as long as the calls all return to
+  ;;    the same place.
   ;;
   ;;    :OPTIONAL
   ;;    a lambda that is an entry point for an OPTIONAL-DISPATCH.
@@ -1103,7 +1085,7 @@
   ;; INLINEP will always be NIL as well.)
   (inline-expansion nil :type list)
   ;; the lexical environment that the INLINE-EXPANSION should be converted in
-  (lexenv *lexenv* :type lexenv :read-only t)
+  (lexenv *lexenv* :type lexenv)
   ;; the original function or macro lambda list, or :UNSPECIFIED if
   ;; this is a compiler created function
   (arg-documentation nil :type (or list (member :unspecified)))
@@ -1236,9 +1218,9 @@
   (tail-set nil :type (or tail-set null))
   ;; the structure which represents the phsical environment that this
   ;; function's variables are allocated in. This is filled in by
-  ;; physical environment analysis. In a LET, this is EQ to our home's
-  ;; physical environment.
-  (physenv nil :type (or physenv null))
+  ;; environment analysis. In a LET, this is EQ to our home's
+  ;; environment.
+  (environment nil :type (or environment null))
   ;; In a LET, this is the NODE-LEXENV of the combination node. We
   ;; retain it so that if the LET is deleted (due to a lack of vars),
   ;; we will still have caller's lexenv to figure out which cleanup is
@@ -1387,26 +1369,26 @@
 ;;; lambda arguments which may ultimately turn out not to be simple
 ;;; and lexical.
 ;;;
-;;; LAMBDA-VARs with no REFs are considered to be deleted; physical
-;;; environment analysis isn't done on these variables, so the back
-;;; end must check for and ignore unreferenced variables. Note that a
-;;; deleted LAMBDA-VAR may have sets; in this case the back end is
-;;; still responsible for propagating the SET-VALUE to the set's CONT.
+;;; LAMBDA-VARs with no REFs are considered to be deleted; environment
+;;; analysis isn't done on these variables, so the back end must check
+;;; for and ignore unreferenced variables. Note that a deleted
+;;; LAMBDA-VAR may have sets; in this case the back end is still
+;;; responsible for propagating the SET-VALUE to the set's CONT.
 (!def-boolean-attribute lambda-var
   ;; true if this variable has been declared IGNORE
   ignore
-  ;; This is set by physical environment analysis if it chooses an
-  ;; indirect (value cell) representation for this variable because it
-  ;; is both set and closed over.
+  ;; This is set by environment analysis if it chooses an indirect
+  ;; (value cell) representation for this variable because it is both
+  ;; set and closed over.
   indirect
   ;; true if the last reference has been deleted (and new references
   ;; should not be made)
   deleted
-  ;; This is set by physical environment analysis if, should it be an
-  ;; indirect lambda-var, an actual value cell object must be
-  ;; allocated for this variable because one or more of the closures
-  ;; that refer to it are not dynamic-extent.  Note that both
-  ;; attributes must be set for the value-cell object to be created.
+  ;; This is set by environment analysis if, should it be an indirect
+  ;; lambda-var, an actual value cell object must be allocated for
+  ;; this variable because one or more of the closures that refer to
+  ;; it are not dynamic-extent.  Note that both attributes must be set
+  ;; for the value-cell object to be created.
   explicit-value-cell
   ;; Do not propagate constraints for this var
   no-constraints
@@ -1658,12 +1640,16 @@
   asserted-type
   type-to-check)
 
-;;; A filter to help order the value semantics of MULTIPLE-VALUE-PROG1
-(defstruct (vestigial-exit-cast (:include cast
-                                          (%type-check nil)
-                                          (asserted-type *wild-type*)
-                                          (type-to-check *wild-type*))
-                                (:copier nil)))
+;;; The DELAY node is interposed between a VALUE's USE and its DEST in
+;;; order to allow the value to be immediately used. This is necessary
+;;; for implementing multiple-use unknown values LVARs, as otherwise,
+;;; a non-moveable dynamic extent object may be allocated between the
+;;; DEST and one of the LVAR's uses but not the others.
+(defstruct (delay (:include cast
+                   (%type-check nil)
+                   (asserted-type *wild-type*)
+                   (type-to-check *wild-type*))
+                  (:copier nil)))
 
 ;;; A cast that always follows %check-bound and they are deleted together.
 ;;; Created via BOUND-CAST ir1-translator by chaining it together with %check-bound.
@@ -1705,8 +1691,8 @@
 ;;; continuation and the exit continuation's DEST. Instead of using
 ;;; the returned value being delivered directly to the exit
 ;;; continuation, it is delivered to our VALUE lvar. The original exit
-;;; lvar is the exit node's LVAR; physenv analysis also makes it the
-;;; lvar of %NLX-ENTRY call.
+;;; lvar is the exit node's LVAR; environment analysis also makes it
+;;; the lvar of %NLX-ENTRY call.
 (defstruct (exit (:include valued-node)
                  (:copier nil))
   ;; the ENTRY node that this is an exit for. If null, this is a
@@ -1722,9 +1708,6 @@
   (entry :test entry)
   (value :test value))
 
-(defstruct (no-op (:include node)
-                  (:copier nil)))
-
 ;;; The ENCLOSE node marks the place at which closure allocation code
 ;;; would be emitted, if necessary.
 (defstruct (enclose (:include valued-node) ; this node uses a dummy lvar for dx analysis
@@ -1752,8 +1735,18 @@
   ;; where this thing was used. Note that we only record the first
   ;; *UNDEFINED-WARNING-LIMIT* calls.
   (warnings () :type list))
+
 (declaim (freeze-type undefined-warning))
 
+(defstruct (argument-mismatch-warning
+            (:print-object (lambda (x s)
+                             (print-unreadable-object (x s :type t)
+                               (prin1 (argument-mismatch-warning-name x) s))))
+            (:copier nil))
+  (name nil :type (or symbol list))
+  ;; a list of (KEYS . COMPILER-ERROR-CONTEXT)
+  (warnings () :type list))
+
 
 ;;; a helper for the POLICY macro, defined late here so that the
 ;;; various type tests can be inlined
@@ -1783,4 +1776,4 @@
 ;;;; Freeze some structure types to speed type testing.
 
 (declaim (freeze-type node lexenv ctran lvar cblock component cleanup
-                      physenv tail-set nlx-info leaf))
+                      environment tail-set nlx-info leaf))
