@@ -1170,14 +1170,12 @@ default-value-8
       (move result null-tn)
       (inst beq done)
 
-    ;; We need to do this atomically.
-    (pseudo-atomic (pa-flag :sync nil)
+    (pseudo-atomic (pa-flag :sync nil :elide-if dx-p)
       ;; Allocate a cons (2 words) for each item.
       (if dx-p
           (progn
             (align-csp temp)
-            (inst clrrdi result csp-tn n-lowtag-bits)
-            (inst ori result result list-pointer-lowtag)
+            (inst ori result csp-tn list-pointer-lowtag)
             (move dst result)
             (inst sldi temp count (1+ (- word-shift n-fixnum-tag-bits)))
             (inst add csp-tn csp-tn temp))
