@@ -46,7 +46,6 @@
   (:temporary (:scs (descriptor-reg) :from (:argument 0)) obj-temp))
 
 ;;; With Symbol-Value, we check that the value isn't the trap object.
-;;; So Symbol-Value of NIL is NIL.
 (define-vop (symbol-value checked-cell-ref)
   (:translate symeval)
   (:generator 9
@@ -290,18 +289,13 @@
          (index :scs (any-reg))
          (value :scs (any-reg descriptor-reg)))
   (:arg-types * tagged-num *)
-  (:temporary (:scs (non-descriptor-reg)) temp #+gencgc card)
+  (:temporary (:scs (non-descriptor-reg)) temp card)
   (:generator 10
-    #+cheneygc
-    (progn (inst add temp index (- other-pointer-lowtag))
-           (inst st value object temp))
-    #+gencgc
-    (progn
     ;; Load the card mask
     (inst li temp (make-fixup "gc_card_table_mask" :foreign-dataref)) ; linkage entry
     (inst ld temp temp) ; address of gc_card_table_mask
     (inst ld temp temp) ; value of gc_card_table_mask
-    (pseudo-atomic ()
+    (pseudo-atomic (temp)
       ;; Compute card mark index
       (inst srl card object gencgc-card-shift)
       (inst and card card temp)
@@ -310,7 +304,7 @@
       (inst ld temp temp) ; address of gc_card_mark
       (inst ld temp temp) ; value of gc_card_mark
       ;; Touch the card mark byte.
-      (inst stb zero-tn temp card)
+      (inst stb null-tn temp card)
       ;; set 'written' flag in the code header
       ;; If two threads get here at the same time, they'll write the same byte.
       (let ((byte #+big-endian (- other-pointer-lowtag) #+little-endian (bug "Wat")))
@@ -318,7 +312,7 @@
         (inst or temp temp #x40)
         (inst stb temp object byte))
       (inst add temp index (- other-pointer-lowtag))
-      (inst st value object temp)))))
+      (inst st value object temp))))
 
 ;;;; raw instance slot accessors
 
