@@ -38,7 +38,7 @@
            has already been partially loaded. This may not work, you may~%~
            need to get a fresh lisp (reboot) and then load PCL."))
 
-#-sb-fluid (declaim (inline gdefinition))
+(declaim (inline gdefinition))
 (defun gdefinition (spec)
   ;; This is null layer right now, but once FDEFINITION stops bypasssing
   ;; fwrappers/encapsulations we can do that here.
@@ -122,7 +122,7 @@
     ((not and or) `(,(car type) ,@(mapcar #'convert-to-system-type
                                           (cdr type))))
     ((class class-eq) ; class-eq is impossible to do right
-     (layout-classoid (class-wrapper (cadr type))))
+     (wrapper-classoid (class-wrapper (cadr type))))
     (eql type)
     (t (if (null (cdr type))
            (car type)
@@ -177,50 +177,52 @@
 
 ;;;; built-in classes
 
-;;; Grovel over SB-KERNEL::+!BUILT-IN-CLASSES+ in order to set
+;;; Grovel over SB-KERNEL::*BUILTIN-CLASSOIDS* in order to set
 ;;; SB-PCL:*BUILT-IN-CLASSES*.
 (/show "about to set up SB-PCL::*BUILT-IN-CLASSES*")
 (define-load-time-global *built-in-classes*
-  (labels ((direct-supers (class)
-             (/noshow "entering DIRECT-SUPERS" (classoid-name class))
-             (if (typep class 'built-in-classoid)
-                 (built-in-classoid-direct-superclasses class)
-                 (let ((inherits (layout-inherits
-                                  (classoid-layout class))))
-                   (/noshow inherits)
-                   (list (svref inherits (1- (length inherits)))))))
-           (direct-subs (class)
-             (/noshow "entering DIRECT-SUBS" (classoid-name class))
-             (collect ((res))
-               (let ((subs (classoid-subclasses class)))
-                 (/noshow subs)
-                 (when subs
-                   (dohash ((sub v) subs)
-                     (declare (ignore v))
-                     (/noshow sub)
-                     (when (member class (direct-supers sub) :test #'eq)
-                       (res sub)))))
-               (res))))
-    (mapcar (lambda (kernel-bic-entry)
-              (/noshow "setting up" kernel-bic-entry)
-              (let* ((name (car kernel-bic-entry))
-                     (class (find-classoid name)))
-                (/noshow name class)
-                `(,name
-                  ,(mapcar #'classoid-name (direct-supers class))
-                  ,(mapcar #'classoid-name (direct-subs class))
-                  ,(map 'list
-                        (lambda (x) (classoid-name (layout-classoid x)))
-                        (reverse (layout-inherits (classoid-layout class))))
-                  ,(eval (getf (cdr kernel-bic-entry) :prototype-form)))))
-            (remove-if (lambda (kernel-bic-entry)
-                         (member (first kernel-bic-entry)
-                                 ;; remove special classes (T and our
-                                 ;; SYSTEM-CLASSes) from the
-                                 ;; BUILT-IN-CLASS list
-                                 '(t function stream sequence
-                                     file-stream string-stream)))
-                       sb-kernel::+!built-in-classes+))))
+    (macrolet
+        ((frob ()
+           (labels ((direct-supers (class)
+                      (/noshow "entering DIRECT-SUPERS" (classoid-name class))
+                      (if (typep class 'built-in-classoid)
+                          (built-in-classoid-direct-superclasses class)
+                          (let ((inherits (wrapper-inherits (classoid-wrapper class))))
+                            (/noshow inherits)
+                            (list (svref inherits (1- (length inherits)))))))
+                    (direct-subs (class)
+                      (/noshow "entering DIRECT-SUBS" (classoid-name class))
+                      (collect ((res))
+                        (let ((subs (classoid-subclasses class)))
+                          (/noshow subs)
+                          (when subs
+                            (sb-kernel::do-subclassoids ((sub wrapper) class)
+                              (declare (ignore wrapper))
+                              (/noshow sub)
+                              (when (member class (direct-supers sub) :test #'eq)
+                                (res sub)))))
+                        (sort (res) #'string< :key #'classoid-name))))
+             `(list ,@(mapcar (lambda (kernel-bic-entry)
+                                (/noshow "setting up" kernel-bic-entry)
+                                (let* ((name (car kernel-bic-entry))
+                                       (class (find-classoid name)))
+                                  (/noshow name class)
+                                  `(list ',name
+                                         ',(mapcar #'classoid-name (direct-supers class))
+                                         ',(mapcar #'classoid-name (direct-subs class))
+                                         ',(map 'list #'wrapper-classoid-name
+                                                (reverse (wrapper-inherits (classoid-wrapper class))))
+                                         ,(getf (cdr kernel-bic-entry) :prototype-form))))
+                              (remove-if (lambda (kernel-bic-entry)
+                                           (member (first kernel-bic-entry)
+                                                   ;; remove special classes (T and our
+                                                   ;; SYSTEM-CLASSes) from the
+                                                   ;; BUILT-IN-CLASS list
+                                                   '(t function stream sequence
+                                                     file-stream string-stream
+                                                     slot-object)))
+                                         sb-kernel::*builtin-classoids*))))))
+      (frob)))
 (/noshow "done setting up SB-PCL::*BUILT-IN-CLASSES*")
 
 ;;;; the classes that define the kernel of the metabraid
@@ -330,6 +332,7 @@
    (lambda-list :initform () :initarg :lambda-list :reader method-lambda-list)
    (%function :initform nil :initarg :function :reader method-function)
    (%documentation :initform nil :initarg :documentation)
+   (%cache :initform nil :accessor method-em-cache)
    ;; True IFF method is known to have no CALL-NEXT-METHOD in it, or
    ;; just a plain (CALL-NEXT-METHOD).
    (simple-next-method-call
@@ -458,6 +461,11 @@
    (info
     :accessor slot-definition-info)))
 
+(defun uninitialized-accessor-function (type slotd)
+  (lambda (&rest args)
+    (declare (ignore args))
+    (error "~:(~A~) function~@[ for ~S ~] not yet initialized."
+           type slotd)))
 ;;; We use a structure here, because fast slot-accesses to this information
 ;;; are critical to making SLOT-VALUE-USING-CLASS &co fast: places that need
 ;;; these functions can access the SLOT-INFO directly, avoiding the overhead
@@ -568,14 +576,12 @@
 ;; gethash on a live key should get the identical specializer, but since
 ;; nothing referenced the old specializer, consing a new one is fine.
 (defglobal *eql-specializer-table*
-  (make-hash-table :test 'eql :weakness :value))
+  (sb-impl::make-system-hash-table :test 'eql :weakness :value :synchronized nil))
 
 (defun intern-eql-specializer (object)
   ;; Avoid style-warning about compiler-macro being unavailable.
   (declare (notinline make-instance))
-  ;; Need to lock, so that two threads don't get non-EQ specializers
-  ;; for an EQL object.
-  (with-locked-system-table (*eql-specializer-table*)
+  (with-system-mutex ((hash-table-lock *eql-specializer-table*))
     (ensure-gethash object *eql-specializer-table*
                     (make-instance 'eql-specializer :object object))))
 
