@@ -40,7 +40,6 @@
 #include "validate.h"
 #include "gc-internal.h"
 #include "gc-private.h"
-#include "getallocptr.h"
 #include "code.h"
 
 #include <errno.h>
@@ -588,8 +587,8 @@ static void relocate_space(uword_t start, lispobj* end, struct heap_adjust* adj)
         adjust_pointers(where+1, nwords-1, adj);
     }
 #if SHOW_SPACE_RELOCATION
-    fprintf(stderr, "space @ %p: fixed %d absolute + %d relative pointers\n",
-            (lispobj*)start, adj->n_relocs_abs, adj->n_relocs_rel);
+    fprintf(stderr, "space @ %p..%p: fixed %d absolute + %d relative pointers\n",
+            (lispobj*)start, end, adj->n_relocs_abs, adj->n_relocs_rel);
 #endif
 }
 
@@ -613,7 +612,8 @@ static void relocate_heap(struct heap_adjust* adj)
 #ifdef LISP_FEATURE_CHENEYGC
     relocate_space(DYNAMIC_0_SPACE_START, (lispobj*)get_alloc_pointer(), adj);
 #else
-    relocate_space(DYNAMIC_SPACE_START, (lispobj*)get_alloc_pointer(), adj);
+    relocate_space(DYNAMIC_SPACE_START, (lispobj*)dynamic_space_highwatermark(),
+                   adj);
 #endif
 #ifdef LISP_FEATURE_IMMOBILE_SPACE
     // Pointers within varyobj space to varyobj space do not need adjustment
@@ -980,9 +980,10 @@ process_directory(int count, struct ndir_entry *entry,
              * at the time the core was saved.
              * For gencgc this is #defined as DYNAMIC_SPACE_START */
             current_dynamic_space = (lispobj *)addr;
+#else
+            next_free_page = ALIGN_UP(entry->nwords<<WORD_SHIFT, GENCGC_PAGE_BYTES)
+              / GENCGC_PAGE_BYTES;
 #endif
-            set_alloc_pointer((lispobj)free_pointer);
-
             anon_dynamic_space_start = (os_vm_address_t)(addr + len);
         }
     }
