@@ -52,6 +52,15 @@
 #define BIGNUM_MARK_BIT MARK_BIT
 #endif
 
+static inline uword_t boxedobj_widetag_to_markbit(int widetag) {
+    return (widetag == FDEFN_WIDETAG || widetag == WEAK_POINTER_WIDETAG)
+         ? FDEFN_MARK_BIT : MARK_BIT;
+}
+static inline uword_t general_widetag_to_markbit(int widetag) {
+    return (widetag == FDEFN_WIDETAG || widetag == WEAK_POINTER_WIDETAG)
+        ? FDEFN_MARK_BIT : ((widetag == BIGNUM_WIDETAG) ? BIGNUM_MARK_BIT : MARK_BIT);
+}
+
 #define interesting_pointer_p(x) \
   (find_page_index((void*)x) >= 0 || immobile_space_p(x))
 
@@ -75,13 +84,13 @@ static page_index_t free_page;
 
 /* The whole-page allocator works backwards from the end of dynamic space.
  * If it collides with 'next_free_page', then you lose.
- * TOOD: It would be reasonably simple to have this request more memory from
+ * TODO: It would be reasonably simple to have this request more memory from
  * the OS instead of failing on overflow */
 static void* get_free_page() {
     --free_page;
     if (free_page < next_free_page)
         lose("Needed more space to GC");
-    page_table[free_page].type = UNBOXED_PAGE_FLAG;
+    page_table[free_page].type = PAGE_TYPE_UNBOXED;
     char* mem = page_address(free_page);
     zero_dirty_pages(free_page, free_page, 0);
     return mem;
@@ -92,18 +101,23 @@ static void* get_free_page() {
  * then one GC page can hold 2K cons cells.
  * One byte marks 8 conses (1 bit per cons), 256 bytes mark 2048 conses.
  * 128 blocks of 256 bytes fit on a 32K GC page. */
-char *suballocator_free_ptr, *suballocator_end_ptr;
+static char *suballocator_free_ptr, *suballocator_end_ptr;
 
+#ifndef LISP_FEATURE_USE_CONS_REGION
+static inline uword_t compute_page_key(lispobj cons) {
+    return ALIGN_DOWN(cons, GENCGC_PAGE_BYTES);
+}
 static void* allocate_cons_mark_bits() {
-    int nbytes = GENCGC_CARD_BYTES / (2 * N_WORD_BYTES) / 8;
+    int nbytes = GENCGC_PAGE_BYTES / (2 * N_WORD_BYTES) / 8;
     if (suballocator_free_ptr + nbytes > suballocator_end_ptr) {
         suballocator_free_ptr = get_free_page();
-        suballocator_end_ptr = suballocator_free_ptr + GENCGC_CARD_BYTES;
+        suballocator_end_ptr = suballocator_free_ptr + GENCGC_PAGE_BYTES;
     }
     void* mem = suballocator_free_ptr;
     suballocator_free_ptr += nbytes;
     return mem;
 }
+#endif
 
 static void gc_enqueue(lispobj object)
 {
@@ -152,18 +166,20 @@ static lispobj gc_dequeue()
  * for headerless objects (conses) */
 struct hopscotch_table mark_bits;
 
-static inline uword_t compute_page_key(lispobj cons) {
-    return ALIGN_DOWN(cons, GENCGC_CARD_BYTES);
-}
 static inline int compute_dword_number(lispobj cons) {
-    return (cons & (GENCGC_CARD_BYTES - 1)) >> (1+WORD_SHIFT);
+    return (cons & (GENCGC_PAGE_BYTES - 1)) >> (1+WORD_SHIFT);
 }
 
 static inline int cons_markedp(lispobj pointer) {
+    unsigned int index = compute_dword_number(pointer);
+#ifdef LISP_FEATURE_USE_CONS_REGION /* all conses must be on cons pages */
+    unsigned char* bits = (unsigned char*)ALIGN_DOWN(pointer, GENCGC_PAGE_BYTES)
+                          + CONS_PAGE_USABLE_BYTES;
+#else
     unsigned char* bits = (unsigned char*)
         hopscotch_get(&mark_bits, compute_page_key(pointer), 0);
     if (!bits) return 0;
-    int index = compute_dword_number(pointer);
+#endif
     return (bits[index / 8] >> (index % 8)) & 1;
 }
 
@@ -178,7 +194,8 @@ static inline int pointer_survived_gc_yet(lispobj pointer)
     int widetag = header_widetag(header);
     switch (widetag) {
     case BIGNUM_WIDETAG: return (header & BIGNUM_MARK_BIT) != 0;
-    case FDEFN_WIDETAG : return (header & FDEFN_MARK_BIT) != 0;
+    case WEAK_POINTER_WIDETAG:
+    case FDEFN_WIDETAG: return (header & FDEFN_MARK_BIT) != 0;
     }
     if (embedded_obj_p(widetag))
         header = *fun_code_header(native_pointer(pointer));
@@ -203,21 +220,39 @@ void __mark_obj(lispobj pointer)
                 pointer = make_lispobj(base, OTHER_POINTER_LOWTAG);
                 header = *base;
             }
-            uword_t markbit = (widetag == FDEFN_WIDETAG) ? FDEFN_MARK_BIT : MARK_BIT;
+            uword_t markbit = boxedobj_widetag_to_markbit(widetag);
             if (header & markbit) return; // already marked
             *base |= markbit;
         }
+#ifdef LISP_FEATURE_UBSAN
+        if (specialized_vector_widetag_p(widetag) && is_lisp_pointer(base[1]))
+            gc_mark_obj(base[1]);
+        else if (widetag == SIMPLE_VECTOR_WIDETAG && fixnump(base[1])) {
+            char *origin_pc = (char*)(base[1]>>4);
+            lispobj* code = component_ptr_from_pc(origin_pc);
+            if (code) gc_mark_obj(make_lispobj(code, OTHER_POINTER_LOWTAG));
+            /* else lose("can't find code containing %p (vector=%p)", origin_pc, base); */
+        }
+#endif
         if (leaf_obj_widetag_p(widetag)) return;
     } else {
+        unsigned int index = compute_dword_number(pointer);
+#ifdef LISP_FEATURE_USE_CONS_REGION /* all conses must be on cons pages */
+        unsigned char* bits = (unsigned char*)ALIGN_DOWN(pointer, GENCGC_PAGE_BYTES)
+                              + CONS_PAGE_USABLE_BYTES;
+#else
+        /* Only _some_ conses are on cons pages depending if they got moved there.
+         * We could use a hybrid of storing mark bits on the page if possible,
+         * or the hash-table if not. However, I'd rather not complicate things.
+         * All architectures should be made to allocate directly on cons pages! */
         uword_t key = compute_page_key(pointer);
-        int index = compute_dword_number(pointer);
         unsigned char* bits = (unsigned char*)hopscotch_get(&mark_bits, key, 0);
         if (!bits) {
             bits = allocate_cons_mark_bits();
             hopscotch_insert(&mark_bits, key, (sword_t)bits);
-        } else if (bits[index / 8] & (1 << (index % 8))) {
-            return;
         }
+#endif
+        if (bits[index / 8] & (1 << (index % 8))) return;
         // Mark the cons
         bits[index / 8] |= 1 << (index % 8);
     }
@@ -249,7 +284,11 @@ static void trace_using_layout(lispobj layout, lispobj* where, int nslots)
     // Apart from the allowance for untagged pointers in lockfree list nodes,
     // this contains almost none of the special cases that gencgc does.
     if (!layout) return;
+#ifdef LISP_FEATURE_METASPACE
+    gc_mark_obj(LAYOUT(layout)->friend);
+#else
     gc_mark_obj(layout);
+#endif
     if (lockfree_list_node_layout_p(LAYOUT(layout))) { // allow untagged 'next'
         struct instance* node = (struct instance*)where;
         lispobj next = node->slots[INSTANCE_DATA_START];
@@ -269,26 +308,30 @@ static void trace_object(lispobj* where)
     lispobj header = *where;
     int widetag = header_widetag(header);
 
-    switch (widetag) {
-    case INSTANCE_WIDETAG:
-        return trace_using_layout(instance_layout(where),
-                                  where, instance_length(header));
-    case FUNCALLABLE_INSTANCE_WIDETAG:
-        return trace_using_layout(funinstance_layout(where),
-                                  where, HeaderValue(header) & SHORT_HEADER_MAX_WORDS);
-    }
+    if (instanceoid_widetag_p(widetag))
+        return trace_using_layout(layout_of(where), where,
+                                  instanceoid_length(header));
     sword_t scan_from = 1;
     sword_t scan_to = sizetab[widetag](where);
     sword_t i;
     struct weak_pointer *weakptr;
     switch (widetag) {
     case SIMPLE_VECTOR_WIDETAG:
+#ifdef LISP_FEATURE_UBSAN
+        if (is_lisp_pointer(where[1])) gc_mark_obj(where[1]);
+#endif
         // non-weak hashtable kv vectors are trivial in fullcgc. Keys don't move
         // so the table will not need rehash as a result of gc.
-        if ((vector_subtype(header) & ~subtype_VectorAddrHashing)
-            == subtype_VectorHashing + subtype_VectorWeak) { // weak table
+        // Ergo, those may be treated just like ordinary simple vectors.
+        // However, weakness remains as a special case.
+        if (vector_flagp(header, VectorWeak)) {
+            if (!vector_flagp(header, VectorHashing)) {
+                add_to_weak_vector_list(where, header);
+                return;
+            }
+            // Ok, we're looking at a weak hash-table.
             struct vector* v = (struct vector*)where;
-            lispobj lhash_table = v->data[fixnum_value(v->length)-1];
+            lispobj lhash_table = v->data[vector_len(v)-1];
             gc_dcheck(instancep(lhash_table));
             __mark_obj(lhash_table);
             struct hash_table* hash_table
@@ -308,12 +351,8 @@ static void trace_object(lispobj* where)
             }
             return;
         }
-        if (is_vector_subtype(header, VectorWeak)) {
-            add_to_weak_vector_list(where, header);
-            return;
-        }
         break;
-#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
+#if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64) || defined (LISP_FEATURE_ARM64)
     /* on x86[-64], closure->fun is a fixnum-qua-pointer. Convert it to a lisp
      * pointer to mark it, but not on platforms where it's already a descriptor */
     case CLOSURE_WIDETAG:
@@ -325,7 +364,8 @@ static void trace_object(lispobj* where)
         scan_to = code_header_words((struct code*)where);
 #ifdef LISP_FEATURE_UNTAGGED_FDEFNS
         struct code* code = (struct code*)where;
-        lispobj* fdefns_start = code->constants + code_n_funs(code) * 4;
+        lispobj* fdefns_start = code->constants
+                                + code_n_funs(code) * CODE_SLOTS_PER_SIMPLE_FUN;
         lispobj* fdefns_end  = fdefns_start + code_n_named_calls(code);
         lispobj* limit = where + scan_to;
         where = where + scan_from;
@@ -338,6 +378,19 @@ static void trace_object(lispobj* where)
         return;
 #endif
         break;
+#ifdef LISP_FEATURE_COMPACT_SYMBOL
+    case SYMBOL_WIDETAG:
+        {
+        struct symbol* s = (void*)where;
+        gc_mark_obj(decode_symbol_name(s->name));
+        gc_mark_obj(s->value);
+        gc_mark_obj(s->info);
+        gc_mark_obj(s->fdefn);
+        // process the unnamed slot of augmented symbols
+        if ((s->header & 0xFF00) == (SYMBOL_SIZE<<8)) gc_mark_obj(*(1+&s->fdefn));
+        }
+        return;
+#endif
     case FDEFN_WIDETAG:
         gc_mark_obj(fdefn_callee_lispobj((struct fdefn*)where));
         scan_to = 3;
@@ -356,15 +409,12 @@ static void trace_object(lispobj* where)
 
 void prepare_for_full_mark_phase()
 {
-    // FIXME: Estimate how large to create mark_bits based on dynamic space size.
-    // Guess 8 words per object, and X% of the objects are conses.
-    // The problem is guessing how localized the conses are: guess that N conses
-    // will reside on fraction*N different pages, which guides us as to how many
-    // hash table entries are needed.
+#ifndef LISP_FEATURE_USE_CONS_REGION
     hopscotch_create(&mark_bits, HOPSCOTCH_HASH_FUN_DEFAULT,
                      N_WORD_BYTES, /* table values are machine words */
                      65536, /* initial size */
                      0);
+#endif
 
     free_page = page_table_pages;
     suballocator_free_ptr = suballocator_end_ptr = 0;
@@ -382,6 +432,7 @@ void execute_full_mark_phase()
     struct rusage before, after;
     getrusage(RUSAGE_SELF, &before);
 #endif
+    trace_object((lispobj*)NIL_SYMBOL_SLOTS_START);
     lispobj* where = (lispobj*)STATIC_SPACE_OBJECTS_START;
     lispobj* end = static_space_free_pointer;
     while (where < end) {
@@ -389,6 +440,16 @@ void execute_full_mark_phase()
         gc_enqueue(obj);
         where += listp(obj) ? 2 : sizetab[widetag_of(where)](where);
     }
+#ifdef LISP_FEATURE_METASPACE
+    where = (lispobj*)METASPACE_START;
+    end = (lispobj*)READ_ONLY_SPACE_END;
+    while (where < end) {
+        lispobj obj = compute_lispobj(where);
+        gc_enqueue(obj);
+        where += listp(obj) ? 2 : sizetab[widetag_of(where)](where);
+    }
+#endif
+    gc_mark_obj(lisp_package_vector);
     do {
         lispobj ptr = gc_dequeue();
         gc_dcheck(ptr != 0);
@@ -418,8 +479,8 @@ static void local_smash_weak_pointers()
     struct weak_pointer *wp, *next_wp;
     for (wp = weak_pointer_chain; wp != WEAK_POINTER_CHAIN_END; wp = next_wp) {
         gc_assert(widetag_of(&wp->header) == WEAK_POINTER_WIDETAG);
-        next_wp = wp->next;
-        wp->next = NULL;
+        next_wp = get_weak_pointer_next(wp);
+        reset_weak_pointer_next(wp);
         lispobj pointee = wp->value;
         gc_assert(is_lisp_pointer(pointee));
         if (!pointer_survived_gc_yet(pointee))
@@ -432,7 +493,7 @@ static void local_smash_weak_pointers()
         struct vector* vector = (struct vector*)vectors->car;
         vectors = (struct cons*)vectors->cdr;
         UNSET_WEAK_VECTOR_VISITED(vector);
-        sword_t len = fixnum_value(vector->length);
+        sword_t len = vector_len(vector);
         sword_t i;
         for (i = 0; i<len; ++i) {
             lispobj obj = vector->data[i];
@@ -484,7 +545,7 @@ static void sweep_fixedobj_pages(long *zeroed)
         lispobj *limit = (lispobj*)((char*)obj + IMMOBILE_CARD_BYTES - obj_spacing);
         for ( ; obj <= limit ; obj = (lispobj*)((char*)obj + obj_spacing) ) {
             lispobj header = *obj;
-            uword_t markbit = (header_widetag(header) == FDEFN_WIDETAG) ? FDEFN_MARK_BIT : MARK_BIT;
+            uword_t markbit = boxedobj_widetag_to_markbit(header_widetag(header));
             if (fixnump(header)) { // is a hole
             } else if (header & markbit) { // live object
                 *obj = header ^ markbit;
@@ -494,6 +555,15 @@ static void sweep_fixedobj_pages(long *zeroed)
             }
         }
     }
+    // The reserved fixedobj page has the vector of primitive object layouts.
+    lispobj* obj = fixedobj_page_address(0);
+    lispobj* limit = fixedobj_page_address(FIXEDOBJ_RESERVED_PAGES);
+    while (obj < limit) {
+        lispobj header = *obj;
+        uword_t markbit = boxedobj_widetag_to_markbit(header_widetag(header));
+        if (header & markbit) *obj = header ^ markbit;
+        obj += sizetab[widetag_of(obj)](obj);
+    }
 }
 #endif
 
@@ -507,11 +577,7 @@ static uword_t sweep(lispobj* where, lispobj* end, uword_t arg)
         lispobj word = *where;
         if (is_header(word)) {
             nwords = sizetab[header_widetag(word)](where);
-            lispobj markbit = MARK_BIT;
-            switch (header_widetag(word)) {
-            case BIGNUM_WIDETAG: markbit = BIGNUM_MARK_BIT; break;
-            case FDEFN_WIDETAG : markbit = FDEFN_MARK_BIT; break;
-            }
+            lispobj markbit = general_widetag_to_markbit(header_widetag(word));
             if (word & markbit)
                 *where = word ^ markbit;
             else {
@@ -529,6 +595,9 @@ static uword_t sweep(lispobj* where, lispobj* end, uword_t arg)
                         code->boxed_size = 0;
                         code->header = (nwords << CODE_HEADER_SIZE_SHIFT)
                                      | CODE_HEADER_WIDETAG;
+                        /* This memset() is needed for strictly boxed pages,
+                         * but probably not for other page types. No big deal.
+                         * fullcgc isn't really used much */
                         memset(where+2, 0, (nwords - 2) * N_WORD_BYTES);
                     })
                 }
@@ -592,23 +661,23 @@ void execute_full_sweep_phase()
             fprintf(stderr, "%ld%s", words_zeroed[i], i?"+":"");
         fprintf(stderr, " words zeroed]\n");
     }
+#ifdef LISP_FEATURE_USE_CONS_REGION
+    page_index_t page;
+    for (page = 0; page < next_free_page; ++page)
+        if (page_table[page].type == PAGE_TYPE_CONS) {
+            char* base = page_address(page);
+            char* bits = base + CONS_PAGE_USABLE_BYTES;
+            char* end  = base + GENCGC_PAGE_BYTES;
+            memset(bits, 0, end-bits);
+        }
+#else
     hopscotch_destroy(&mark_bits);
+#endif
+
     if (sweeplog)
         fflush(sweeplog);
 
-    page_index_t first_page, last_page;
-    for (first_page = 0; first_page < next_free_page; ++first_page)
-        if (page_table[first_page].write_protected
-            && protection_mode(first_page) == PHYSICAL) {
-            last_page = first_page;
-            while (page_table[last_page+1].write_protected
-                   && protection_mode(last_page+1) == PHYSICAL)
-                ++last_page;
-            os_protect(page_address(first_page),
-                       (last_page - first_page + 1) * GENCGC_CARD_BYTES,
-                       OS_VM_PROT_READ | OS_VM_PROT_EXECUTE);
-            first_page = last_page;
-        }
+    free_page = next_free_page;
     while (free_page < page_table_pages) {
         page_table[free_page++].type = FREE_PAGE_FLAG;
     }
