@@ -63,7 +63,7 @@
 #include "gc.h"
 #include "alloc.h"
 #include "dynbind.h"
-#include "getallocptr.h"
+#include "pseudo-atomic.h"
 #include "genesis/fdefn.h"
 #include "genesis/simple-fun.h"
 #include "genesis/cons.h"
@@ -524,8 +524,12 @@ check_deferrables_blocked_or_lose(sigset_t *sigset)
         lose("deferrables unblocked");
 }
 
+int sigaction_does_not_mask;
 static void assert_blockables_blocked()
 {
+#ifdef LISP_FEATURE_RISCV
+    if (sigaction_does_not_mask) return; // assert nothing
+#endif
 #if !defined(LISP_FEATURE_WIN32)
     /* On Windows, there are no actual signals, but since the win32 port
      * tracks the sigmask and checks it explicitly, some functions are
@@ -847,7 +851,11 @@ build_fake_control_stack_frames(struct thread __attribute__((unused)) *th,
             access_control_frame_pointer(th)[0] =
                 *os_context_register_addr(context, reg_OCFP);
             access_control_frame_pointer(th)[1] =
-                *os_context_register_addr(context, reg_LRA);
+#ifdef reg_LRA
+              *os_context_register_addr(context, reg_LRA);
+#else
+              *os_context_register_addr(context, reg_RA);
+#endif
             access_control_frame_pointer(th) += 2;
             /* Build our frame on top of it. */
             oldcont = (lispobj)(*os_context_register_addr(context, reg_CFP));
@@ -887,32 +895,17 @@ build_fake_control_stack_frames(struct thread __attribute__((unused)) *th,
 }
 #endif
 
-/* Stores the context for gc to scavange and builds fake stack
+/* Stores the context for gc to scavenge and builds fake stack
  * frames. */
 void fake_foreign_function_call_noassert(os_context_t *context)
 {
     int context_index;
     struct thread *thread=get_sb_vm_thread();
 
-#ifdef reg_ALLOC
-#ifdef LISP_FEATURE_SB_THREAD
-    thread->pseudo_atomic_bits =
-#else
-    dynamic_space_free_pointer =
-        (lispobj *)(uword_t)
-#endif
-            (*os_context_register_addr(context, reg_ALLOC));
-/*     fprintf(stderr,"dynamic_space_free_pointer: %p\n", */
-/*             dynamic_space_free_pointer); */
-#if defined(LISP_FEATURE_MIPS)
-    if ((sword_t)dynamic_space_free_pointer & 1) {
-        lose("dead in fake_foreign_function_call, context = %x", context);
-    }
-#endif
-#endif
 #ifdef reg_BSP
     set_binding_stack_pointer(thread,
-        *os_context_register_addr(context, reg_BSP));
+       // registers can be wider than uword_t (on some 64-bit machines compiling to 32-bit code)
+       (uword_t)*os_context_register_addr(context, reg_BSP));
 #endif
 
 #if defined(LISP_FEATURE_ARM)
@@ -983,30 +976,6 @@ undo_fake_foreign_function_call(os_context_t __attribute__((unused)) *context)
                    thread);
     unbind(thread);
 #endif
-
-#if defined(reg_ALLOC) && !defined(LISP_FEATURE_SB_THREAD)
-    /* Put the dynamic space free pointer back into the context. */
-    *os_context_register_addr(context, reg_ALLOC) =
-        (uword_t) dynamic_space_free_pointer
-        | (*os_context_register_addr(context, reg_ALLOC)
-           & LOWTAG_MASK);
-    /*
-      ((uword_t)(*os_context_register_addr(context, reg_ALLOC))
-      & ~LOWTAG_MASK)
-      | ((uword_t) dynamic_space_free_pointer & LOWTAG_MASK);
-    */
-#endif
-#if defined(reg_ALLOC) && defined(LISP_FEATURE_SB_THREAD)
-    /* Put the pseudo-atomic bits and dynamic space free pointer back
-     * into the context (p-a-bits for p-a, and dynamic space free
-     * pointer for ROOM). */
-    *os_context_register_addr(context, reg_ALLOC) =
-        (uword_t) dynamic_space_free_pointer
-        | (thread->pseudo_atomic_bits & LOWTAG_MASK);
-    /* And clear them so we don't get bit later by call-in/call-out
-     * not updating them. */
-    thread->pseudo_atomic_bits = 0;
-#endif
 }
 
 /* a handler for the signal caused by execution of a trap opcode
@@ -1555,7 +1524,7 @@ interrupt_handle_now_handler(int signal, siginfo_t *info, void *void_context)
  */
 
 #if (defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64))
-extern int *context_eflags_addr(os_context_t *context);
+extern int *os_context_flags_addr(os_context_t *context);
 #endif
 
 extern lispobj call_into_lisp(lispobj fun, lispobj *args, int nargs);
@@ -1644,7 +1613,7 @@ arrange_return_to_c_function(os_context_t *context,
     *(register_save_area + 5) = *os_context_register_addr(context,reg_ECX);
     *(register_save_area + 6) = *os_context_register_addr(context,reg_EBX);
     *(register_save_area + 7) = *os_context_register_addr(context,reg_EAX);
-    *(register_save_area + 8) = *context_eflags_addr(context);
+    *(register_save_area + 8) = *os_context_flags_addr(context);
 
     set_os_context_pc(context, (os_context_register_t) funptr);
     *os_context_register_addr(context,reg_ECX) =
@@ -1668,7 +1637,7 @@ arrange_return_to_c_function(os_context_t *context,
     *(sp-6)=*os_context_register_addr(context,reg_EDX);
     *(sp-5)=*os_context_register_addr(context,reg_ECX);
     *(sp-4)=*os_context_register_addr(context,reg_EAX);
-    *(sp-3)=*context_eflags_addr(context);
+    *(sp-3)=*os_context_flags_addr(context);
     *(sp-2)=*os_context_register_addr(context,reg_EBP);
     *(sp-1)=os_context_pc(context);
 
@@ -1695,7 +1664,7 @@ arrange_return_to_c_function(os_context_t *context,
     *(sp-6)=*os_context_register_addr(context,reg_RDX);
     *(sp-5)=*os_context_register_addr(context,reg_RCX);
     *(sp-4)=*os_context_register_addr(context,reg_RAX);
-    *(sp-3)=*context_eflags_addr(context);
+    *(sp-3)=*os_context_flags_addr(context);
     *(sp-2)=*os_context_register_addr(context,reg_RBP);
     *(sp-1)=os_context_pc(context);
 
