@@ -35,8 +35,6 @@
 #include "genesis/static-symbols.h"
 #include "genesis/fdefn.h"
 
-#include <sys/socket.h>
-#include <sys/utsname.h>
 #include <errno.h>
 
 #include <sys/types.h>
@@ -51,11 +49,7 @@
 #include "thread.h"
 #include "gc-internal.h"
 #include <fcntl.h>
-#ifdef LISP_FEATURE_SB_WTIMER
-# include <sys/timerfd.h>
-#endif
-
-int sb_GetTID() { return syscall(SYS_gettid); }
+#include <sys/prctl.h>
 
 #ifdef LISP_FEATURE_X86
 /* Prototype for personality(2). Done inline here since the header file
@@ -66,11 +60,98 @@ int personality (unsigned long);
 #include <sys/personality.h>
 #endif
 
-#if defined(LISP_FEATURE_SB_THREAD) && defined(LISP_FEATURE_SB_FUTEX) && !defined(LISP_FEATURE_SB_PTHREAD_FUTEX)
+#ifdef LISP_FEATURE_SB_FUTEX
 #include <sys/syscall.h>
 #include <unistd.h>
 #include <errno.h>
 
+#ifdef MUTEX_EVENTRECORDING
+#include "genesis/mutex.h"
+#define MAXEVENTS 200
+static struct {
+    struct thread* th;
+    struct timespec ts;
+    char *label;
+    char *mutex_name;
+    sword_t timeout;
+} events[MAXEVENTS];
+static int record_mutex_events;
+static int eventcount;
+
+void lisp_mutex_event(char *string) {
+    if (record_mutex_events) {
+        int id = __sync_fetch_and_add(&eventcount, 1);
+        if (id >= MAXEVENTS) lose("event buffer overflow");
+        clock_gettime(CLOCK_REALTIME, &events[id].ts);
+        events[id].th = get_sb_vm_thread();
+        events[id].label = string;
+        events[id].mutex_name = 0;
+        events[id].timeout = -1;
+    }
+}
+void lisp_mutex_event1(char *string, char *string2) {
+    if (record_mutex_events) {
+        int id = __sync_fetch_and_add(&eventcount, 1);
+        if (id >= MAXEVENTS) lose("event buffer overflow");
+        clock_gettime(CLOCK_REALTIME, &events[id].ts);
+        events[id].th = get_sb_vm_thread();
+        events[id].label = string;
+        events[id].mutex_name = string2;
+        events[id].timeout = -1;
+    }
+}
+void lisp_mutex_event2(char *string, char *string2, uword_t usec) {
+    if (record_mutex_events) {
+        int id = __sync_fetch_and_add(&eventcount, 1);
+        if (id >= MAXEVENTS) lose("event buffer overflow");
+        clock_gettime(CLOCK_REALTIME, &events[id].ts);
+        events[id].th = get_sb_vm_thread();
+        events[id].label = string;
+        events[id].mutex_name = string2;
+        events[id].timeout = usec;
+    }
+}
+void lisp_mutex_start_eventrecording() {
+    eventcount = 0;
+    record_mutex_events = 1;
+}
+void lisp_mutex_done_eventrecording() {
+    record_mutex_events = 0;
+    int i;
+    fprintf(stderr, "event log:\n");
+    struct timespec basetime = events[0].ts;
+    for(i=0; i<eventcount;++i) {
+        struct thread *th = events[i].th;
+        struct thread_instance *ti = (void*)native_pointer(th->lisp_thread);
+        struct timespec rel_time = events[i].ts;
+        rel_time.tv_sec -= basetime.tv_sec;
+        rel_time.tv_nsec -= basetime.tv_nsec;
+        if (rel_time.tv_nsec<0) rel_time.tv_nsec += 1000 * 1000 * 1000, rel_time.tv_sec--;
+        lispobj threadname = ti->name;
+        if (events[i].timeout >= 0) // must also have mutex_name in this case
+            fprintf(stderr, "[%d.%09ld] %s: %s '%s' timeout %ld\n",
+                    (int)rel_time.tv_sec, rel_time.tv_nsec,
+                    (char*)VECTOR(threadname)->data,
+                    events[i].label, events[i].mutex_name, events[i].timeout);
+        else if (events[i].mutex_name)
+            fprintf(stderr, "[%d.%09ld] %s: %s '%s'\n",
+                    (int)rel_time.tv_sec, rel_time.tv_nsec,
+                    (char*)VECTOR(threadname)->data,
+                    events[i].label, events[i].mutex_name);
+        else
+            fprintf(stderr, "[%d.%09ld] %s: %s\n",
+                    (int)rel_time.tv_sec, rel_time.tv_nsec,
+                    (char*)VECTOR(threadname)->data,
+                    events[i].label);
+    }
+    fprintf(stderr, "-----\n");
+}
+#else
+#define lisp_mutex_event(x)
+#define lisp_mutex_event1(x,y)
+#define lisp_mutex_event2(x,y,z)
+#endif
+
 /* values taken from the kernel's linux/futex.h.  This header file
    doesn't exist in userspace, which is our excuse for not grovelling
    them automatically */
@@ -140,14 +221,21 @@ futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
   struct timespec timeout;
   int t;
 
+#ifdef MUTEX_EVENTRECORDING
+    struct mutex* m = (void*)((char*)lock_word - offsetof(struct mutex,state));
+    char *name = m->name != NIL ? (char*)VECTOR(m->name)->data : "(unnamed)";
+#endif
   if (sec<0) {
+      lisp_mutex_event1("start futex wait", name);
       t = sys_futex(lock_word, futex_wait_op(), oldval, 0);
   }
   else {
       timeout.tv_sec = sec;
       timeout.tv_nsec = usec * 1000;
+      lisp_mutex_event2("start futex timedwait", name, usec);
       t = sys_futex(lock_word, futex_wait_op(), oldval, &timeout);
   }
+  lisp_mutex_event1("back from sys_futex", name);
   if (t==0)
       return 0;
   else if (errno==ETIMEDOUT)
@@ -162,23 +250,23 @@ futex_wait(int *lock_word, int oldval, long sec, unsigned long usec)
 int
 futex_wake(int *lock_word, int n)
 {
+#ifdef MUTEX_EVENTRECORDING
+    struct mutex* m = (void*)((char*)lock_word - offsetof(struct mutex,state));
+    char *name = m->name != NIL ? (char*)VECTOR(m->name)->data : "(unnamed)";
+    lisp_mutex_event1("waking futex", name);
+#endif
     return sys_futex(lock_word, futex_wake_op(),n,0);
 }
 #endif
 
 
-void os_init(char __attribute__((unused)) *argv[],
-             char __attribute__((unused)) *envp[])
+void os_init()
 {
-#if defined(LISP_FEATURE_SB_THREAD) && defined(LISP_FEATURE_SB_FUTEX) && !defined(LISP_FEATURE_SB_PTHREAD_FUTEX)
+#ifdef LISP_FEATURE_SB_FUTEX
     futex_init();
 #endif
-
-#ifdef LISP_FEATURE_X86
-    /* Use SSE detector.  Recent versions of Linux enable SSE support
-     * on SSE capable CPUs.  */
-    /* FIXME: Are there any old versions that does not support SSE?  */
-    fast_bzero_pointer = fast_bzero_detect;
+#ifdef LISP_FEATURE_SB_DEVEL
+    prctl(PR_SET_PTRACER, PR_SET_PTRACER_ANY);
 #endif
 }
 
@@ -191,6 +279,28 @@ void os_init(char __attribute__((unused)) *argv[],
 
 int os_preinit(char *argv[], char *envp[])
 {
+#ifdef LISP_FEATURE_RISCV
+    extern int riscv_user_emulation, mmap_does_not_zero, sigaction_does_not_mask;
+    /* Accomodate buggy mmap() emulation, but detect up front whether it may be.
+     * Full system emulation running a RISCV kernel is generally fine. User mode is not.
+     * There's no way to know what it _will_ do, so we have to guess based on
+     * whether the emulation looks bad. */
+    char buf[100];
+    FILE *f = fopen("/proc/cpuinfo", "r");
+    fgets(buf, sizeof buf, f);
+    fgets(buf, sizeof buf, f);
+    if (!strstr(buf, "hart")) { // look for "hardware thread" string
+        fprintf(stderr, "WARNING: enabling mmap() workaround. GC time may be affected\n");
+        rewind(f);
+        fprintf(stderr, "Contents of /proc/cpuinfo:\n");
+        while (fgets(buf, sizeof buf, f) && strlen(buf)>1) fprintf(stderr, " | %s", buf);
+        fprintf(stderr, "----\n");
+        riscv_user_emulation = 1;
+        mmap_does_not_zero = 1;
+        sigaction_does_not_mask = 1;
+    }
+    fclose(f);
+#endif
 
 #if ALLOW_PERSONALITY_CHANGE
     if (getenv("SBCL_IS_RESTARTING")) {
@@ -254,13 +364,18 @@ int os_preinit(char *argv[], char *envp[])
 void
 os_protect(os_vm_address_t address, os_vm_size_t length, os_vm_prot_t prot)
 {
+#ifdef LISP_FEATURE_SOFT_CARD_MARKS
+    // dynamic space should not have protections manipulated
+    if (find_page_index(address) >= 0)
+        lose("unexpected call to os_protect with software card marks");
+#endif
     if (mprotect(address, length, prot)) {
         if (errno == ENOMEM) {
             lose("An mprotect call failed with ENOMEM. This probably means that the maximum amount\n"
                  "of separate memory mappings was exceeded. To fix the problem, either increase\n"
                  "the maximum with e.g. 'echo 262144 > /proc/sys/vm/max_map_count' or recompile\n"
-                 "SBCL with a larger value for GENCGC-CARD-BYTES in\n"
-                 "'src/compiler/target/backend-parms.lisp'.");
+                 "SBCL with a larger value for GENCGC-PAGE-BYTES in\n"
+                 "'src/compiler/"SBCL_TARGET_ARCHITECTURE_STRING"/parms.lisp'.");
         } else {
             perror("mprotect");
         }
@@ -290,32 +405,16 @@ sigsegv_handler(int signal, siginfo_t *info, os_context_t *context)
 {
     os_vm_address_t addr = arch_get_bad_addr(signal, info, context);
 
-#ifdef LISP_FEATURE_ALPHA
-    /* Alpha stuff: This is the end of a pseudo-atomic section during
-       which a signal was received.  We must deal with the pending
-       interrupt (see also interrupt.c, ../code/interrupt.lisp)
-
-       (how we got here: when interrupting, we set bit 63 in reg_ALLOC.
-       At the end of the atomic section we tried to write to reg_ALLOC,
-       got a SIGSEGV (there's nothing mapped there) so ended up here. */
-    if (addr != NULL &&
-        *os_context_register_addr(context, reg_ALLOC) & (1L<<63)) {
-        *os_context_register_addr(context, reg_ALLOC) -= (1L<<63);
-        interrupt_handle_pending(context);
-        return;
-    }
-#endif
-
 #ifdef LISP_FEATURE_SB_SAFEPOINT
-    if (!handle_safepoint_violation(context, addr))
+    if (handle_safepoint_violation(context, addr)) return;
 #endif
 
 #ifdef LISP_FEATURE_GENCGC
-    if (!gencgc_handle_wp_violation(addr))
+    if (gencgc_handle_wp_violation(addr)) return;
 #else
-    if (!cheneygc_handle_wp_violation(context, addr))
+    if (cheneygc_handle_wp_violation(context, addr)) return;
 #endif
-        if (!handle_guard_page_triggered(context, addr))
+    if (!handle_guard_page_triggered(context, addr))
             sbcl_fallback_sigsegv_handler(signal, info, context);
 }
 
@@ -323,22 +422,8 @@ void
 os_install_interrupt_handlers(void)
 {
     if (INSTALL_SIG_MEMORY_FAULT_HANDLER) {
-    undoably_install_low_level_interrupt_handler(SIG_MEMORY_FAULT,
-                                                 sigsegv_handler);
+    ll_install_handler(SIG_MEMORY_FAULT, sigsegv_handler);
     }
-
-    /* OAOOM c.f. sunos-os.c.
-     * Should we have a reusable function gc_install_interrupt_handlers? */
-#ifdef LISP_FEATURE_SB_THREAD
-# ifdef LISP_FEATURE_SB_SAFEPOINT
-#  ifdef LISP_FEATURE_SB_THRUPTION
-    undoably_install_low_level_interrupt_handler(SIGPIPE, thruption_handler);
-#  endif
-# else
-    undoably_install_low_level_interrupt_handler(SIG_STOP_FOR_GC,
-                                                 sig_stop_for_gc_handler);
-# endif
-#endif
 }
 
 char *os_get_runtime_executable_path()
@@ -354,61 +439,3 @@ char *os_get_runtime_executable_path()
 
     return copied_string(path);
 }
-
-#ifdef LISP_FEATURE_SB_WTIMER
-/*
- * Waitable timer implementation for the safepoint-based (SIGALRM-free)
- * timer facility using timerfd_create().
- */
-int
-os_create_wtimer()
-{
-    int fd = timerfd_create(CLOCK_MONOTONIC, 0);
-    if (fd == -1)
-        lose("os_create_wtimer: timerfd_create");
-
-    /* Cannot count on TFD_CLOEXEC availability, so do it manually: */
-    if (fcntl(fd, F_SETFD, FD_CLOEXEC) == -1)
-        lose("os_create_wtimer: fcntl");
-
-    return fd;
-}
-
-int
-os_wait_for_wtimer(int fd)
-{
-    unsigned char buf[8];
-    int n = read(fd, buf, sizeof(buf));
-    if (n == -1) {
-        if (errno == EINTR)
-            return -1;
-        lose("os_wtimer_listen failed");
-    }
-    if (n != sizeof(buf))
-        lose("os_wtimer_listen read too little");
-    return 0;
-}
-
-void
-os_close_wtimer(int fd)
-{
-    if (close(fd) == -1)
-        lose("os_close_wtimer failed");
-}
-
-void
-os_set_wtimer(int fd, int sec, int nsec)
-{
-    struct itimerspec spec = { {0,0}, {0,0} };
-    spec.it_value.tv_sec = sec;
-    spec.it_value.tv_nsec = nsec;
-    if (timerfd_settime(fd, 0, &spec, 0) == -1)
-        lose("timerfd_settime");
-}
-
-void
-os_cancel_wtimer(int fd)
-{
-    os_set_wtimer(fd, 0, 0);
-}
-#endif
