@@ -47,7 +47,7 @@
 #include "interr.h"             /* for lose() */
 #include "alloc.h"
 #include "gc-internal.h"
-#include "getallocptr.h"
+#include "pseudo-atomic.h"
 #include "interrupt.h"
 #include "lispregs.h"
 
@@ -88,11 +88,7 @@ static pthread_mutex_t in_gc_lock = PTHREAD_MUTEX_INITIALIZER;
 #endif
 
 #if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
-extern lispobj call_into_lisp_first_time(lispobj fun, lispobj *args, int nargs)
-# ifdef LISP_FEATURE_X86_64
-    __attribute__((sysv_abi))
-# endif
-    ;
+extern lispobj call_into_lisp_first_time(lispobj fun, lispobj *args, int nargs);
 #endif
 
 static void
@@ -116,12 +112,14 @@ unlink_thread(struct thread *th)
         th->next->prev = th->prev;
 }
 
-#define get_thread_state(thread) (int)__sync_val_compare_and_swap(&thread->state, -1, -1)
+#define get_thread_state(thread) \
+ (int)__sync_val_compare_and_swap(&thread->state_word.state, -1, -1)
 
 #ifndef LISP_FEATURE_SB_SAFEPOINT
 
 void
-set_thread_state(struct thread *thread, lispobj state,
+set_thread_state(struct thread *thread,
+                 char state,
                  boolean signals_already_blocked) // for foreign thread
 {
     struct extra_thread_data *semaphores = thread_extra_data(thread);
@@ -131,7 +129,7 @@ set_thread_state(struct thread *thread, lispobj state,
     if (!signals_already_blocked)
         block_blockable_signals(&old);
     os_sem_wait(&semaphores->state_sem, "set_thread_state");
-    if (thread->state != state) {
+    if (thread->state_word.state != state) {
         if ((STATE_STOPPED==state) ||
             (STATE_DEAD==state)) {
             waitcount = semaphores->state_not_running_waitcount;
@@ -146,7 +144,7 @@ set_thread_state(struct thread *thread, lispobj state,
             for (i=0; i<waitcount; i++)
                 os_sem_post(&semaphores->state_not_stopped_sem, "set_thread_state (not stopped)");
         }
-        thread->state = state;
+        thread->state_word.state = state;
     }
     os_sem_post(&semaphores->state_sem, "set_thread_state");
     if (!signals_already_blocked)
@@ -205,14 +203,33 @@ int sb_GetTID() { return syscall(SYS_gettid); }
 #elif defined __DragonFly__
 #include <sys/lwp.h>
 lwpid_t sb_GetTID() { return lwp_gettid(); }
+#elif defined __FreeBSD__
+#include <sys/thr.h>
+int sb_GetTID()
+{
+    long id;
+    thr_self(&id);
+    // man thr_self(2) says: the thread identifier is an integer in the range
+    // from PID_MAX + 2 (100001) to INT_MAX. So casting to int is safe.
+    return (int)id;
+}
 #else
 #define sb_GetTID() 0
 #endif
 
+/* Our futex-based lisp mutex needs an OS-assigned unique ID.
+ * Why not use pthread_self? I think the reason is that that on linux,
+ * the TID is 4 bytes, and the futex lock word is 4 bytes.
+ * If the unique ID needed 8 bytes, there could be spurious aliasing
+ * that would make the code behave incorrectly. */
 static int get_nonzero_tid()
 {
     int tid = sb_GetTID();
-    gc_dcheck(tid != 0);
+#ifdef LISP_FEATURE_SB_FUTEX
+    // If no futexes, don't need or want to assert that the TID is valid.
+    // (macOS etc)
+    gc_assert(tid != 0);
+#endif
     return tid;
 }
 
@@ -231,6 +248,12 @@ struct thread *alloc_thread_struct(void*,lispobj);
     DuplicateHandle(GetCurrentProcess(), GetCurrentThread(), \
                     GetCurrentProcess(), (LPHANDLE)&thread->os_thread, 0, TRUE, \
                     DUPLICATE_SAME_ACCESS)
+#elif defined LISP_FEATURE_GS_SEG
+#include <asm/prctl.h>
+#include <sys/prctl.h>
+extern int arch_prctl(int code, unsigned long *addr);
+#define ASSOCIATE_OS_THREAD(thread) arch_prctl(ARCH_SET_GS, (uword_t*)thread), \
+      thread->os_thread = thread_self()
 #else
 #define ASSOCIATE_OS_THREAD(thread) thread->os_thread = thread_self()
 #endif
@@ -246,15 +269,42 @@ struct thread *alloc_thread_struct(void*,lispobj);
 #endif
 
 #ifdef LISP_FEATURE_WIN32
-// Need a function callable from assembly code. The inline one won't do.
-// This is basically pthread_getspecific without an argument.
-void* get_current_vm_thread() {
-  return arch_os_get_current_thread();
+// Need a function callable from assembly code, where the inline one won't do.
+void* read_current_thread() {
+  return get_sb_vm_thread();
 }
 #endif
 
 #if defined LISP_FEATURE_DARWIN && defined LISP_FEATURE_SB_THREAD
-    extern pthread_key_t sigwait_bug_mitigation;
+extern pthread_key_t foreign_thread_ever_lispified;
+#endif
+
+#if !defined COLLECT_GC_STATS && !defined STANDALONE_LDB && \
+  defined LISP_FEATURE_LINUX && defined LISP_FEATURE_SB_THREAD && defined LISP_FEATURE_64_BIT
+#define COLLECT_GC_STATS
+#endif
+#ifdef COLLECT_GC_STATS
+static struct timespec gc_start_time;
+static long stw_elapsed,
+    stw_min_duration = LONG_MAX, stw_max_duration, stw_sum_duration,
+    gc_min_duration = LONG_MAX, gc_max_duration, gc_sum_duration;
+int show_gc_stats, n_gcs_done;
+static void summarize_gc_stats(void) {
+    // TODO: also collect things like number of root pages,bytes scanned
+    // and number of pages,bytes copied on average per GC cycle.
+    if (show_gc_stats && n_gcs_done)
+        fprintf(stderr,
+                "\nGC: stw_delay=%ld,%ld,%ld \u00B5s (min,avg,max) pause=%ld,%ld,%ld \u00B5s (sum=%ld) over %d GCs\n",
+                stw_min_duration/1000, stw_sum_duration/n_gcs_done/1000, stw_max_duration/1000,
+                gc_min_duration/1000, gc_sum_duration/n_gcs_done/1000, gc_max_duration/1000,
+                gc_sum_duration/1000, n_gcs_done);
+}
+void reset_gc_stats() { // after sb-posix:fork
+    stw_min_duration = LONG_MAX; stw_max_duration = stw_sum_duration = 0;
+    gc_min_duration = LONG_MAX; gc_max_duration = gc_sum_duration = 0;
+    n_gcs_done = 0;
+    show_gc_stats = 1; // won't show if never called reset
+}
 #endif
 
 void create_main_lisp_thread(lispobj function) {
@@ -266,14 +316,15 @@ void create_main_lisp_thread(lispobj function) {
     struct thread *th = alloc_thread_struct(0, NO_TLS_VALUE_MARKER_WIDETAG);
     if (!th || arch_os_thread_init(th)==0 || !init_shared_attr_object())
         lose("can't create initial thread");
+    th->state_word.sprof_enable = 1;
 #if defined LISP_FEATURE_SB_THREAD && !defined LISP_FEATURE_GCC_TLS && !defined LISP_FEATURE_WIN32
     pthread_key_create(&specials, 0);
 #endif
 #if defined LISP_FEATURE_DARWIN && defined LISP_FEATURE_SB_THREAD
-    pthread_key_create(&sigwait_bug_mitigation, 0);
+    pthread_key_create(&foreign_thread_ever_lispified, 0);
 #endif
 #if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
-    lispobj *args = NULL;
+    __attribute__((unused)) lispobj *args = NULL;
 #endif
     ASSOCIATE_OS_THREAD(th);
     ASSIGN_CURRENT_THREAD(th);
@@ -300,6 +351,9 @@ void create_main_lisp_thread(lispobj function) {
     set_thread_stack(th->control_stack_end);
 #endif
 
+#ifdef COLLECT_GC_STATS
+    atexit(summarize_gc_stats);
+#endif
     /* WIN32 has a special stack arrangement, calling
      * call_into_lisp_first_time will put the new stack in the middle
      * of the current stack */
@@ -309,6 +363,14 @@ void create_main_lisp_thread(lispobj function) {
 #else
     funcall0(function);
 #endif
+    // If we end up returning, clean up the initial thread.
+#ifdef LISP_FEATURE_SB_THREAD
+    unlink_thread(th);
+#else
+    all_threads = NULL;
+#endif
+    arch_os_thread_cleanup(th);
+    ASSIGN_CURRENT_THREAD(NULL);
 }
 
 #ifdef LISP_FEATURE_SB_THREAD
@@ -350,10 +412,10 @@ init_new_thread(struct thread *th,
 #ifdef LISP_FEATURE_SB_SAFEPOINT
     csp_around_foreign_call(th) = (lispobj)scribble;
 #endif
-    lock_ret = thread_mutex_lock(&all_threads_lock);
-    gc_assert(lock_ret == 0);
+    lock_ret = mutex_acquire(&all_threads_lock);
+    gc_assert(lock_ret);
     link_thread(th);
-    thread_mutex_unlock(&all_threads_lock);
+    ignore_value(mutex_release(&all_threads_lock));
 
     /* Kludge: Changed the order of some steps between the safepoint/
      * non-safepoint versions of this code.  Can we unify this more?
@@ -372,55 +434,37 @@ unregister_thread(struct thread *th,
 {
     int lock_ret;
 
-    /* Kludge: Changed the order of some steps between the safepoint/
-     * non-safepoint versions of this code.  Can we unify this more?
-     */
+    block_blockable_signals(0);
+    gc_close_thread_regions(th);
 #ifdef LISP_FEATURE_SB_SAFEPOINT
-
-    block_blockable_signals(0);
-    ensure_region_closed(&th->alloc_region, BOXED_PAGE_FLAG);
     pop_gcing_safety(&scribble->safety);
-    lock_ret = thread_mutex_lock(&all_threads_lock);
-    gc_assert(lock_ret == 0);
-    unlink_thread(th);
-    lock_ret = thread_mutex_unlock(&all_threads_lock);
-    gc_assert(lock_ret == 0);
-
 #else
-
-    /* Block GC */
-    block_blockable_signals(0);
     /* This state change serves to "acknowledge" any stop-the-world
      * signal received while the STOP_FOR_GC signal is blocked */
     set_thread_state(th, STATE_DEAD, 1);
-
+#endif
     /* SIG_STOP_FOR_GC is blocked and GC might be waiting for this
      * thread, but since we are either exiting lisp code as a lisp
      * thread that is dying, or exiting lisp code to return to
      * former status as a C thread, it won't wait long. */
-    lock_ret = thread_mutex_lock(&all_threads_lock);
-    gc_assert(lock_ret == 0);
-
-    /* FIXME: this nests the free_pages_lock inside the all_threads_lock.
-     * There's no reason for that, so closing of regions should be done
-     * sooner to eliminate an ordering constraint. */
-    ensure_region_closed(&th->alloc_region, BOXED_PAGE_FLAG);
+    lock_ret = mutex_acquire(&all_threads_lock);
+    gc_assert(lock_ret);
     unlink_thread(th);
-    thread_mutex_unlock(&all_threads_lock);
-    gc_assert(lock_ret == 0);
-
-#endif
+    lock_ret = mutex_release(&all_threads_lock);
+    gc_assert(lock_ret);
 
     arch_os_thread_cleanup(th);
 
+    __attribute__((unused)) struct extra_thread_data *semaphores = thread_extra_data(th);
+#ifdef LISP_FEATURE_UNIX
+    os_sem_destroy(&semaphores->sprof_sem);
+#endif
 #ifndef LISP_FEATURE_SB_SAFEPOINT
-    struct extra_thread_data *semaphores = thread_extra_data(th);
     os_sem_destroy(&semaphores->state_sem);
     os_sem_destroy(&semaphores->state_not_running_sem);
     os_sem_destroy(&semaphores->state_not_stopped_sem);
 #endif
 
-
 #ifdef LISP_FEATURE_MACH_EXCEPTION_HANDLER
     mach_lisp_thread_destroy(th);
 #endif
@@ -433,7 +477,7 @@ unregister_thread(struct thread *th,
 #endif
 
     /* Undo the association of the current pthread to its `struct thread',
-     * such that we can call arch_os_get_current_thread() later in this
+     * such that we can call get_sb_vm_thread() later in this
      * thread and cleanly get back NULL. */
     /* FIXME: what if, after we blocked signals, someone uses INTERRUPT-THREAD
      * on this thread? It's no longer a lisp thread; I suspect the signal
@@ -473,6 +517,7 @@ void* new_thread_trampoline(void* arg)
     // 'th->lisp_thread' remains valid despite not being in all_threads
     // due to the pinning via *STARTING-THREADS*.
     struct thread_instance *lispthread = (void*)native_pointer(th->lisp_thread);
+    if (lispthread->_ephemeral_p == T) th->state_word.user_thread_p = 0;
 
     /* Potentially set the externally-visible name of this thread,
      * and for a whole pile of crazy, look at get_max_thread_name_length_impl() in
@@ -486,14 +531,14 @@ void* new_thread_trampoline(void* arg)
          *  restricted to 16 characters, including the terminating null byte ('\0').
          *  The pthread_setname_np() function can fail with the following error:
          *  ERANGE The length of the string ... exceeds the allowed limit." */
-        if (fixnum_value(VECTOR(name)->length) <= 15)
+        if (vector_len(VECTOR(name)) <= 15)
             pthread_setname_np(pthread_self(), (char*)VECTOR(name)->data);
 #endif
 #ifdef LISP_FEATURE_NETBSD
         /* This constant is an upper bound on the length including the NUL.
          * Exceeding it will fail the call. It happens to be 32.
          * Also, don't want to printf-format a name containing a '%' */
-        if (fixnum_value(VECTOR(name)->length) < PTHREAD_MAX_NAMELEN_NP)
+        if (vector_len(VECTOR(name)) < PTHREAD_MAX_NAMELEN_NP)
             pthread_setname_np(pthread_self(), "%s", (char*)VECTOR(name)->data);
 #endif
 #if defined LISP_FEATURE_FREEBSD || defined LISP_FEATURE_OPENBSD
@@ -502,7 +547,7 @@ void* new_thread_trampoline(void* arg)
         pthread_set_name_np(pthread_self(), (char*)VECTOR(name)->data);
 #endif
 #ifdef LISP_FEATURE_DARWIN
-        if (fixnum_value(VECTOR(name)->length) < 64)
+        if (vector_len(VECTOR(name)) < 64)
             pthread_setname_np((char*)VECTOR(name)->data);
 #endif
     }
@@ -556,9 +601,7 @@ void* new_thread_trampoline(void* arg)
                     GUARD_CONTROL_STACK|GUARD_BINDING_STACK|GUARD_ALIEN_STACK);
     funcall0(function);
     unregister_thread(th, &scribble);
-#ifdef LISP_FEATURE_WIN32
     free_thread_struct(th); // no recycling of 'struct thread'
-#endif
 
 #endif
     return 0;
@@ -577,23 +620,23 @@ static struct thread* get_recyclebin_item()
 {
     struct thread* result = 0;
     int rc;
-    rc = thread_mutex_lock(&recyclebin_lock);
-    gc_assert(!rc);
+    rc = mutex_acquire(&recyclebin_lock);
+    gc_assert(rc);
     if (recyclebin_threads) {
         result = recyclebin_threads;
         recyclebin_threads = result->next;
     }
-    thread_mutex_unlock(&recyclebin_lock);
+    ignore_value(mutex_release(&recyclebin_lock));
     return result ? result->os_address : 0;
 }
 static void put_recyclebin_item(struct thread* th)
 {
     int rc;
-    rc = thread_mutex_lock(&recyclebin_lock);
-    gc_assert(!rc);
+    rc = mutex_acquire(&recyclebin_lock);
+    gc_assert(rc);
     th->next = recyclebin_threads;
     recyclebin_threads = th;
-    thread_mutex_unlock(&recyclebin_lock);
+    ignore_value(mutex_release(&recyclebin_lock));
 }
 void empty_thread_recyclebin()
 {
@@ -601,11 +644,7 @@ void empty_thread_recyclebin()
     sigset_t old;
     block_deferrable_signals(&old);
     // no big deal if already locked (recursive GC?)
-#ifdef LISP_FEATURE_WIN32
     if (TryEnterCriticalSection(&recyclebin_lock)) {
-#else
-    if (!pthread_mutex_trylock(&recyclebin_lock)) {
-#endif
         struct thread* this = recyclebin_threads;
         while (this) {
             struct thread* next = this->next;
@@ -613,7 +652,7 @@ void empty_thread_recyclebin()
             this = next;
         }
         recyclebin_threads = 0;
-        thread_mutex_unlock(&recyclebin_lock);
+        ignore_value(mutex_release(&recyclebin_lock));
     }
     thread_sigmask(SIG_SETMASK, &old, 0);
 }
@@ -688,8 +727,10 @@ static void attach_os_thread(init_thread_data *scribble)
 
 static void detach_os_thread(init_thread_data *scribble)
 {
-    struct thread *th = arch_os_get_current_thread();
-
+    struct thread *th = get_sb_vm_thread();
+#ifdef LISP_FEATURE_DARWIN
+    pthread_setspecific(foreign_thread_ever_lispified, (void*)1);
+#endif
     unregister_thread(th, scribble);
 
     /* We have to clear a STOP_FOR_GC signal if pending. Consider:
@@ -706,28 +747,13 @@ static void detach_os_thread(init_thread_data *scribble)
      *  - but STOP_FOR_GC is pending because it was in the blocked set.
      * Bad things happen unless we clear the pending GC signal.
      */
-#ifndef LISP_FEATURE_SB_SAFEPOINT
+#if !defined LISP_FEATURE_SB_SAFEPOINT && !defined LISP_FEATURE_DARWIN
     sigset_t pending;
     sigpending(&pending);
     if (sigismember(&pending, SIG_STOP_FOR_GC)) {
         int sig, rc;
         rc = sigwait(&gc_sigset, &sig);
         gc_assert(rc == 0 && sig == SIG_STOP_FOR_GC);
-#ifdef LISP_FEATURE_DARWIN
-        sigpending(&pending);
-        if (sigismember(&pending, SIG_STOP_FOR_GC)) {
-            // fprintf(stderr, "Trying sigwait bug mitigation\n");
-            pthread_setspecific(sigwait_bug_mitigation, (void*)1);
-            sigfillset(&pending);
-            sigdelset(&pending, SIG_STOP_FOR_GC);
-            // This might hang forever now, because the signal disappears
-            // despite that we just observed it to be pending.
-            // Basically you're screwed one way or the other - either
-            // by a spurious signal or a lost signal.
-            sigsuspend(&pending);
-            // fprintf(stderr, "Back from sigsuspend\n");
-        }
-#endif
     }
 #endif
     put_recyclebin_item(th);
@@ -755,7 +781,7 @@ callback_wrapper_trampoline(
 #endif
     lispobj arg0, lispobj arg1, lispobj arg2)
 {
-    struct thread* th = arch_os_get_current_thread();
+    struct thread* th = get_sb_vm_thread();
     if (!th) {                  /* callback invoked in non-lisp thread */
         init_thread_data scribble;
         attach_os_thread(&scribble);
@@ -830,10 +856,6 @@ callback_wrapper_trampoline(
 
 struct thread *
 alloc_thread_struct(void* spaces, lispobj start_routine) {
-#if defined(LISP_FEATURE_SB_THREAD) || defined(LISP_FEATURE_WIN32)
-    unsigned int i;
-#endif
-
     /* May as well allocate all the spaces at once: it saves us from
      * having to decide what to do if only some of the allocations
      * succeed. SPACES must be appropriately aligned, since the GC
@@ -851,7 +873,7 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
         // if any newly started thread could refer a dead thread's heap objects.
         zeroize_stack = 1;
     } else {
-        spaces = os_validate(MOVABLE|IS_THREAD_STRUCT, NULL, THREAD_STRUCT_SIZE);
+        spaces = os_validate(MOVABLE|IS_THREAD_STRUCT, NULL, THREAD_STRUCT_SIZE, 0, 0);
         if (!spaces) return NULL;
     }
     /* Aligning up is safe as THREAD_STRUCT_SIZE has
@@ -863,7 +885,7 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
     // Refer to the ASCII art in the block comment above
     struct thread *th = (void*)(csp_page + THREAD_CSP_PAGE_SIZE
                                 + THREAD_HEADER_SLOTS*N_WORD_BYTES);
-    __attribute((unused)) lispobj* tls = (lispobj*)th;
+
 #ifdef LISP_FEATURE_SB_SAFEPOINT
     // Out of caution I'm supposing that the last thread to use this memory
     // might have left this page as read-only. Could it? I have no idea.
@@ -871,10 +893,15 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
 #endif
 
 #ifdef LISP_FEATURE_SB_THREAD
-    for(i = 0; i < (unsigned int)(dynamic_values_bytes/N_WORD_BYTES); i++)
-        tls[i] = NO_TLS_VALUE_MARKER_WIDETAG;
-    th->lisp_thread = 0; // force it to be always-thread-local, of course
+    memset(th, 0, sizeof *th);
+    lispobj* ptr = (lispobj*)(th + 1);
+    lispobj* end = (lispobj*)((char*)th + dynamic_values_bytes);
+    while (ptr < end) *ptr++ = NO_TLS_VALUE_MARKER_WIDETAG;
     th->tls_size = dynamic_values_bytes;
+#endif
+    __attribute((unused)) lispobj* tls = (lispobj*)th;
+#ifdef THREAD_T_NIL_CONSTANTS_SLOT
+    tls[THREAD_T_NIL_CONSTANTS_SLOT] = (NIL << 32) | T;
 #endif
 #if defined LISP_FEATURE_X86_64 && defined LISP_FEATURE_LINUX
     tls[THREAD_MSAN_XOR_CONSTANT_SLOT] = 0x500000000000;
@@ -882,17 +909,12 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
 #ifdef LAYOUT_OF_FUNCTION
     tls[THREAD_FUNCTION_LAYOUT_SLOT] = LAYOUT_OF_FUNCTION << 32;
 #endif
-#ifdef LISP_FEATURE_GENCGC
 #ifdef THREAD_VARYOBJ_CARD_MARKS_SLOT
     extern unsigned int* varyobj_page_touched_bits;
     tls[THREAD_VARYOBJ_SPACE_ADDR_SLOT] = VARYOBJ_SPACE_START;
     tls[THREAD_VARYOBJ_CARD_COUNT_SLOT] = varyobj_space_size / IMMOBILE_CARD_BYTES;
     tls[THREAD_VARYOBJ_CARD_MARKS_SLOT] = (lispobj)varyobj_page_touched_bits;
 #endif
-    th->dynspace_addr       = DYNAMIC_SPACE_START;
-    th->dynspace_card_count = page_table_pages;
-    th->dynspace_pte_base   = (lispobj)page_table;
-#endif
 
     th->os_address = spaces;
     th->control_stack_start = (lispobj*)aligned_spaces;
@@ -926,7 +948,7 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
 #endif
     }
 
-    th->control_stack_guard_page_protected = T;
+    th->state_word.control_stack_guard_page_protected = 1;
     th->alien_stack_start=
         (lispobj*)((char*)th->binding_stack_start+BINDING_STACK_SIZE);
     set_binding_stack_pointer(th,th->binding_stack_start);
@@ -949,8 +971,16 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
     os_sem_init(&extra_data->state_not_running_sem, 0);
     os_sem_init(&extra_data->state_not_stopped_sem, 0);
 #endif
+#if defined LISP_FEATURE_UNIX && defined LISP_FEATURE_SB_THREAD
+    os_sem_init(&extra_data->sprof_sem, 0);
+#endif
+    extra_data->sprof_lock = 0;
+    th->sprof_data = 0;
+
+    th->state_word.state = STATE_RUNNING;
+    th->state_word.sprof_enable = 0;
+    th->state_word.user_thread_p = 1;
 
-    th->state=STATE_RUNNING;
 #ifdef ALIEN_STACK_GROWS_DOWNWARD
     th->alien_stack_pointer=(lispobj*)((char*)th->alien_stack_start
                                        + ALIEN_STACK_SIZE-N_WORD_BYTES);
@@ -958,15 +988,16 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
     th->alien_stack_pointer=(lispobj*)((char*)th->alien_stack_start);
 #endif
 
-#ifdef LISP_FEATURE_SB_THREAD
-    th->pseudo_atomic_bits=0;
+#ifdef HAVE_THREAD_PSEUDO_ATOMIC_BITS_SLOT
+    memset(&th->pseudo_atomic_bits, 0, sizeof th->pseudo_atomic_bits);
 #elif defined LISP_FEATURE_GENCGC
     clear_pseudo_atomic_atomic(th);
     clear_pseudo_atomic_interrupted(th);
 #endif
 
 #ifdef LISP_FEATURE_GENCGC
-    gc_init_region(&th->alloc_region);
+    gc_init_region(&th->mixed_tlab);
+    gc_init_region(&th->cons_tlab);
 #endif
 #ifdef LISP_FEATURE_SB_THREAD
     /* This parallels the same logic in globals.c for the
@@ -974,7 +1005,7 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
      * all. */
 #if defined(LISP_FEATURE_X86) || defined(LISP_FEATURE_X86_64)
     th->foreign_function_call_active = 0;
-#else
+#elif !defined(LISP_FEATURE_ARM64) // uses control_stack_start
     th->foreign_function_call_active = 1;
 #endif
 #endif
@@ -1000,9 +1031,18 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
 
     thread_interrupt_data(th).pending_handler = 0;
     thread_interrupt_data(th).gc_blocked_deferrables = 0;
-#if GENCGC_IS_PRECISE
+#if HAVE_ALLOCATION_TRAP_CONTEXT
     thread_interrupt_data(th).allocation_trap_context = 0;
 #endif
+#if defined LISP_FEATURE_PPC64
+    /* Storing a 0 into code coverage mark bytes or GC card mark bytes
+     * can be done from the low byte of the thread base register.
+     * The thread alignment is BACKEND_PAGE_BYTES (from thread.h), but seeing as this is
+     * a similar-but-different requirement, it pays to double-check */
+    if ((lispobj)th & 0xFF) lose("Thread struct not at least 256-byte-aligned");
+    extern unsigned char* gc_card_mark;
+    th->card_table = (lispobj)gc_card_mark;
+#endif
 
 #ifdef LISP_FEATURE_SB_THREAD
 // This macro is the same as "write_TLS(sym,val,th)" but can't be spelled thus.
@@ -1018,6 +1058,7 @@ alloc_thread_struct(void* spaces, lispobj start_routine) {
     th->no_tls_value_marker = start_routine;
 
 #if defined(LISP_FEATURE_WIN32)
+    int i;
     for (i = 0; i<NUM_PRIVATE_EVENTS; ++i)
         thread_private_events(th,i) = CreateEvent(NULL,FALSE,FALSE,NULL);
     thread_extra_data(th)->synchronous_io_handle_and_flag = 0;
@@ -1034,7 +1075,7 @@ uword_t create_thread(struct thread_instance* instance, lispobj start_routine)
     struct thread *th;
 
     /* Must defend against async unwinds. */
-    if (read_TLS(INTERRUPTS_ENABLED, arch_os_get_current_thread()) != NIL)
+    if (read_TLS(INTERRUPTS_ENABLED, get_sb_vm_thread()) != NIL)
         lose("create_thread is not safe when interrupts are enabled.");
 
     /* Assuming that a fresh thread struct has no lisp objects in it,
@@ -1068,7 +1109,7 @@ uword_t create_thread(struct thread_instance* instance, lispobj start_routine)
                      CREATE_SUSPENDED, &tid);
     boolean success = th->os_thread != 0;
     if (success) {
-        instance->primitive_thread = (lispobj)th;
+        instance->uw_primitive_thread = (lispobj)th;
         th->os_kernel_tid = tid;
         ResumeThread((HANDLE)th->os_thread);
     } else {
@@ -1079,13 +1120,8 @@ uword_t create_thread(struct thread_instance* instance, lispobj start_routine)
 }
 #endif
 
-#ifdef LISP_FEATURE_WIN32
 int try_acquire_gc_lock() { return TryEnterCriticalSection(&in_gc_lock); }
-void release_gc_lock() { LeaveCriticalSection(&in_gc_lock); }
-#else
-int try_acquire_gc_lock() { return !pthread_mutex_trylock(&in_gc_lock); }
-void release_gc_lock() { pthread_mutex_unlock(&in_gc_lock); }
-#endif
+int release_gc_lock() { return mutex_release(&in_gc_lock); }
 
 /* stopping the world is a two-stage process.  From this thread we signal
  * all the others with SIG_STOP_FOR_GC.  The handler for this signal does
@@ -1095,7 +1131,7 @@ void release_gc_lock() { pthread_mutex_unlock(&in_gc_lock); }
 /*
  * (With SB-SAFEPOINT, see the definitions in safepoint.c instead.)
  */
-#ifndef LISP_FEATURE_SB_SAFEPOINT
+#if !defined LISP_FEATURE_SB_SAFEPOINT && !defined STANDALONE_LDB
 
 /* To avoid deadlocks when gc stops the world all clients of each
  * mutex must enable or disable SIG_STOP_FOR_GC for the duration of
@@ -1122,35 +1158,36 @@ void release_gc_lock() { pthread_mutex_unlock(&in_gc_lock); }
  * the memory shouldn't be there) */
 void gc_stop_the_world()
 {
-    struct thread *th, *me = arch_os_get_current_thread();
+#ifdef COLLECT_GC_STATS
+    struct timespec stw_begin_time, stw_end_time;
+    // Measuring the wait time has to use a realtime clock, not a thread clock
+    // because sleeping below in a sem_wait needs to accrue time.
+    clock_gettime(CLOCK_MONOTONIC, &stw_begin_time);
+#endif
+    struct thread *th, *me = get_sb_vm_thread();
     int rc;
 
     /* Keep threads from registering with GC while the world is stopped. */
-    rc = thread_mutex_lock(&all_threads_lock);
-    gc_assert(rc == 0);
+    rc = mutex_acquire(&all_threads_lock);
+    gc_assert(rc);
 
     /* stop all other threads by sending them SIG_STOP_FOR_GC */
     for_each_thread(th) {
-        gc_assert(th->os_thread != 0);
-        /* We were going to extremes here to use acquire semantics on the state,
-         * while NOT ACTUALLY PREVENTING CHANGE OF THAT STATE. i.e. we released the
-         * state lock before returning with the determined state. That's no help.
-         * Just do an atomic load and we're good */
-        int state = get_thread_state(th);
-        if (th != me && state == STATE_RUNNING) {
-            /* This pthread_kill is safe.
-             * A RUNNING thread can transition to DEAD but can't disappear either
-             * in terms of 'struct thread' being unmapped or the pthread exiting,
-             * because it'll be blocked on the all_threads lock */
-            rc = pthread_kill(th->os_thread,SIG_STOP_FOR_GC);
-            /* This used to bogusly check for ESRCH.
-             * I changed the ESRCH case to just fall into lose() */
-            if (rc) {
-                lose("cannot suspend thread %p: %d, %s",
+        if (th != me) {
+            gc_assert(th->os_thread != 0);
+            struct extra_thread_data *semaphores = thread_extra_data(th);
+            os_sem_wait(&semaphores->state_sem, "notify stop");
+            int state = get_thread_state(th);
+            if (state == STATE_RUNNING) {
+                rc = pthread_kill(th->os_thread,SIG_STOP_FOR_GC);
+                /* This used to bogusly check for ESRCH.
+                 * I changed the ESRCH case to just fall into lose() */
+                if (rc) lose("cannot suspend thread %p: %d, %s",
                      // KLUDGE: assume that os_thread can be cast as pointer.
                      // See comment in 'interr.h' about that.
                      (void*)th->os_thread, rc, strerror(rc));
             }
+            os_sem_post(&semaphores->state_sem, "notified stop");
         }
     }
     for_each_thread(th) {
@@ -1160,11 +1197,35 @@ void gc_stop_the_world()
         }
     }
     FSHOW_SIGNAL((stderr,"/gc_stop_the_world:end\n"));
+#ifdef COLLECT_GC_STATS
+    clock_gettime(CLOCK_MONOTONIC, &stw_end_time);
+    stw_elapsed = (stw_end_time.tv_sec - stw_begin_time.tv_sec)*1000000000L
+                + (stw_end_time.tv_nsec - stw_begin_time.tv_nsec);
+    gc_start_time = stw_end_time;
+#endif
 }
 
 void gc_start_the_world()
 {
-    struct thread *th, *me = arch_os_get_current_thread();
+#ifdef COLLECT_GC_STATS
+    struct timespec gc_end_time;
+    clock_gettime(CLOCK_MONOTONIC, &gc_end_time);
+    long gc_elapsed = (gc_end_time.tv_sec - gc_start_time.tv_sec)*1000000000L
+                      + (gc_end_time.tv_nsec - gc_start_time.tv_nsec);
+    if (stw_elapsed < 0 || gc_elapsed < 0) {
+        char errmsg[] = "GC: Negative times?\n";
+        ignore_value(write(2, errmsg, sizeof errmsg-1));
+    } else {
+        stw_sum_duration += stw_elapsed;
+        if (stw_elapsed < stw_min_duration) stw_min_duration = stw_elapsed;
+        if (stw_elapsed > stw_max_duration) stw_max_duration = stw_elapsed;
+        gc_sum_duration += gc_elapsed;
+        if (gc_elapsed < gc_min_duration) gc_min_duration = gc_elapsed;
+        if (gc_elapsed > gc_max_duration) gc_max_duration = gc_elapsed;
+        ++n_gcs_done;
+    }
+#endif
+    struct thread *th, *me = get_sb_vm_thread();
     int lock_ret;
     /* if a resumed thread creates a new thread before we're done with
      * this loop, the new thread will be suspended waiting to acquire
@@ -1184,8 +1245,8 @@ void gc_start_the_world()
         }
     }
 
-    lock_ret = thread_mutex_unlock(&all_threads_lock);
-    gc_assert(lock_ret == 0);
+    lock_ret = mutex_release(&all_threads_lock);
+    gc_assert(lock_ret);
 }
 
 #endif /* !LISP_FEATURE_SB_SAFEPOINT */
@@ -1237,8 +1298,8 @@ void wake_thread(struct thread_instance* lispthread)
          * that leave_region safepoint will acknowledge the signal, so
          * there is no need to take locks, roll thread to safepoint
          * etc. */
-        struct thread* thread = (void*)lispthread->primitive_thread;
-        if (thread == arch_os_get_current_thread()) {
+        struct thread* thread = (void*)lispthread->uw_primitive_thread;
+        if (thread == get_sb_vm_thread()) {
             sb_pthr_kill(thread, 1); // can't fail
             check_pending_thruptions(NULL);
             return;
@@ -1247,17 +1308,17 @@ void wake_thread(struct thread_instance* lispthread)
         // but without them, make-target-contrib hangs in bsd-sockets.
         sigset_t oldset;
         block_deferrable_signals(&oldset);
-        thread_mutex_lock(&all_threads_lock);
+        mutex_acquire(&all_threads_lock);
         sb_pthr_kill(thread, 1); // can't fail
-# ifdef LISP_FEATURE_SB_THRUPTION
+# ifdef LISP_FEATURE_SB_SAFEPOINT
         wake_thread_impl(lispthread);
 # endif
-        thread_mutex_unlock(&all_threads_lock);
+        mutex_release(&all_threads_lock);
         thread_sigmask(SIG_SETMASK,&oldset,0);
-#elif defined LISP_FEATURE_SB_THRUPTION
+#elif defined LISP_FEATURE_SB_SAFEPOINT
     wake_thread_impl(lispthread);
 #else
-    pthread_kill(lispthread->os_thread, SIGURG);
+    pthread_kill(lispthread->uw_os_thread, SIGURG);
 #endif
 }
 #endif
