@@ -26,7 +26,7 @@
 #include "interr.h"
 #include "breakpoint.h"
 #include "thread.h"
-#include "getallocptr.h"
+#include "pseudo-atomic.h"
 #include "unaligned.h"
 #include "search.h"
 #include "var-io.h"
@@ -160,7 +160,7 @@ arch_get_bad_addr(int __attribute__((unused)) sig,
  */
 
 os_context_register_t *
-context_eflags_addr(os_context_t *context)
+os_context_flags_addr(os_context_t *context)
 {
 #if defined __linux__
     /* KLUDGE: As of kernel 2.2.14 on Red Hat 6.2, there's code in the
@@ -312,7 +312,7 @@ arch_do_displaced_inst(os_context_t *context, unsigned int orig_inst)
     *(pc-2) = 0x00240c81;
     *(pc-1) = 0x9d000001;
 #else
-    *context_eflags_addr(context) |= 0x100;
+    *os_context_flags_addr(context) |= 0x100;
 #endif
 
     single_stepping = pc;
@@ -355,7 +355,7 @@ restore_breakpoint_from_single_step(os_context_t * context)
     *(single_stepping-2) = single_step_save2;
     *(single_stepping-1) = single_step_save3;
 #else
-    *context_eflags_addr(context) &= ~0x100;
+    *os_context_flags_addr(context) &= ~0x100;
 #endif
     /* Re-install the breakpoint if possible. */
     if (((char *)OS_CONTEXT_PC(context) > (char *)single_stepping) &&
@@ -611,18 +611,18 @@ lispobj fdefn_callee_lispobj(struct fdefn* fdefn) {
     // This logic is rather nasty, but I don't know what else to do.
     lispobj word;
     if (header_widetag(word = raw_addr[-2]) == SIMPLE_FUN_WIDETAG
-        || (word == 1 &&
+        || (word == FORWARDING_HEADER &&
             widetag_of(native_pointer(forwarding_pointer_value(raw_addr-2)))
             == SIMPLE_FUN_WIDETAG))
         return make_lispobj(raw_addr - 2, FUN_POINTER_LOWTAG);
     int widetag;
     if ((widetag = header_widetag(word = raw_addr[-4])) == CODE_HEADER_WIDETAG
-        || (word == 1 &&
+        || (word == FORWARDING_HEADER &&
             widetag_of(native_pointer(forwarding_pointer_value(raw_addr-4)))
             == CODE_HEADER_WIDETAG))
         return make_lispobj(raw_addr - 4, OTHER_POINTER_LOWTAG);
     if (widetag == FUNCALLABLE_INSTANCE_WIDETAG
-        || (word == 1 &&
+        || (word == FORWARDING_HEADER &&
             widetag_of(native_pointer(forwarding_pointer_value(raw_addr-4)))
             == FUNCALLABLE_INSTANCE_WIDETAG))
         return make_lispobj(raw_addr - 4, FUN_POINTER_LOWTAG);
@@ -788,4 +788,16 @@ allocation_tracker_sized(uword_t* sp)
     gc_assert(ret);
 }
 
+__attribute__((sysv_abi)) lispobj call_into_lisp(lispobj fun, lispobj *args, int nargs) {
+    extern lispobj call_into_lisp_(lispobj, lispobj *, int, struct thread *)
+        __attribute__((sysv_abi));
+    return call_into_lisp_(fun, args, nargs, get_sb_vm_thread());
+}
+
+lispobj call_into_lisp_first_time(lispobj fun, lispobj *args, int nargs) {
+    extern lispobj call_into_lisp_first_time_(lispobj, lispobj *, int, struct thread *)
+        __attribute__((sysv_abi));
+    return call_into_lisp_first_time_(fun, args, nargs, get_sb_vm_thread());
+}
+
 #include "x86-arch-shared.inc"
