@@ -23,10 +23,11 @@
 #include "interr.h"
 #include "breakpoint.h"
 #include "thread.h"
-#include "getallocptr.h"
+#include "pseudo-atomic.h"
 #include "forwarding-ptr.h"
 #include "var-io.h"
 #include "code.h"
+#include "unaligned.h"
 
 #include "genesis/static-symbols.h"
 #include "genesis/symbol.h"
@@ -53,7 +54,7 @@ arch_get_bad_addr(int sig, siginfo_t *code, os_context_t *context)
  */
 
 int *
-context_eflags_addr(os_context_t *context)
+os_context_flags_addr(os_context_t *context)
 {
 #if defined __linux__
     /* KLUDGE: As of kernel 2.2.14 on Red Hat 6.2, there's code in the
@@ -89,7 +90,7 @@ void arch_skip_instruction(os_context_t *context)
     int code;
 
     /* Get and skip the Lisp interrupt code. */
-    code = *(char*)(*os_context_pc_addr(context))++;
+    code = *(char*)(OS_CONTEXT_PC(context)++);
     switch (code)
         {
         case trap_Error:
@@ -118,33 +119,32 @@ void arch_skip_instruction(os_context_t *context)
         }
 
     FSHOW((stderr,
-           "/[arch_skip_inst resuming at %x]\n",
-           *os_context_pc_addr(context)));
+           "/[arch_skip_inst resuming at %x]\n", OS_CONTEXT_PC(context)));
 }
 
 unsigned char *
 arch_internal_error_arguments(os_context_t *context)
 {
-    return 1 + (unsigned char *)(*os_context_pc_addr(context));
+    return 1 + (unsigned char *)(OS_CONTEXT_PC(context));
 }
 
 boolean
 arch_pseudo_atomic_atomic(os_context_t *context)
 {
-    return get_pseudo_atomic_atomic(arch_os_get_current_thread());
+    return get_pseudo_atomic_atomic(get_sb_vm_thread());
 }
 
 void
 arch_set_pseudo_atomic_interrupted(os_context_t *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     set_pseudo_atomic_interrupted(thread);
 }
 
 void
 arch_clear_pseudo_atomic_interrupted(os_context_t *context)
 {
-    struct thread *thread = arch_os_get_current_thread();
+    struct thread *thread = get_sb_vm_thread();
     clear_pseudo_atomic_interrupted(thread);
 }
 
@@ -180,7 +180,7 @@ unsigned int  single_step_save3;
 void
 arch_do_displaced_inst(os_context_t *context, unsigned int orig_inst)
 {
-    unsigned int *pc = (unsigned int*)(*os_context_pc_addr(context));
+    unsigned int *pc = (unsigned int*)OS_CONTEXT_PC(context);
 
     /* Put the original instruction back. */
     arch_remove_breakpoint(pc, orig_inst);
@@ -195,13 +195,13 @@ arch_do_displaced_inst(os_context_t *context, unsigned int orig_inst)
     *(pc-2) = 0x00240c81;
     *(pc-1) = 0x9d000001;
 #else
-    *context_eflags_addr(context) |= 0x100;
+    *os_context_flags_addr(context) |= 0x100;
 #endif
 
     single_stepping = pc;
 
 #ifdef CANNOT_GET_TO_SINGLE_STEP_FLAG
-    *os_context_pc_addr(context) = (os_context_register_t)((char *)pc - 9);
+    OS_CONTEXT_PC(context) = (os_context_register_t)((char *)pc - 9);
 #endif
 }
 
@@ -215,13 +215,11 @@ restore_breakpoint_from_single_step(os_context_t * context)
     *(single_stepping-2) = single_step_save2;
     *(single_stepping-1) = single_step_save3;
 #else
-    *context_eflags_addr(context) &= ~0x100;
+    *os_context_flags_addr(context) &= ~0x100;
 #endif
     /* Re-install the breakpoint if possible. */
-    if (((char *)*os_context_pc_addr(context) >
-         (char *)single_stepping) &&
-        ((char *)*os_context_pc_addr(context) <=
-         (char *)single_stepping + BREAKPOINT_WIDTH)) {
+    if (((char *)OS_CONTEXT_PC(context) > (char *)single_stepping) &&
+        ((char *)OS_CONTEXT_PC(context) <= (char *)single_stepping + BREAKPOINT_WIDTH)) {
         fprintf(stderr, "warning: couldn't reinstall breakpoint\n");
     } else {
         arch_install_breakpoint(single_stepping);
@@ -234,16 +232,15 @@ restore_breakpoint_from_single_step(os_context_t * context)
 void
 arch_handle_breakpoint(os_context_t *context)
 {
-    *os_context_pc_addr(context) -= BREAKPOINT_WIDTH;
+    OS_CONTEXT_PC(context) -= BREAKPOINT_WIDTH;
     handle_breakpoint(context);
 }
 
 void
 arch_handle_fun_end_breakpoint(os_context_t *context)
 {
-    *os_context_pc_addr(context) -= BREAKPOINT_WIDTH;
-    *os_context_pc_addr(context) =
-        (int)handle_fun_end_breakpoint(context);
+    OS_CONTEXT_PC(context) -= BREAKPOINT_WIDTH;
+    OS_CONTEXT_PC(context) = (int)handle_fun_end_breakpoint(context);
 }
 
 void
@@ -268,7 +265,7 @@ sigtrap_handler(int signal, siginfo_t *info, os_context_t *context)
 
     /* This is just for info in case the monitor wants to print an
      * approximation. */
-    access_control_stack_pointer(arch_os_get_current_thread()) =
+    access_control_stack_pointer(get_sb_vm_thread()) =
         (lispobj *)*os_context_sp_addr(context);
 
 #ifdef LISP_FEATURE_SUNOS
@@ -286,20 +283,20 @@ sigtrap_handler(int signal, siginfo_t *info, os_context_t *context)
      * 'kind' value (eg trap_Cerror). For error-trap and Cerror-trap a
      * number of bytes will follow, the first is the length of the byte
      * arguments to follow. */
-    trap = *(unsigned char *)(*os_context_pc_addr(context));
+    trap = *(unsigned char *)OS_CONTEXT_PC(context);
     handle_trap(context, trap);
 }
 
 void
 sigill_handler(int signal, siginfo_t *siginfo, os_context_t *context) {
 #ifndef LISP_FEATURE_MACH_EXCEPTION_HANDLER
-    if (*((unsigned short *)*os_context_pc_addr(context)) == UD2_INST) {
-        *os_context_pc_addr(context) += 2;
+    if (*(unsigned short *)OS_CONTEXT_PC(context) == UD2_INST) {
+        OS_CONTEXT_PC(context) += 2;
         return sigtrap_handler(signal, siginfo, context);
     }
 #endif
     fake_foreign_function_call(context);
-    lose("Unhandled SIGILL at %p.", (void*)*os_context_pc_addr(context));
+    lose("Unhandled SIGILL at %p.", (void*)OS_CONTEXT_PC(context));
 }
 #endif /* not LISP_FEATURE_WIN32 */
 
@@ -385,7 +382,8 @@ gencgc_apply_code_fixups(struct code *old_code, struct code *new_code)
 void
 arch_write_linkage_table_entry(int index, void *target_addr, int datap)
 {
-    char *reloc_addr = (char*)LINKAGE_TABLE_SPACE_START + index * LINKAGE_TABLE_ENTRY_SIZE;
+    // 'volatile' works around a spurious GCC warning
+    volatile char *reloc_addr = (char*)LINKAGE_TABLE_SPACE_START + index * LINKAGE_TABLE_ENTRY_SIZE;
     if (datap) {
         *(unsigned long *)reloc_addr = (unsigned long)target_addr;
         return;
@@ -405,3 +403,5 @@ arch_write_linkage_table_entry(int index, void *target_addr, int datap)
     /* write a nop for good measure. */
     *reloc_addr = 0x90;
 }
+
+#include "x86-arch-shared.inc"
